
/*
	dispatcher executes the needed actions at the scheduled events.
	Called from eventqueue.c
	Author: Marcell Gal	
	
	Table-configurable port_act->op; was implemented upto revision 1.39
	It was never used, but might be in the future, so check it out
	before reinventing. (some changes might be wise).
	Please keep this comment.

	Now that we have the dispatcher_arr[] table it would make sense
	to have every function in the *.c file where it belongs.
*/

#include "global.h"
#include "heap.h"
#include "fuelcontrol.h"
#include "eventqueue.h"
#include "dispatcher.h"
#include "hwconf.h"
#include "comm.h"
#include "benchmark.h"
#include "wbo2.h"
#include "knock.h"
#include "lbus.c"

#include <inttypes.h>
#include <avr/signal.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "edis.c"
#include "injector.c"

/*!
 * \ingroup g_event_queues
 * \brief Dispatcher executes the needed actions at the scheduled events.
 *
 * TODO: 
 * The functions could return heaplast, 
 * or the fact that heap ROOT element was deleted or not
 * \return none
 */

#ifdef GENBOARDv3
void disp_inj_start(uint8_t upper_v){	// INJ_START
   // take care, check dispatcher_arr
	injector_start_destroy(upper_v);
}

void disp_inj_pwm(uint8_t upper_v){		// INJ_PWM:
	pwminj(upper_v);

	if (!COND_STOP(inj_port[upper_v].stopat)) {
		event_insert_destroy(inj_port[upper_v].stopat, v_shiftup(upper_v) + INJ_STOP);
	} else { // otherwise we would schedule a very long PWM-ed period :-))))
		stopinj(upper_v);
		heap_del1st(heap_normal);
	}
}

void disp_inj_stop(uint8_t upper_v){	// INJ_STOP:
	if (COND_STOP(inj_port[upper_v].stopat)) {
		stopinj(upper_v);
	}
	heap_del1st(heap_normal);
}

void disp_inj_all(uint8_t upper_v){		// INJ_ALL:
	switch (upper_v) {
	case INJ_ALL_START:
		injector_start_all();
		heap_del1st(heap_normal);
		break;

	case INJ_ALL_PWM:
		pwminj_all();

		if (!COND_STOP(inj_port[0].stopat)) {
			event_insert_destroy(inj_port[0].stopat, v_shiftup(INJ_ALL_STOP) + INJ_ALL);
		} else { // otherwise we would schedule a very long PWM-ed period :-))))
			stopinj_all();
			heap_del1st(heap_normal);
		}
		break;

	case INJ_ALL_STOP:
		stopinj_all();
		heap_del1st(heap_normal);
		break;

	default:
		heap_del1st(heap_normal);
		break;
	}
}

void disp_knock_start(uint8_t upper_v){	// KNOCK_START:
#ifdef KNOCK_8101
	// do a final check to see if the integration should be started (window_length could have changed in the mean time)
	if (knock.window_length != 0xffff) {
		PORTC |= _BV(KNOCK_INTEGRATE);
		// take care, knock IC ignores spi activity while KNOCK_INTEGRATE is high
		knock_set_state(KNOCK_WINDOW_START);
		knock.igncount = upper_v;
		event_insert_destroy(etime.now + knock.window_length, KNOCK_STOP);
	} else {
		heap_del1st(heap_normal);
	}
#endif
}

void disp_knock_stop(uint8_t upper_v){		// KNOCK_STOP:
#ifdef KNOCK_8101
	PORTC &= ~_BV(KNOCK_INTEGRATE);
	knock_nextstate();
	heap_del1st(heap_normal);
#endif
}

#else	// genboard v2

void disp_injastop(uint8_t upper_v){	// INJASTOP:
	if(COND_STOP(inj_port[0].stopat))
		stopinj1();
	heap_del1st(heap_normal);
}

void disp_injbstop(uint8_t upper_v){	// INJBSTOP:	
	if(COND_STOP(inj_port[1].stopat))
		stopinj2();
	heap_del1st(heap_normal);
}

void disp_injagowpm(uint8_t upper_v){	// INJAGOPWM:
	injAgopwm();
	if(!COND_STOP(inj_port[0].stopat)){
		event_insert_destroy(inj_port[0].stopat, INJASTOP);
	} else { // otherwise we would schedule a very long PWM-ed period :-))))
		stopinj1();
		heap_del1st(heap_normal);
	}
}

void disp_injbgopwm(uint8_t upper_v){	// INJBGOPWM:
	injBgopwm();
	if(!COND_STOP(inj_port[1].stopat)){
		event_insert_destroy(inj_port[1].stopat, INJBSTOP);
	} else {
		stopinj2();
		heap_del1st(heap_normal);
	}
}
#endif	// not GENBOARDv3

// #ifndef NOIGN
void disp_edis_stop_saw(uint8_t upper_v){	// EDIS_STOP_SAW:
	ign_deactivate();
	heap_del1st(heap_normal);	
	// no need to insert an event (next PIP pulse will) 
}

void disp_ignact(uint8_t upper_v){	// IGNACT:
		// now one pin, later the right channel:
		ign_activate();
#ifdef KNOCK_8101
		// check that the mainloop can keep up
		if (knock_stats.ve_igncount != engine.igncount)
			knock_stats.dwell_err++;
#endif
		// only schedule the firing if it's not yet done:
		if(engine.ignstate != 3){
			if(engine.ignstate == 1){
				event_insert_destroy(engine.sparkat, IGNDEACT);
				engine.ignstate = 2;
#ifdef V3_IGN_DRIVE_A
				test_ign_deactivate_at(engine.sparkat);
#endif
			} else {	// state is probably 2, IGNDEACT already scheduled
				// state cannot be 0, but just in case:
				if(engine.ignstate==0){
					ign_deactivate();
				}
				heap_del1st(heap_normal);
				// do not touch ignstate
			}
		} else { // ign_trigger_here() will schedule IGNDEACT more precisely than we could
			heap_del1st(heap_normal);
			engine.ignstate = 4;
			// if trigger is stopped, dummyign coils would fry?
			// see megasquirt.c protection in sevenms that
			// deactivates ignition.
			// Releasing a spark at arbitrary time is also not very good.
			// sudden outage of trigger signal at high RPM 
			// (only case when we can get here) is not good in general.
		}
}

void disp_igndeact(uint8_t upper_v){	// IGNDEACT:
	uint8_t t8;
		// London's burning:
		ign_deactivate();

#ifdef KNOCK_8101
		// check that the mainloop can keep up
		if (knock_stats.ve_igncount != engine.igncount)
			knock_stats.adv_err++;
// t8 is used by the knock detection to determine which cylinder is fired.
		t8 = engine.igncount; // remember which cylinder is on fire
#endif

		if( engine.status_trig & _BV(cam_sync2) ){
			engine.status_trig &= ~_BV(cam_sync2);
			engine.igncount = config.ignchmax;			
		} else if(engine.igncount == 0){
			engine.igncount = config.ignchmax;
		} else {
			engine.igncount--;
		}

#ifdef BENCHMARK
		benchmark_histogram(ic3_histogram, etime.now - engine.sparkat, IC3HISTOGRAM_BUCKETS);
#endif
		// engine.ignstate is 2 here, I guess
		// always let 400 usec for off:
#define IGN_MIN_INACTIVE	(100)
		// early_dwellstart only set if last rpm_period was small or trig2spark was small
		if((engine.status_ext & _BV(early_dwellstart)) 
		&& ((engine.rpm_period >> 16)==0)){
			uint16_t inactive_time, per, de;
#ifdef LATE_CORRECTION
			uint8_t late;
		
			// we are a bit late: should be 0..10 for 0..40 usec
			late = (uint8_t)etime.now - (uint8_t)engine.sparkat;
			// this can never hold. "now" is always after sparkat
			if(late > 20)	// prevents overflow. TODO: drop this, never can cause big problem
				late = 0;
#define late_correction ((uint16_t)late)
#else
#define late_correction (0)
#endif
			de = engine.dwell + late_correction;		//de is "Dwell Enlarged"
			per = (uint16_t)engine.rpm_period;
			if(per > (de + IGN_MIN_INACTIVE))	// enough time for the user and VBatt determined dwell:
				inactive_time = per - de;
			else	// we  shorten the dwell if we don't have enough time:
				inactive_time = IGN_MIN_INACTIVE;

//			event_insert_destroy(engine.sparkat + inactive_time, IGNACT);
			event_insert_destroy(etime.now + inactive_time, IGNACT);
#ifdef V3_IGN_DRIVE_A
			test_ign_activate_at(engine.sparkat + inactive_time);
#endif

#ifdef KNOCK_8101
			if (knock_ready())
				event_insert(etime.now + knock.window_start, v_shiftup(t8) + KNOCK_START);
#endif
			engine.ignstate = 3;
		} else {
#ifdef KNOCK_8101
			if (knock_ready())
				event_insert_destroy(etime.now + knock.window_start, v_shiftup(t8) + KNOCK_START);
			else
#endif
				heap_del1st(heap_normal);

			engine.ignstate = 0;
		}
}
// #endif	// NOIGN

void disp_stim_low(uint8_t upper_v){	// STIM_LOW:
		if( debug.stim_period >= 20 ){
			uint16_t newtime;
			uint8_t dstn;
#ifdef GENBOARDv3
			digitalout(config.act_wot_channel, DLOW);
#else	// on v2 using panel LED for stim-out
			LEDPORT &= ~_BV(LED_STATUS);
#endif
			newtime = debug.stim_nexttime;
			dstn = debug.stim_toothnow;
			if( dstn == 0 ){
				// let's insert some missing teeth
				uint8_t halfperiods = debug.stim_tooth >> 5;
				// we want mst.._upper2bits => halfperiods like 1 => 3 and 2 => 5 so:
				halfperiods |= 1; // redundant for 36-1 or 60-2, only needed for 24-1
				dstn = debug.stim_tooth & 63; // -1 was needed for old behavior 
				newtime += mult16_8(debug.stim_period, halfperiods );
			} else {
				dstn--;
				newtime += debug.stim_period;
			}

			event_insert_destroy(newtime, STIM_HIGH);
			// save local variables to global:
			debug.stim_nexttime = newtime;
			debug.stim_toothnow = dstn;
		} else {
			heap_del1st(heap_normal);
		}
}

void disp_stim_high(uint8_t upper_v){	// STIM_HIGH:
	uint16_t newtime;
		if(debug.stim_period >= 32){
#ifdef GENBOARDv3
			digitalout(config.act_wot_channel, DHIGH);
#else	// on v2 using panel LED for stim-out
			LEDPORT |= _BV(LED_STATUS);
#endif
			newtime = debug.stim_nexttime;
#ifdef BENCHMARK
			// 4: small offset so slightly negative values don't overflow
			benchmark_histogram(evq_histogram, 4+etime.now - newtime, EVQHISTOGRAM_BUCKETS);
#endif
			newtime += debug.stim_period;
			event_insert_destroy(newtime, STIM_LOW);
			debug.stim_nexttime = newtime;
		} else {
			heap_del1st(heap_normal);
		}
}

void disp_ptrig_reenable(uint8_t upper_v){	// PRIMARY_TRIGGER_REENABLE:
#ifdef GENBOARDv3
		TIFR = _BV(ICF1);  // timer1
		TIMSK |= _BV(TICIE1);
#else
		ETIFR = _BV(ICF3);  // timer3
		ETIMSK |= _BV(TICIE3);
#endif
		heap_del1st(heap_normal);
}

void disp_strig_reenable(uint8_t upper_v){	// SECONDARY_TRIGGER_REENABLE:
#ifdef GENBOARDv3
		ETIFR = _BV(ICF3);  // timer3
		ETIMSK |= _BV(TICIE3);
#else
		EIFR = _BV(INTF6);  // int6
		EIMSK |= _BV(INT6);
#endif
		heap_del1st(heap_normal);
}

void disp_event_benchmark(uint8_t upper_v){	// EVENT_BENCHMARK:
	uint16_t newtime;
	// we insert dummy events if number of events < 7
	if (heap_normal->heaplast < 7) {
		newtime = rand();
		if (newtime > 16000)
			newtime = etime.now + (newtime & 0x1FF);	// we want small delays to be higher probability
		event_insert_destroy(newtime, EVENT_BENCHMARK);
	} else {
		heap_del1st(heap_normal);
	}
}
		
void disp_boostcontrol_actuate(uint8_t upper_v){	// EVENT_BOOSTACT:
#ifdef BOOSTCONTROL
	// we have to take care of duty inversion 
	// (eliminate effect of race condition)
	if( COND_STOP( engine.boost_stopat ) ){
		digitalout( config.boost_channel, DLOW );
	}
#endif
	heap_del1st(heap_normal);
}

void disp_eventbad(uint8_t upper_v){	// BAD EVENT:
	heap_del1st(heap_normal);
#ifdef BENCHMARK
	benchstats.unknownevent++;
#endif
}

#ifdef GENBOARDv3
PGM_VOID_P dispatcher_arr[] PROGMEM = {
	disp_ignact, disp_igndeact, disp_ptrig_reenable, disp_strig_reenable,
	disp_stim_low,	disp_stim_high, disp_event_benchmark, disp_edis_stop_saw,
	injector_start_destroy, disp_inj_pwm, disp_inj_stop, disp_inj_all,
	disp_knock_start, disp_knock_stop, disp_boostcontrol_actuate, disp_eventbad
	};
#else // genboard v2
PGM_VOID_P dispatcher_arr[] PROGMEM = {	
	disp_ignact, disp_igndeact, disp_ptrig_reenable, disp_strig_reenable,
	disp_stim_low,	disp_stim_high, disp_event_benchmark, disp_edis_stop_saw,
	disp_injastop, disp_injbstop, disp_injagowpm, disp_injbgopwm,
	disp_eventbad, disp_eventbad, disp_eventbad, disp_eventbad
	};
#endif // not GENBOARDv3

