
/* - Efficient eventqueue routines
 * Marcell Gal
 * license: GPL
 */

   /*!
   * \addtogroup g_event_queues Event Queues
   */
/*@{*/


#ifndef __EVENTQUEUE_C__
#define __EVENTQUEUE_C__

#include "global.h"
#include "heap.h"
#include "eventqueue.h"
#include "fuelcontrol.h"
#include "dispatcher.h"
#include "kb.h"
#include "iac.h"

#ifdef BENCHMARK
#include "benchmark.h"
#endif

#include <inttypes.h>
#include <avr/signal.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>

extern volatile struct engine_t engine;
struct heap_t h1;

#define event_not_too_far(x) (1)

/* 
 * For the most efficient use of the eventqueue (eg. from dispatcher()) do:
 * event_session_start();
 * event_insert_insert_takecare(); (any number of times) Will delete heap_normal HEAPROOT max once!
 * event_session_end(); will delete heap_normal HEAPROOT if necessary
 *
 * OR optionally (I prefer the previous):
 * event_insert_destroy()  just once
 * event_insert() any number of times
 * 
 * This is a slight optimization since heap_replace1st(heap_normal,k,v) is 
 * faster than heap_del1st(heap_normal) and heap_insert(heap_normal, k, v);
 * BUT in dispatcher we want to destroy heap_normal's HEAPROOT just once.
 * Therefore when inserting several events, we have to track
 * if one of the previous inserts have deleted the heap_normal HEAPROOT...
 */

struct etime_t etime;

/* Called from dispatcher:
 * inserts an event.
 * Does NOT destroy heap_normal's HEAPROOT like the event_insert_destroy()
 */
void event_insert(key_t k, val_t v)
{
	if(event_not_too_far(k)) {
		heap_insert(heap_normal, k, v);
	}
}

static uint8_t heap_hn_root_deleted ;

// This can be left out since this variable is reset at _init and 
// also in event_session_end()
void event_session_start(void)
{
	// heap_hn_root_deleted = 0;
};

// deletes the HEAPROOT if heap_normal if appropriate
void event_insert_takecare(key_t k, val_t v)
{
	if(event_not_too_far(k)) {
		if(heap_hn_root_deleted){
			heap_insert(heap_normal, k, v);
		} else {
			heap_hn_root_deleted = 1;
			heap_replace1st(heap_normal, k, v);
		}
	}
}

void event_session_end(void)
{
	if(heap_hn_root_deleted == 0){
		heap_del1st(heap_normal); // none of the event_insert_takecare()-s deleted it, so
	} else {
		heap_hn_root_deleted = 0;
	}
}

void event_insert_destroy(key_t k, val_t v)
{
	if(event_not_too_far(k)) {
		heap_replace1st(heap_normal, k, v);
		return;
	} 
	heap_del1st(heap_normal);
}

// this is not very common, but __inline__ does not work across file boundaries
// Why here? There is circular calling between event_insert..() and dispatcher() => both cannot be inlined
// This is the faster "direction", considering calling probability
#include "dispatcher.c"
#include "handler2k.c"

// 16 usec unit (not 4) for TCNT2-TCNT0 hibrid:
#define FORWARD_CNT 1
// 4 usec units for TCNT3:
#define FORWARD3_CNT 4

/*
the forw parameter will cheat, forward time a little

void readTMR_f(uint8_t forw)
{
	uint16_t t0;

	t0 = TCNT16 + forw;

	if((t0>>8) != (etime.h2klast)){
		etime.h2klast = t0>>8;
		if((t0>>8) == 0){
			etime.now_upper ++;
		}
		etime.now = t0;	// handler2k is interested in it
		handler2k();
		t0 = TCNT16 + forw;	// read again
	} 
	etime.now = t0;
}
*/

/*
 MUST be called with interrupts disabled.
 uint16_t etime.now and now_upper are updated
 NOTE: hardware time TCNT is read at the ___beginning and NOT reread___ even if 
 a long action is taken. This is intentional: you 
 get back the time of the beginning of the function call, not at the end

void readTMR(void)
{
	etime.now = TCNT16; // + FORWARD3_CNT;
}
*/

// 12 usec:
#define HANDLER2K_MAX_DELAY 3
// handler2k() run at 0x..05 time. DO NOT set it too high as 
// handler2k() can be executed soon after 0x..00 anyway.
#define HANDLER2K_TIMED_AT 5
#define HANDLER2K_DELAY_COND (HANDLER2K_TIMED_AT + HANDLER2K_MAX_DELAY)

/* we always disable interrupts since the heap is manipulated from interrupt
 * routine. A conflict would result in trashed structures.
 *
 * Operation: after all the current events are served, we set OCR and return.
 * NOTE: the task of (delete current event and insert any wanted events) is the 
 * responsibility of the dispatcher !!!!
 *
 * It seems that the action needed after an event_insert from userspace is
 * exactly the same as when an OCR compare match interrupt was fired.
 */

#ifndef READAHEAD

#ifdef LESS_READTMR
#define READAHEAD 4
// not used: #define READAHEADSHORT 3
#else	// no LESS_READTMR:
	// this branch allows less READAHEAD cause the last minute readTMR() will help:
	// in fact 1 is no "readahead" just compensate for earlieroe16 => earlier16 cheat
#define READAHEAD 1
// 3 or 4 are supposed to be nice values:
#define READAHEADSHORT 4
#endif

#endif // READAHEAD


__inline__ void alarm2core(void)
{
	key_t k, t0;
	uint8_t t0_u; 
	heapindex_t heaplast;

#ifndef HANDLER2K_CHANGES_HEAP
	// this is split from the while, because of the "goto" after handler2k:
	// only dispatcher changes heap currently.
	heaplast = heap_normal->heaplast;
#endif

back:
	while(	
	// + 3 should be enough for everyone: allows (3-1) * 4 usec to reach OCR16 = k;
		( (t0 = readTMR() + READAHEAD) || 1)
		&& (heaplast
#ifdef HANDLER2K_CHANGES_HEAP
			= heap_normal->heaplast
#endif
			) 
		&& earlier16(k = heapkey(heap_normal, HEAPROOT), t0 )){
		void (*func)(uint8_t);
		uint8_t trimmed_v, upper_v;
		val_t v = heapval(heap_normal, HEAPROOT);
		upper_v = v_shiftdown(v);	// retreive upper bits
		trimmed_v = v_trim(v);
		
		// dispatcher(v);
		trimmed_v <<= 1;	// this avoids unnecessary (adc) operation on the upper-byte
		func = (PGM_VOID_P)pgm_read_word(((uint16_t)dispatcher_arr) + trimmed_v);
		func(upper_v);
		
#ifndef HANDLER2K_CHANGES_HEAP
		heaplast = heap_normal->heaplast;
#endif
	}

	// at this point, all t0 and heaplast are uptodate.
	// k also, if(heaplast != 0)

	//t0 = TCNT16;	
	// we must not update etime.now outside readTMR() (hint: etime.now_upper++)

	t0_u = t0>>8;
	// this shouldn't be checked again if already run.
	// Do we gain or lose if we add a local variable for etime.h2klast ?
	// the good solution is to kill handler2k altogether
	if(t0_u != (etime.h2klast)){
		etime.h2klast = t0_u;
		handler2k(); // oops, this might take 9 usec, so be safe:
// #ifdef HANDLER2K_CHANGES_HEAP
		// it is not necessary to go back if handler2k aint change heap
		// since the last minute check will catch any due events
		// It was removed when sloppy msp in menu.c tricked us to
		// believe there is event scheduling into the past
		// Just put it back, do some testing.
		goto back ;
// #endif
	}
	
	t0_u ++;
	// find out when next event/handler2k is due
	// check if event is closer and move handler2k out of the way of events:
	// NEGLECT the "k might be uninitialized" warning, we are cleverer than gcc
	if( heaplast && earlier16(k, ((t0_u << 8) | HANDLER2K_DELAY_COND))){
		// event is close, schedule that
		// k = k; 
	} else { // schedule h2k
		k = (t0_u << 8) | HANDLER2K_TIMED_AT;
	}
	
#ifndef LESS_READTMR
	// last minute readTMR()
	t0 = readTMR() + READAHEADSHORT;
	if(earlier16(k, t0)) goto back;
	// if k is an event (not handler2k), we could jump right to
	// the execution of the event!
#endif

	// if k points to the PAST (hint: readTMR_f() gone), we are screwed
	OCR16 = k;
	// this is the dangerous instruction:
	ETIFR = ETIFR_CLR;	
	// It is intended to clear the flag _after_ setting OCR	
	// to a value in the _past_
	// if the k value is the current time, we are screwed.
}

SIGNAL (SIGNAL_OC16) {
	alarm2core();
}

/* 
relative scheduling, from now. For conveniency.
alarm2core() still needs to be called afterwards,
when dif is very small ( < 256, which is 1024 usec)...
*/
void event_insert_rel(key_t dif, val_t v)
{
	key_t k, now;
	
	readTMR();
	/* another function would be needed to use from userspace (cli/sei).
	it would reenable the irq _here_ for 2-3 nops if  dif>100
	*/
	now = etime.now;
	k = now + dif;
	
	if(event_not_too_far(k)) { // insert to heap_normal
		heap_insert(heap_normal, k, v);
	}
}

void event_insert_destroy_rel(key_t dif, val_t v)
{
	key_t k, now;
	
	readTMR();
	now = etime.now;
	k = now + dif;
	
	if(event_not_too_far(k)) {
		heap_replace1st(heap_normal, k, v);
		return;
	}
	heap_del1st(heap_normal);
}

// deletes the HEAPROOT if heap_normal if appropriate
void event_insert_takecare_rel(key_t dif, val_t v)
{
	key_t k, now;
	
	readTMR();
	now = etime.now;
	k = now + dif;
	
	if(event_not_too_far(k)) { // we need to insert into heap_normal
		if(heap_hn_root_deleted){ //heap_normal HEAPROOT already deleted
			heap_insert(heap_normal, k, v);
		} else {
			heap_hn_root_deleted = 1;
			heap_replace1st(heap_normal, k, v);
		}
	}
}

/* We will pin down TCNT0 to the 1/4 freq of TCNT2 (correct for IDLEPWM_HW as well)
This synchronizes TCNT2 and TCNT0 counters (also resets the other timers: for fun)
*/
void counter_sync(void)
{
	SFIOR |= _BV(TSM) | _BV(PSR321);
	TCNT3=0;
	TCNT2=0; // this isn't strictly required
	TCNT1=0;
	SFIOR &= ~_BV(TSM);
}

void eventqueue_init(void)
{

#if HEAPROOT != 1
#warning "HEAPROOT VALUE is NOT, 1 THIS WILL CAUSE PROBLEM"
#endif

#ifdef	VAL_T_16BIT
#warning "why do you make heap copy the upper byte?"
#endif	// VAL_T_16BIT

	TCCR16B = TCCR16B_VAL ; /* set prescaler to division 64 (4usec) page 133 */
	/* set mode (up, ...) NOTHING to do, using default */
	ETIMSK |= ETIMSK_OR;

	/* reset struct: done by avr-libc (all 0) */
	
	heap_hn_root_deleted = 0;
	// heap_normal = &h1; 
	heap_init(heap_normal); 

#ifdef BENCHMARK
	benchmark_init();
//	event_insert(0x3456, EVENT_BENCHMARK); // must start somewhere
#endif

//	counter_sync();
	TCNT3=0;
	TCNT1=0;

	etime.now = 0;
	// even if we do not insert an event we want eventqueue to start tracking the time
	alarm2core();	// without this the epoch would be offsetted? big deal?
}

#endif

/*@}*/

