   /*!
   * \addtogroup g_heap Memory Heap Routines
   */
/*@{*/

#ifndef __HEAP_C__
#define __HEAP_C__
#include <inttypes.h>
#include "heap.h"

/*
 * inlined from eventqueue.c
 * designed as the base of an eventqueue to efficiently find nearest event
 *
 * author: Marcell Gal
 * copyright: GPL
 */

/* 
 * IMPORTANT: userspace should cli/sei around all functions here  !!
 * an eventqueue will use 2 heaps, since 16 bit time overflows fast.
 * Imagine: now = 37000
 * 38000 is in the future, 6000 is also in the future, actually later.
 * 
 */

STATICHEAP void heap_init(struct heap_t *h)
{
	h->heaplast = HEAPROOT - 1; /* it will be HEAPROOT after adding 1 element */
	h->heapunderrun = h->heapoverrun = 0;
}

/* --------------------
 * heap is 1 based (first index is HEAPROOT) 
 * change if going to base 0
 */
#if HEAPROOT==1
static inline heapindex_t parent(heapindex_t i)
{
	return i >> 1;
}

/* returns child with smaller index */
static inline heapindexx_t child0(heapindex_t i)
{
	return i + i ; 
	/* second child is next */
}
#else 
#warning check that HEAPROOT and subroutines match

#endif

/* 
 * Returns child index that have smaller key,
 * or smaller index child if higher index child is invalid. 
 * Warning: smaller index child not checked against heaplast.

static inline heapindexx_t childsmaller(struct heap_t *h, heapindex_t i, heapindex_t last)
{
	heapindexx_t s, s1;

	s = child0(i);
	if (((s1=s+1) <= last) 
		&& (earlier16(heapkey(h,s1), heapkey(h,s)))){
		return s1;
	}

	return s;
}
*/

static inline void heapitemcopy(struct heap_t *h, heapindex_t from, heapindex_t to)
{
	heapkey(h, to) = heapkey(h, from);
	heapval(h, to) = heapval(h, from);
}

/* Swaps 2 event_t in the heap:
 * first swap the keys, than the vals for less register contamination:
 * check the asm, this could be probably improved
 */
static inline void heapswap(struct heap_t *h, heapindex_t i, heapindex_t ii)
{
	key_t dk;
	val_t dt;
	dk = heapkey(h, ii);
	heapkey(h, ii) = heapkey(h, i);
	heapkey(h, i) = dk;

	dt = heapval(h, ii);
	heapval(h, ii) = heapval(h, i);
	heapval(h, i) = dt;
}

static inline void heap_fix(struct heap_t *p, heapindex_t h, heapindex_t last){
	heapindexx_t hh, s1;	
	do {
		hh = child0(h);
		if(hh > last) return;
		
		s1 = hh + 1;
		if (( s1 <= last) && (earlier16(heapkey(p,s1), heapkey(p,hh)))){
			hh = s1;
		}

		// earlieroe16() could be sligthly faster here, but 
		// we haven't verified earlieroe16() asm yet,
		// and there is no difference unless conflicting events in the heap
		if (earlier16(heapkey(p, h), heapkey(p, hh))) 
			return;

		heapswap(p, hh, h);
		h = hh;
	} while(1);
}

/*
 * inserts an event into the heap
 * From userspace ALWAYS call this with interrupts disabled,
 * since timer interrupt also manipulates events 
 */
STATICHEAP void heap_insert(struct heap_t *h, uint16_t key, uint16_t val) 
{
	heapindex_t hole, hh;

	hole = h->heaplast;
	if (hole >= (HEAPSIZE-1) ){
		h->heapoverrun ++;
		return;		/* drop without much warning */
	}

	hole++; /* still < HEAPSIZE */
	h->heaplast = hole;

	for ( ; (hole > HEAPROOT) && earlier16(key, heapkey(h, hh=parent(hole)))
		; hole = hh )
		heapitemcopy(h, hh, hole);

	heapkey(h, hole) = key;
	heapval(h, hole) = val;
}

/* 
 * Deletes root (smallest key) element.
 * Copies last to root element, and makes corrections if necessary
 */
STATICHEAP void heap_del1st(struct heap_t *p)
{
	heapindex_t h, last;
	h = last = p->heaplast;

	if(h < HEAPROOT) { /* it was empty */
		p->heapunderrun++;
		return;
	}

	last--;
	p->heaplast = last;
	if( last < HEAPROOT) { // now empty; TODO: check if gcc sees this is h==0
		return;
	}

	heapitemcopy(p, h, HEAPROOT);
	heap_fix(p, HEAPROOT, last);
}
/*  heap_del1st and heap_insert in one EFFICIENT step.
 * often used from one event to fire next (eg. cylinder, 
 * or matching inj-stop event at inj-start
 */
STATICHEAP void heap_replace1st(struct heap_t *p, uint16_t key, uint16_t val) 
{
	heapkey(p, HEAPROOT) = key;
	heapval(p, HEAPROOT) = val;
	heap_fix(p, HEAPROOT, p->heaplast);
}

#endif // __HEAP_C__
/*@}*/

