 /***********************************************************
  *   Program:    LCD driver
  *   Author:     Marcell Gal <cell@x-dsl.hu>
  *   Comments:   HITACHI LCD driver
  ************************************************************/

   /*!
   * \addtogroup g_lcd LCD Control Routines
   */
/*@{*/

#include "global.h"
#include "lcd.h"
#include "comm.h"
#include "eventqueue.h"
#include "benchmark.h"

#include <inttypes.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/delay.h>

#include "softpwm.c"
#include "lbus.c"
#include "arithm.c"

char hex[] PROGMEM = "0123456789ABCDEF";
const uint8_t BV_rev_arr[] PROGMEM = {0x80, 0x40, 0x20, 0x10, 8, 4, 2, 1};

struct softpwm_t lcd_backlight;

__inline__ uint8_t hex_hi(uint8_t i)
{
	return PRG_RDB(&hex[i>>4]);
}

__inline__ uint8_t hex_lo(uint8_t i)
{
	return PRG_RDB(&hex[i & 0x0F]);
}

#ifdef LCD
struct lcd_t lcd;

/* 
we start each line with \r\n so on terminal it looks like the real LCD.
Nice when extending menu.c
This implies that LCD_CHARS+2*LCD_LINES bytes must be queried
*/
uint8_t lcd_cache2comm(uint8_t i)
{
	uint8_t cachepos;
// Instead of doing a division we step line by line and see where we are
	for( cachepos = 0; i >= (LCD_COLS+2); cachepos += LCD_COLS, i -= (LCD_COLS+2))
		;
		
	if(i){
		if(i>1)
			return lcd.cache[cachepos + i-2];
		else 
			return '\n';	// pos 1 on terminal line
	} else 
		return '\r';		// pos 0 on terminal line
		
}

/* the char at index should be sent to the LCD */
static __inline__ void dirtyat(lcdpos_t index)
{
	lcd.dirty[index>>3] |= BV_rev_l(index) ;
}

/* the char at index is synchronized with LCD thus sent. - no need to send again -
 * return old flag (Test and Clear, effectively) */
static __inline__ uint8_t TaC(lcdpos_t index)
{
	uint8_t t, ret, mask;
	t = index>>3;
	mask = BV_rev_l(index) ;
	ret = lcd.dirty[t] & mask;
	lcd.dirty[t] &= ~mask;
	return ret;
}

/* checks if the char at index is dirty */
__inline__ char isdirty(lcdpos_t index)
{
	return lcd.dirty[index>>3] & BV_rev_l(index) ;
}

static __inline__ void lcd_pos_to_peek_inc(void)
{
	lcd.pos_to_peek++;
	if (lcd.pos_to_peek >= LCD_CHARS) lcd.pos_to_peek = 0;
}

// setting lcd.pos_to_peek to a dirty position (starting to look from appr. lpk)
static __inline__ void find_dirty(uint8_t lpk)
{
	uint8_t p, t, maxpeek;	// we limit the maximum of "lookaround"
	
	p = lpk >> 3;
	maxpeek = LCD_CHARS / 8;
	if( p >= maxpeek) p = 0;
	
	while ((t=lcd.dirty[p]) == 0){
		maxpeek--;
		if(maxpeek == 0){
			lcd.pos_to_peek = p<<3;
			return;
		}
		p++;
		if(p >= LCD_CHARS / 8) p = 0;
	}
	
	lpk = p << 3;
#ifdef PHASESIGNAL
	phase_check_ext();
#endif

	// found something, now check the highest set bit
	if((t & 0xf0) == 0) {	// somewhere in the lower half (means index 4..7)
		lpk += 4;
		t <<= 4;
	}
	// this is compiled amazingly efficiently, but it could use a 16 byte flash lookup table as well:
	if((t & 0xc0) == 0) {
		lpk += 2;
		t <<= 2;
	}
	if((t & 0x80) == 0) {
		lpk ++;
		// t <<= 1;
	}
	
	lcd.pos_to_peek = lpk;
	// Note, that lcd.went_to is not touched. If it accidentally parks
	// at the position we found, it will be used.
}	// skip_some_cleans()
		
__inline__ void toggle_E(void)
{
  _delay_loop_1(config.lcd_delay >> 4);
  LCD_CTRL_EN |= ENABLE;
  _delay_loop_1(config.lcd_delay & 0xf);  // enable pulsewidth must be >230 ns
  LCD_CTRL_EN &= ~ENABLE;
#ifdef PHASESIGNAL
	phase_check_ext();
#endif
}

// Used to set RS pin to backlight control after it was used for RS :-)
static inline void lcd_rs_backlight(void)
{
#ifdef GENBOARDv3
	// s259 out1 our out2 could be used for backlight PWM
#else
	cli();	// assuming turnedoff is nicer behaviour
	LCD_CTRL_RS_RW &= ~RS;
	sei();
	setleds();
	if(softpwm_act(&lcd_backlight)){
		cli();
		LCD_CTRL_RS_RW |= RS;
		sei();
	}
#endif
}

// Enable is automatically written atomic
// RS/RW *must* be written with cli/sei
void lcd_data(uint8_t cmd)
//static __inline__ void lcd_data(uint8_t cmd)
{
  uint8_t t;

#ifdef GENBOARDv3
	/* The LBUS is used from interrupt. Therefore the usage must 
	follow strict rules:
	- Using cli() before accessing it.
	- When interrupts are enabled, the AVR LBUS pins must be output.
	- The LCD shouldn't be allowed to drive the LBUS (DDIR=0)
	- all the CS signals must be inactive (high)
	
	That's all for LBUS, I guess.
	
	LCD_RS signal:
	We keep the RS signal generally low except in this
	lcd_data function. This policy would not be necessary, but this 
	way other functions need not care setting it low. Most 
	efficient solution would be to have a mirror of the RS bit inside SRAM
	- which cannot be read back, since it's the output of 259 -
	and set RS if necessary. That would prevent unnecessary 
	toggling of RS when lcd_data(); lcd_data() ... is used without
	lcd_cmd(); in between (the autoincrement allows this)
	*/
	cli();
	// RS = high
	l259(DHIGH, LCD_RS, (uint8_t *)&LCD_RS_259_PORT, LCD_RS_259);
	sei_nop_cli();
	t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
	LCD_DATA = t | (cmd & 0xf0);
	toggle_E();	/* latch high bits */
	sei_nop_cli();
	t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
	LCD_DATA = t | (cmd << 4); /* TODO: if interrupt changes PORT value? */
	toggle_E();	/* latch low bits */
	sei_nop_cli();
	// RS = low, so the cmd..() functions need not be concerned about RS
	l259(0, LCD_RS, (uint8_t *)&LCD_RS_259_PORT, LCD_RS_259);
	sei();
	
#else
	// LCD_EN is kept low anyway, this shouldn't be needed
  LCD_CTRL_EN &= ~ENABLE;    /* 0_0  NOTE that for data RS=1 */
  cli();
  t = LCD_CTRL_RS_RW;
  t &= ~DDIR;    /* 0_0  NOTE that for data RS=1 */
  t |= RS; /* we write data, not command */
  LCD_CTRL_RS_RW = t;
  sei();

  t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
  LCD_DATA = t | (cmd & 0xf0);
  toggle_E();	/* latch high bits */
  LCD_DATA = t | (cmd << 4); /* TODO: if interrupt changes PORT value? */
  toggle_E();	/* latch low bits */

	// RS is high now
	if(softpwm_act(&lcd_backlight) == 0){
		cli();
		LCD_CTRL_RS_RW &= ~RS;
		sei();
	}
#endif	// GENBOARD v2.2
	setleds();
}

// used after lcd_data() when cursorpos-autoincrease is turned on. 
static __inline__ void lcd_step(void)
{
	uint8_t t;

/* we always force goto if config bit is set to 0 */
	if((config.lcd_c0 & _BV(lcd_easygoto)) == 0){
		lcd.went_to = 255;
		return;
	}
	
	t = ++lcd.went_to; 
/* This fails at the end of line, 
since LCD does NOT proceed to next line
For 16 Col display we do it faster!
*/
#if LCD_COLS == 16
	if ((t & (LCD_COLS - 1)) == 0)
#elif LCD_COLS == 20
	if(t==LCD_COLS || t==LCD_COLS*2 || t==LCD_COLS*3)
#else
#warning "Unknown LCD_COLS, the LCD won't work"
#endif
		lcd.went_to = 255;
}

/* lcd_refresh should be called frequently, eg. 32..100 times/sec 
 * LCD is completely updated after  appr. LCD_CHARS invocations
 */
void lcd_refresh(void)
{
	uint8_t t;
// polling the busy flag can be disabled (DDIR can be connected to GND */
	if(config.lcd_c0 & _BV(lcd_busypolled)){
		if(lcd_busy()){ // oh, LCD busy
			return;
		}
	}
	
	t = lcd.pos_to_peek;
	if(isdirty(t)){ /* it was dirty, we have to send char */
		if(lcd.went_to == t){
			TaC(t);
			t = lcd.cache[t];
			lcd_pos_to_peek_inc();
			lcd_step();
			lcd_data(t);	// reordered to be last (=>common lcd_rs_backlight() possible)
		} else { /* cursor was not there */
			lcd_go2index(t);
			/* we will lcd_putch at next invocation */
		}
	} else { /* it was clear */
		// any value would do, but constant might cause starvation of some positions
		find_dirty(t+17);
		lcd_rs_backlight();
	};
}

/* writes to lcd_cache[ index ] and steps to next position */
__inline__ void update_char(char c, lcdpos_t index)
{
	lcd.next_pos = index + 1;
	if (lcd.next_pos >= LCD_CHARS)
	  lcd.next_pos=0;

	if(lcd.cache[index] == c) return; /* nothing to do */
	lcd.cache[index] = c;
	dirtyat(index);
}

/* this gets a string */
void update_chars(char *c, lcdpos_t index)
{
	lcdpos_t i, z;
	z=0;
	for(i=index; i<LCD_CHARS && c[z]; i++){
		update_char(c[z++], i);
	}
}

/* puts a sign character to index, and returns unsigned value */
uint16_t stripsign(int bin, lcdpos_t index)
{
	if(bin<0){
		update_char('-', index);
		return 0-bin;
	}
	update_char('+', index);	/* or ' ' */
	return bin;
}

/* FAST conversion!
 * result is exactly 5 LCD chars updated starting from index ...
 * no sign, see stripsign() if you need it
 */
uint16_t BCDbits[] = {80,800,8000};
void int2BCD5(uint16_t bin, lcdpos_t index)
{
  int8_t i;
  uint8_t c, tenpower8;
  uint16_t result, t;

  tenpower8 = 2;

  t=40000; c=4;
  result = '0';	/* 1 digit of the number */
  for (i=14; i>=0; i--) {  // we don't start at 15, and we also want the last (0) element
    if (bin>=t) {  // greater than _and equal_
      bin -= t;
      result += c;
    }
    t >>= 1; c >>= 1;
    if (!(i&3)) {
      update_char(result, index++);
      t = BCDbits[tenpower8--];
      c=8;
      result = '0';
    }
  }
  update_char('0' + bin, index);	/* last digit is easy */
}

/* tenpower8= 2(4 digits), 1(3 digits), 0(2 digit) see named constants in lcd.h */
void int2BCDn(uint16_t bin, lcdpos_t index, char tenpower8)
{
  uint8_t c;
  uint16_t result, t;

  while (tenpower8>=0) {  
      t = BCDbits[(uint8_t)tenpower8--];
      c=8; result = '0';	/* 1 digit of the number */

    if (bin>=t) {  // greater than _and equal_ 8..
      bin -= t; result += c;
    }
    t >>= 1; c >>= 1;

    if (bin>=t) {  // greater than _and equal_ 4..
      bin -= t; result += c;
    }
    t >>= 1; c >>= 1;

    if (bin>=t) {  // greater than _and equal_ 2..
      bin -= t; result += c;
    }
    t >>= 1; c >>= 1;

    if (bin>=t) {  // greater than _and equal_ 1..
      bin -= t; result += c;
    }
#ifdef PHASESIGNAL
	phase_check_ext();
#endif
      update_char(result, index++);
  } // while

  if(bin > 9){
	update_char(OUT_OF_RANGE_CHAR, index);	/* out of range error */
  } else {
  	update_char('0' + bin, index);	/* last digit is easy */
  }
}


/* write command 7:4 bits assuming 8 bit interface */
// Enable is automatically written atomic
// RS/RW *must* be written with cli/sei
void lcd_cmd8(uint8_t cmd)
{
  uint8_t t;
    
  LCD_CTRL_EN &= ~ENABLE;	/* 000 */

	cli();
#ifdef GENBOARDv3
	// ENABLE, RS and DDIR kept low
	t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
	LCD_DATA = t | (cmd & 0xf0);
	toggle_E();	/* latch high bits */
	sei();
#else	// GENBOARD v2.2
  LCD_CTRL_RS_RW &= ~( RS | DDIR );	/* 000 */
  sei();
  t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
  LCD_DATA = t | (cmd & 0xf0);
  toggle_E();	/* latch high bits */
#endif	// GENBOARD v2.2
}

/* write command assuming 4 bit interface
TODO: should allow compile-time config 8 or 4 bits */
// Enable is automatically written atomic
// RS/RW *must* be written with cli/sei
__inline__ void lcd_cmd(uint8_t cmd)
{
  uint8_t t;

#ifdef GENBOARDv3
	// ENABLE, RS and DDIR kept low
	cli();
	t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
	LCD_DATA = t | (cmd & 0xf0);
	toggle_E();	/* latch high bits */
	sei_nop_cli();
	t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
	LCD_DATA = t | (cmd << 4);
	toggle_E();	/* latch low bits */
	sei();
#else	// GENBOARD v2.2
  LCD_CTRL_EN &= ~ENABLE;	/* 000 */
  cli();
  LCD_CTRL_RS_RW &= ~( RS | DDIR );	/* 000 */
  sei();
  t=LCD_DATA & 0x0f;	/* clear 7:4 bits */
  LCD_DATA = t | (cmd & 0xf0);
  toggle_E();	/* latch high bits */
  // MIK TODO: read t again, and cli/sei it. An enable cycle must min be 500 ns anyway
  LCD_DATA = t | (cmd << 4); /* TODO: if interrupt changes PORT value? */
  toggle_E();	/* latch low bits */
  
  	// RS is low now
	if(softpwm_act(&lcd_backlight)){
		cli();
		LCD_CTRL_RS_RW |= RS;
		sei();
	}
#endif	// GENBOARD v2.2
	setleds();
}

/*-----------------27.5.99 20:33--------------------
 *  Clear display
 *--------------------------------------------------*/
__inline__ void lcd_cls()
{
  lcd_cmd(0x01);
  // TODO: this should clear the cache too
}

/*-----------------27.5.99 20:33--------------------
 *  Home position
 *--------------------------------------------------*/
void lcd_home()
{
  lcd.pos_to_peek = 0;
  lcd_cmd(0x02);
  lcd.next_pos = 0;
}


/*-----------------27.5.99 20:33--------------------
 *  LCD & Cursor control (Blinking, ON/OFF,...)
 *--------------------------------------------------*/
void lcd_control(byte disonoff, byte curonoff, byte curblink)
{
  byte temp;
  temp=0x08;
  if (disonoff==1) temp|=0x04;
  if (curonoff==1) temp|=0x02;
  if (curblink==1) temp|=0x01;
  lcd_cmd(temp);
}

/*-----------------27.5.99 20:31--------------------
 *  Goto position   (0x00 - 1.row, 1.col)
 *                  (0x40 - 2.row, 1.col)
 *--------------------------------------------------*/
void lcd_goto(byte mesto)
{
  lcd_cmd(0x80 | (mesto & 0x7f));
}

/* wrapper around lcd_goto indexed as lcd_cache[] */
void lcd_go2index(lcdpos_t index)
{
  byte t, l;

#if LCD_COLS == 16
/* line beginning addresses: 0,0x40,0x10,0x50 */

#define L1ADDR	0x30
#define L2ADDR	0xf0
#define L3ADDR	0x20
#else 

  // the designer of the hd44780 4x20 char display propably had one of those monday mornings...
  // arghhhh, it is convoluted!

  // the 4x20 display works like a 2x40 display cut in half. Address 0 is the start 
  // of line 1. At the end of line 1, the cursor continues to line 3! The same is the
  // case for line 2 / line 4.
/* line beginning addresses: 0,64,20,84 */
#define L1ADDR	0x2c
#define L2ADDR	0xec
#define L3ADDR	0x18
#endif

  l = config.lcd_offs[0];

  if (l == 0xff){	// default line addresses used, that Mik's 4x20 LCD follows
  if (index >= (3*LCD_COLS)) {
    t = index + L3ADDR;
  } else if (index >= (2*LCD_COLS)) {
    t = index + L2ADDR;
  } else if (index >= (LCD_COLS)) {
    t = index + L1ADDR;
  } else {
    t = index;
  }
  } else { // line_address - cache_address offsets are stored in config
    if (index >= (3*LCD_COLS)) {
    t = index + config.lcd_offs[3]; // magic constants, look in the datasheet for an explanation
  } else if (index >= (2*LCD_COLS)) {
    t = index + config.lcd_offs[2];
  } else if (index >= (LCD_COLS)) {
    t = index + config.lcd_offs[1];
  } else {
    t = index + config.lcd_offs[0];
  }
  }
  
  // these address calculations makes it a lot easier to use the lcd,
  // now line 0 has address 0..19, line 1 has address 20..39 etc.

//#endif

  lcd_goto(t);
  lcd.went_to = index;
}


__inline__ void lcd_setpos(lcdpos_t index)
{
  lcd.next_pos = index;
}

__inline__ lcdpos_t lcd_getpos(void)
{
  return lcd.next_pos;
}

/*-----------------27.5.99 20:31--------------------
 *  Put character on LCD (in cache)
 *--------------------------------------------------*/
__inline__ void lcd_putch(byte data)
{ 
  update_char(data, lcd.next_pos);
}

/*-----------------27.5.99 20:30--------------------
 *  Display null terminated string
 *  You always want to use lcd_putstr_P() for constant strings to
 *  waste less SRAM. (also avrstudio does not initialize section data properly?)
 *--------------------------------------------------*/
void lcd_putstr(byte *data)
{
	uint8_t t;
  for( ;(t = *data); data++) {
    lcd_putch(t);
  }
}

// fetch string from flash memory and print to lcd
void lcd_putstr_P(const char *data)
{
  char c;
  // it might be necessary to use __elpm_inline()
  while( (c = PRG_RDB(data++)) ) {
    lcd_putch(c);
  }
}

// fetch string from flash memory and print to lcd (max 'max' characters)
void lcd_putstr_P_max(const char *data, uint8_t max)
{
  char c;
  while( (c = PRG_RDB(data++)) && max--) {
    lcd_putch(c);
  }
}


void lcd_clr_to_end(void)
{
	while(lcd.next_pos && lcd.next_pos < LCD_CHARS){
		update_char(' ', lcd.next_pos);
	}
}

void lcd_clr_to_idx(uint8_t idx)
{
	while(lcd.next_pos && (lcd.next_pos <= idx) && (lcd.next_pos < LCD_CHARS)) {
		update_char(' ', lcd.next_pos);
	}
}

void lcd_clr_rest_of_line(void) {
  uint8_t i,k;

  k = lcd.next_pos;

  if (k >= 3*LCD_COLS)
    k = 4*LCD_COLS;
  else if (k >= 2*LCD_COLS)
    k = 3*LCD_COLS;
  else if (k >= LCD_COLS)
    k = 2*LCD_COLS;
  else
    k = 1*LCD_COLS;

  for (i=lcd.next_pos; i<k; i++)
    update_char(' ', lcd.next_pos);
}


/*-----------------27.5.99 20:29--------------------
 *  Display 8bit bin value
 *--------------------------------------------------*/
void printbin(byte x)
{
  byte i;
  for (i=128;i>0;i>>=1)
  {
    if ((x&i)==0) lcd_putch('0');
    else lcd_putch('1');
  }
}

/*-----------------1.6.99 16:39---------------------
 *  Display byte in hex
 *--------------------------------------------------*/
void printhex(byte i)
{
  lcd_putch(hex_hi(i));
  lcd_putch(hex_lo(i));
}

/*-----------------28.08.99 23:00-------------------
 *   Display 0-99
 * --------------------------------------------------*/
void print2(uint8_t x)
{
  uint8_t y;
   if (x < 100)
     {
       y = x/10;
       lcd_putch(y+0x30);
       x -= y*10;
       lcd_putch(x+0x30);
     }
   else 
     lcd_putstr_P(PSTR("??"));  // better to write a two character string
   //     lcd_putstr("Err");
}

/*-----------------27.5.99 20:29--------------------
 *  Display 0-999 unsigned char or int (8 or 16bit)
 *--------------------------------------------------*/
void print3(uint16_t x)
{
  uint16_t y;
  if (x<1000)
  {
    y=x/100;
    lcd_putch(y+0x30);

    x-=(y*100);
    y=x/10;
    lcd_putch(y+0x30);

    x-=(y*10);
    lcd_putch(x+0x30);
  }
  else lcd_putstr_P(PSTR("Err"));
}

/*-----------------27.5.99 20:30--------------------
 *  Display integer 0-65535 (16bit)
 *--------------------------------------------------*/
/*void print5(unsigned int x)
{
  unsigned int y;
    y=x/10000;lcd_putch(y+0x30);x-=(y*10000);
    y=x/1000;lcd_putch(y+0x30);x-=(y*1000);
    y=x/100;lcd_putch(y+0x30);x-=(y*100);
    y=x/10;lcd_putch(y+0x30);x-=(y*10);
    lcd_putch(x+0x30);
}*/

/*-----------------31.5.99 18:18--------------------
 *  Display unsigned long 0-4294967296 (32bit)
 *--------------------------------------------------*/
/*void print10(unsigned long x)
{
  unsigned int y;
    y=x/1000000000;lcd_putch(y+0x30);x-=(y*1000000000);
    y=x/100000000;lcd_putch(y+0x30);x-=(y*100000000);
    y=x/10000000;lcd_putch(y+0x30);x-=(y*10000000);
    y=x/1000000;lcd_putch(y+0x30);x-=(y*1000000);
    y=x/100000;lcd_putch(y+0x30);x-=(y*100000);
    y=x/10000;lcd_putch(y+0x30);x-=(y*10000);
    y=x/1000;lcd_putch(y+0x30);x-=(y*1000);
    y=x/100;lcd_putch(y+0x30);x-=(y*100);
    y=x/10;lcd_putch(y+0x30);x-=(y*10);
    lcd_putch(x+0x30);
}*/


/* ioinit() does this. Should not be needed */
__inline__ void lcd_ctrl_as_output(void)
{
	LCD_CTRLDDR_EN |= ENABLE;     /* outputs */
	LCD_CTRLDDR_RS_RW |= (
#ifndef GENBOARDv3	// actually, EN and DDIR are on the same port on v3...
	RS |
#endif
	DDIR);     /* outputs */
}

/* we drive the control lines LOW */
/*void  lcd_testLOW(void)
{
	lcd_ctrl_as_output();
	LCD_CTRL &= ~( ENABLE | RS | DDIR );
}*/

/* we drive the control lines HIGH */
/*void  lcd_testHIGH(void)
{
	lcd_ctrl_as_output();
	LCD_CTRL |= ( ENABLE | RS | DDIR );
}*/
/*void lcd_testrun(void)
{
	uint8_t t;
	for (t=0; t<64; t++){
		lcd_goto(t<<1);
		printhex(t);
	}
}*/

/* this dirties the whole LCD, and makes sure it will be refreshed: */
__inline__ void lcd_dirty(void)
{
	uint8_t i;
	
	for (i=0; i<(LCD_CHARS/8); i++)
		lcd.dirty[i] = 0xFF;  // set all dirty bits
	
	lcd.went_to = 255;	// not necessary
}

/*  Display initialization:
	1) Setting outputs
	2) Clean the cache
	3) lcd.init = startcountfrom
	
	After this,
	lcd.init variable will start counting up in lcd_init_seq()
	appr. every 10msec, starting from startcountfrom.
	Real initialization will begin at (LCD_INIT_STATE + 1)
*/
void lcd_initfrom(uint8_t startcountfrom)
{
  uint8_t i;
  lcd.init = startcountfrom ;
  // non-0 lcd.init also disables lcd_refresh() until the lcd has been initialized
  lcd.pos_to_peek=0;
  lcd.next_pos=0;
  // if software control of the brightness is enabled then the backlight value should
  // be it's brightness
  if ((config.lcd_c0 & _BV(swbrightness))==0)
    softpwm_init(&lcd_backlight, config.lcd_backlight);

  for (i=0; i<LCD_CHARS; i++)
    lcd.cache[i] = ' '; // this will clear the cache

	lcd_dirty();

  i = SREG;
  cli(); // not needed at init; but mli could theoretically screw a flybackpulse otherwise
  lcd_ctrl_as_output();
  LCD_CTRL_EN &= ~ENABLE;
  LCD_CTRL_RS_RW &= ~(
#ifndef GENBOARDv3
	RS |
#endif
	DDIR );	
  LCD_DATADDR |= 0xf0;
  SREG = i;

// lcd_ready is cleared, the mainloop and 
// lcd_charseq_test() will take care of the rest
}

// compatibility wrapper with default waiting-time
void lcd_init()
{
	lcd_initfrom(LCD_INIT_STATE - 25);
}

void lcd_start_address_test(void)
{
	lcd.went_to = 255; 
	lcd.init = LCD_ADDRESS_TEST;
}

/* 
prints a character sequence so you can find your favorite (Hiraghana :-) characters
*/
void lcd_charseq_test(uint8_t offs)
{
	uint8_t i;
	// update_char() could be used (bit'ofa waste)
	i = 0;
	do{
		lcd.cache[i] = offs + i;
		i++;
	} while(i < LCD_CHARS);
	lcd_dirty();
}

// lcd initialization without busy wait
// it is based on a 10msec handler, so the complete initialization takes some time, who cares?!
void lcd_init_seq(void) {
  uint8_t idx, h;
  idx = lcd.init++;
  switch(idx) {
  case LCD_INIT_STATE+1: lcd_cmd8(0x30); return;
  case LCD_INIT_STATE+2: lcd_cmd8(0x30); return; // now switched to 8 bit for sure
  case LCD_INIT_STATE+3: lcd_cmd8(0x30); return;
  case LCD_INIT_STATE+4: lcd_cmd8(0x20); return; // now switched to 4 bit
  case LCD_INIT_STATE+5: lcd_cmd(0x28); return;	// 2 lines, 5x7 dots
  case LCD_INIT_STATE+6: lcd_cmd(0x0c); return;	// display on, cursor off, cursor blink off = lcd_control(1,0,0)
  case LCD_INIT_STATE+7: lcd_cmd(0x06); return; //  increment position, no display shift
  case LCD_INIT_STATE+8: 
    lcd_cls();
    lcd.went_to = 255;  // just let lcd_refresh take care of the cursor position
    lcd_setpos(0);
    if ((config.lcd_c0 & _BV(swbrightness))==0){
	softpwm_init(&lcd_backlight, config.lcd_backlight);
    }
    lcd.init = 0;  // init sequence finished
    return;
  case LCD_ADDRESS_TEST:	// sending an address-test sequence 00..7f
	++lcd.went_to;
	h = lcd.went_to >> 1;
	if(h < 0x7f) lcd.init--;	// we stay in this state
	if(lcd.went_to & 1){	// sending data
		lcd_data(h&1 ? hex_lo(h) : hex_hi(h));
	} else {	// going to position
		lcd_goto(h);
	}
	return;
  default: 
	if(idx > LCD_ADDRESS_TEST){
		if(++lcd.next_pos) lcd.init--;	// delay for 4*255*10 msec
		else {	// doing it appr. 4 times? No problem.
			lcd_dirty();
		}
	}
	return;
  }
}

// Enable is automatically written atomic
// RS/RW *must* be written with cli/sei
uint8_t lcd_busy(void) // read back busy flag (0x80)
{
  uint8_t t;
  // LCD_DATADDR &= 0x0F; // LBUS (pins 7:4) input
  /* By inventing LCD_DATADDR30 constant for the DDR of 3:0 pins, that
  way we can avoid a readback and an AND/OR operation
  => becomes simple LCD_DATADDR = ... at compiletime
  */
	cli();
	LCD_DATADDR = LCD_DATADDR30;	/* 7:4 are let 0, thus input */
	LCD_CTRL_EN &= ~ENABLE;		// TODO: remove after v3.0 busypoll debugged
#ifdef GENBOARDv3
	LCD_CTRL_RS_RW |= DDIR; /* we read back. From now, the LCD drives LBUS.  */

	// first delay for a short while letting the control signals stabilize
	_delay_loop_1(config.lcd_delay >> 4);

	// set the enable
	LCD_CTRL_EN |= ENABLE;

	// wait for some time, be sure that the data has arrived to the avr
	_delay_loop_1(config.lcd_delay & 0xf);  // enable pulsewidth must be >230 ns

	// read the 7:4 bits
	t = LCD_DATAPIN ;
	// clear enable
	LCD_CTRL_EN &= ~ENABLE;

	// at this point pending interrupts could let sneak through:
	// DDIR low; DATADDR output;sei();nop();cli(); DATADDR input; DDIR high

	toggle_E();  // 4 bit mode; we need to toggle EN two times, else the lcd gets confused

	LCD_CTRL_RS_RW &= ~DDIR;
  	
#else

  t = LCD_CTRL_RS_RW;
  t &= ~RS;
  t |= DDIR;  /* we read back */
  LCD_CTRL_RS_RW = t;
  sei();

  // first delay for a short while letting the control signals stabilize
  _delay_loop_1(config.lcd_delay >> 4);

  // set the enable
  LCD_CTRL_EN |= ENABLE;

  // wait for some time, be sure that the data has arrived to the avr
  _delay_loop_1(config.lcd_delay & 0xf);  // enable pulsewidth must be >230 ns

  // read the busy-flag
  t = LCD_DATAPIN ;  /* we just return the BUSY flag, not all 4 bits with 0xF0 */

  // clear enable
  LCD_CTRL_EN &= ~ENABLE;

  toggle_E();  // 4 bit mode; we need to toggle EN two times, else the lcd gets confused
  	
	// setleds(); not needed here, since PORTA was not screwed
	
	if(softpwm_act(&lcd_backlight)){
		uint8_t t2;
		cli();
		t2 = LCD_CTRL_RS_RW;
		t2 &= ~DDIR;
		t2 |= RS;
		LCD_CTRL_RS_RW = t2;
	} else {
		cli();
		LCD_CTRL_RS_RW &= ~( RS | DDIR );
	}
#endif
	/* the GENBOARDv3 interrupts are already impatient, but they expect
	LBUS to be output, so this need to be done before sei():
	This must be _after_ DDIR=0, we don't want to drive outputs against each other.
	Moving before softpwm_act() could decrease flickering of good-old v2 LEDs
	*/
	LCD_DATADDR = LCD_DATADDR30 | 0xF0;  /* data out */
	
	// show must go on:
	sei();

	// we just return the BUSY flag	neglect other bits we read from PINA/PINC 
	t &= 0x80;
	return t;
}

/*-----------------28.08.99 22:49-------------------
 *   Simple printf function (no fp, and strings)
 *--------------------------------------------------*/
/*
int printf(const char *format, ...){
  char format_flag;
  unsigned int u_val, div_val, base;
//   char *ptr;
  va_list ap;

  va_start (ap, format);
  for (;;) {
     while ((format_flag = *format++) != '%') {      // Until '%' or '\0'
       if (!format_flag) {
         va_end (ap);
         return (0);
       }
       lcd_putch (format_flag);
     }

     switch (format_flag = *format++) {
       case 'c': format_flag = va_arg(ap,int);
       default:  lcd_putch(format_flag); continue;

       case 'd': base = 10; div_val = 10000; goto CONVERSION_LOOP;
       case 'x': base = 16; div_val = 0x10;

       CONVERSION_LOOP:
       u_val = va_arg(ap,int);
       if (format_flag == 'd') {
         if (((int)u_val) < 0) {
           u_val = - u_val;
           lcd_putch ('-');
         }
         while (div_val > 1 && div_val > u_val) div_val /= 10;
       }
       do {
         lcd_putch (hex[u_val / div_val]);
         u_val %= div_val;
         div_val /= base;
       } while (div_val);
    }
  }
}
*/

/* 
* similar as int2BCNn() but bin is a signed number 
* Note the lack of lcdpos_t, we use lcd.next_pos 
*/
void int_i2BCDn(int16_t bin, char tenpower8)
{
	if(bin < 0){
		lcd_putch('-');
		int2BCDn(((uint16_t)0) - bin, lcd.next_pos, tenpower8);
	} else {
		lcd_putch('+');
		int2BCDn(bin, lcd.next_pos, tenpower8);
	}
}

#endif

/*@}*/

