/*********************************************************\
*  LCD functions                                          *
*  (C) 2003 [Marcell Gal, cell at x-dsl.hu                *
***********************************************************
*  -- History --                                          *
*    release 0.0 - 20030601
*            splitout from vems.c
*    release 0.1 - 20050409
*	     maintenance changed from page to line
\*********************************************************/
   /*!
   * \addtogroup g_lcd LCD Control Routines
   */
/*@{*/

#ifdef  LCD
#include "vems.h"
#include "global.h"
#include "comm.h"
#include "fuelcalc.h"
#include "helpers.h"
#include "storage.h"
#include "adc.h"
#include "actuators.h"
#include "tables.h"
#include "ve.h"
#include "iac.h"
#include "menu.h"
#include "eventqueue.h"
#include "benchmark.h"
#include "ego.h"
#include "pid.h"
#include "spi.h"
#include "knock.h"
#include "boostcontrol.h"
// #include "multitooth.h"
#include "firmware_revision.h"

#include "lcd.h"
#include "lcd_display.h"

// DEL #include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
// DEL #include <avr/signal.h>
// DEL #include <avr/sfr_defs.h>
// DEL #include <avr/pgmspace.h>

#include "arithm.c"

#ifdef WBO2
#include "wbo2.h"
#endif

#include "magicstr.h"

extern volatile struct engine_t engine;

uint8_t ego_warmup_count;

struct lcd_display_t lcd_disp;

/*
 * val16 = 4096 * 3.72
 * is printed like "3.72"
 * So we have 12 bits of fraction.
 */
static inline void lcd_print_X_XX( uint16_t val16 )
{
	uint8_t val, frac;
	val = div4096( val16 );
	if( val > 9 ){
		val = 9;
		frac = 99;
	} else {
		frac = div4096( mult16_8( val16 & 0xFFF, 100));
	}
	lcd_bcd1( val ); lcd_putch('.'); lcd_bcd2( frac );
}

/*
 * val16 = 256 * 14.7 
 * is printed like "14.7"
 * So decimal dot is on the byte boundary.
 */
static inline void lcd_print_XX_X( uint16_t val16 )
{
	uint8_t val8, frac;

	val8 = val16 >> 8;

	if( val8 > 99 ){
		val8 = 99;
		frac = 9;
	} else {
		frac = ((val16 & 0xff) * 10) >> 8;
	}

	lcd_bcd2( val8 ); lcd_putch('.'); lcd_bcd1( frac );
}
/* writes an intro string up on the display and centers it */
void lcd_introstr(void)
{
	uint8_t i;
	// centre the string
	for (i=(LCD_COLS-sizeof(FIRMWARE_VENDOR))/2;i>0;i--)
		lcd_putch(' '); 
    
	lcd_putstr_P(PSTR(FIRMWARE_VENDOR));
}

/* writes the firmware version string on the display */
void lcd_firmware_versionstr(void)
{
        uint8_t i;
        // centre the string
        for (i=(LCD_COLS-sizeof(FIRMWARE_VERSION)-10)/2;i>0;i--)
                lcd_putch(' ');
	lcd_putstr_P(PSTR("Firmware: "));
        lcd_putstr_P( PSTR( FIRMWARE_VERSION ) );
}

/* displays a message alerting the user that the ECU is dropping
 * into bootloader mode for a firmware update */
void lcd_bootloaderstr(void)
{
#if LCD_COLS==16
	lcd_putstr_P(PSTR("Firmware Update"));
#else
	lcd_putstr_P(PSTR(" Updating Firmware "));
#endif
}

/* 
It is a good idea to have small functions in lcd_funcarr, 
that print only 1..5 characters,
because it will result in smaller overhead on mainloop.
Also the chance for reuse is higher for smaller functions.
Do not worry about slight increase in flashsize
*/

void lcd_rpmstr(void)
{
#if LCD_COLS==16
    lcd_putstr_P(PSTR("RPM"));
#else 
	lcd_putstr_P(PSTR("RPM:"));
#endif
}

void  lcd_rpm_true_4char(void)
{  // true 4 digit RPM in range 0-9999. Over that, approximation will be used (in 2 digit kRPMs)
  uint16_t rpm=engine.rpm_vhr; // Very high resolution RPM
  if (rpm < 10000)
    int2BCDn(rpm, lcd.next_pos, BCD_DIGIT4); // Print true RPM in 4 digits
  else
  { // with rpmstr the display should show "RPM>12k"-like string
    lcd_putch('>');
	// div1024 gives more info (RPM>09k means 10000..10239 RPM)
    int2BCDn(div1024(rpm), lcd.next_pos, BCD_DIGIT2);
    lcd_putch('k');
  }
}    


void lcd_space(void)
{
	lcd_putch(' ');
}

void lcd_2space(void)
{
      lcd_putstr_P(PSTR("  "));
}

void lcd_4space(void)
{
	lcd_2space();
	lcd_2space();
}

void lcd_5space(void)
{
      lcd_putstr_P(PSTR("     "));
}


void lcd_Tstr(void)
{
#if LCD_COLS==16
    lcd_putstr_P(PSTR("AT"));
#else
    lcd_putstr_P(PSTR("T"));
#endif
}

void print_temperature(uint8_t t)
{ // Print t-40 degrees Fahrenheit (or Celsius) - 3 characters
#ifdef CELSIUS
  uint8_t c;

  // print temperature in the Celsius scale
  // limit the display to two digits + sign

  // c = 5/9 * (f-32)
  // c = 5/9 * (t-40-32)
  // c = 5/9 * t - 5/9 * (40+32)
  // c = 5/9 * t - 40

  c = (t*142) >> 8; // * 5/9
  if (c >= 40) { // positive temperature (2 or 3 characters)
    c -= 40;
    if(c > 99) // Some hack to display temperatures over +100C (possible with water, in a closed cooling system, over 100 kPa)
               // TODO: currently using a -40 shifting in the F scale, therefore the maximum displayable temp is 215 F=101 C.
               //   We should extend the maximum to ca. 117 C. Suggested F-scale shifting is -13.
               //   (then the minimum: -25 C instead of current -40 C)
    {
      int2BCDn(c, lcd.next_pos, BCD_DIGIT3);
      return;
    }
    lcd_putch(' '); // this space used for sign ('+')
  } else { // negative temperature (should be 2 characters)
    lcd_putch('-');
    c = 40 - c;
  }

  int2BCDn(c, lcd.next_pos, BCD_DIGIT2);
#else
    // print temperature in the Fahrenheit scale
    int2BCDn(t-40, lcd.next_pos, BCD_DIGIT3);
#endif
}

void  lcd_T(void)
{ // Manifold air temperature - 3 characters
  print_temperature(engine.mat);
  // No unit symbol used
}


/* prints the battery voltage
 * display is like 12.6V padded to the right
 */
void lcd_battery_volts(void) {
	/* engine.batt is always scaled as a fraction of 30V */
	lcd_print_XX_X( engine.batt * 30 );
	lcd_putch( 'V' );
}

void  lcd_Pstr(void)
{
    lcd_putstr_P(PSTR("P")); // means MAP for LCD_4x20, Power for LCD_2x16
}

void lcd_P(void)
{
// Q1: how is kpafac (easily) configured when it cannot be read on the lcd?
// Q2: who cares about the non-corrected MAP value?
    int2BCDn(engine.kpa250 >> 2, lcd.next_pos, BCD_DIGIT3);
} 

// goto lineidx
void lcd_GLx( uint8_t lineidx )
{
	lcd_setpos( LCD_COLS * lineidx );
}

void lcd_space_or_key(void)
{
   	char sp;

#ifdef KEYBOARD
/* TODO: check if not called too often */
	if((sp = cooked_char())){
		if (sp == '9'){
			lcd_init();
			lcd_waitmode(1);
		}
	} else {
		sp=' ';
	}
#else
	sp=' ';
#endif
	update_char(sp, 28);
}

void lcd_print_o2percent_old(void)
{
	lcd_putstr_P(PSTR("O2%="));
}

#ifdef WBO2
void lcd_print_current_o2(void)
{
	uint16_t o2 = wbo2.current_o2;

	if (o2 != 0) {
		uint32_t t32 = mult16_8( o2 >> 1 , 51 );
		// O2% * 10 (*256)
		// so we need a *0.1 here: that is * 51 / 512
		lcd_print_XX_X( t32 >> 8 );
	} else {
		// undefined, maybe sensor not heated
                lcd_putstr_P( PSTR("??.?") );
	}
}
#endif	// WBO2

// ==================================
void lcd_O2str(void)	// 3 (WBO2) or 5(NBO2) char
{
#ifdef WBO2
	if (WBO2_SENSOR_CONFIGURED)
		// this could be made dynamic "La:" / "Af:"
		// preferrably printed same time as the value see lcd_O2()
		lcd_putstr_P(PSTR("La:"));
	else
#endif	// WBO2
		lcd_putstr_P(PSTR("O2:0."));
}

#ifdef WBO2
/*
 * stoich afr= 1024 * 14.7
 * stoich lambda = 4096 = afr * 70 / 256
 */
static inline void lcd_print_lambda(uint16_t afr)
{
	uint16_t lambda;
	uint32_t t32;

	t32 = mult16_8(afr, 70);
	lambda = t32 >> 8;
	lcd_print_X_XX( lambda );
}
#endif // WBO2

/*
 * 4 chars for WBO2 
 * "0.99" (lambda)
 * "14.7" (O2)
 * "??.?" (unknown)
 *
 * 3 chars for NBO2
 * "82V"
 */
void lcd_O2(void)
{
#ifdef WBO2
	if (WBO2_SENSOR_CONFIGURED){ 
		uint16_t afr = wbo2.current_afr;
		if(afr == 0xffff){
			lcd_print_current_o2();
		} else {
			lcd_print_lambda( afr );
		}
	} else
#endif
	{
		uint8_t t;
		t = ego_voltage();

		t <<= 1;
		if (t > 99) t = 99;
		lcd_bcd2( t );
		lcd_putch('V');
	}
}

// ================================

static inline void print_ego_plus_minus(uint8_t ego)
{
    uint8_t t;
    if (ego > EGOCORR_BASE){
		lcd_putch('+');
		t = ego - EGOCORR_BASE;
    }else{
		lcd_putch('-');
		t = EGOCORR_BASE - ego;
    }
    int2BCDn(t, lcd.next_pos, BCD_DIGIT2);
}

void lcd_egocorr(void)
{
    lcd_putstr_P(PSTR("egoC:"));
	print_ego_plus_minus(corr.ego);
}

void lcd_egocorr_average(void)
{
    lcd_putstr_P(PSTR("egoC:"));
	print_ego_plus_minus(ego_corr_filt/HIST_LENGTH);
}

void lcd_coolantstr(void)
{
#if LCD_COLS==16
    lcd_putch('C');
#else
    lcd_putstr_P(PSTR("clt:"));
#endif
}

void lcd_coolant(void)
{ // Coolant temperature - 4 characters
  print_temperature(engine.coolant);
#ifdef CELSIUS
  lcd_putch('C');
#else
  lcd_putch('F');
#endif
}

void lcd_tpsstr (void)
{
    lcd_putstr_P(PSTR(" tps:"));
}

void lcd_tps (void)
{	// print tps position
    int2BCDn(engine.tps, lcd.next_pos, BCD_DIGIT3);
}

void lcd_pwstr(void)
{
	 lcd_putstr_P(PSTR("pw:"));
}

void lcd_tps_percent (void)
{	// print tps position in percentage 0-99% or 'max' at full throttle (always in 3 chars)
    if (engine.tps == 255)
        lcd_putstr_P(PSTR("max"));
    else
    {
      int2BCDn((engine.tps*25)>>6, lcd.next_pos, BCD_DIGIT2);
      lcd_putch('%');
    }
}

void lcd_shiftcut_state(void)
{
	lcd_putch(' ');
	printhex( engine.shiftcut_state );
}

void lcd_idlerpm(void)
{
	lcd_putch(' ');
	int2BCDn(idle_rpm, lcd.next_pos, BCD_DIGIT3);
}

void lcd_iacpos_str(void)
{
	lcd_putstr_P(PSTR("pos: "));
}

void lcd_iacpos(void)
{
	int2BCDn(iac.position, lcd.next_pos, BCD_DIGIT3);
}

void lcd_iacdir_str(void)
{
	lcd_putstr_P(PSTR("dir: "));
}

void lcd_iacdir(void)
{
	if (iac.status & _BV(direction))
		lcd_putch('+');
	else
		lcd_putch('-');

	int2BCDn(iac.count, lcd.next_pos, BCD_DIGIT2);
}

void lcd_iacstr(void)
{
	lcd_putstr_P(PSTR("iac:"));
}

void lcd_iac_status(void)
{
	if (engine.status & _BV(idle))
		lcd_putstr_P(PSTR(" active"));
	else
		lcd_putstr_P(PSTR("       "));

}

void lcd_iac_pid_e(void)
{
	lcd_putch('e');
	int_i2BCDn(iac.error, BCD_DIGIT4);
}

void lcd_iac_pid_i(void)
{
	lcd_putch('i');
	int_i2BCDn(iac.integral, BCD_DIGIT4);
}

void lcd_iac_pid_d(void)
{
	lcd_putch('d');
	int_i2BCDn(iac.rpmdot, BCD_DIGIT4);
}

void lcd_iac_pid(void)
{
	lcd_putch('p');
	int_i2BCDn(iac.pid, BCD_DIGIT4);
}

void lcd_iac_pid_pterm(void)
{
	lcd_putch('P');
	int_i2BCDn(iac.pid_p, BCD_DIGIT4);
}

void lcd_iac_pid_iterm(void)
{
	lcd_putch('I');
	int_i2BCDn(iac.pid_i, BCD_DIGIT4);
}

void lcd_iac_pid_dterm(void)
{
	lcd_putch('D');
	int_i2BCDn(iac.pid_d, BCD_DIGIT4);
}

void lcd_iac_resync(void)
{
//    lcd_putch('S');
//#ifdef BENCHMARK
//    printhex(benchstats.iac_overclose>>8);
//    printhex(benchstats.iac_overclose);
//#endif
	printhex(0);
	printhex(0);
}

void lcd_iac_advance(void)
{
	uint8_t t;
	int16_t adjust = iac.adv_adjust;

	if (adjust < 0) {
		lcd_putch('-');
		t = -adjust;
	} else {
		lcd_putch('+');
		t = adjust;
	}
	int2BCDn(t, lcd.next_pos, BCD_DIGIT3);
}


void lcd_ve_learn_str(void)
{
	lcd_putstr_P(PSTR("VE learn:"));
}

void lcd_ve_learn_status(void)
{
	if ((config.ve_learn_conf & _BV(ve_learn_on)) &&
	    (engine.status_ext & _BV(o2_closed_loop)) && 
	    (engine.coolant >= config.ve_learn_coolant)) {
		
		lcd_putstr_P(PSTR(" active"));
		
		if (config.ve_learn_conf & _BV(ve_learn_sim))
			lcd_putstr_P(PSTR(" sim"));
		else
			lcd_clr_rest_of_line();

	} else {
		lcd_clr_rest_of_line();
	}
}

void lcd_ve_learn_stat_r(void)
{
	lcd_putch('r');
	printhex(engine.rpm);
	lcd_putch('-');
	printhex(tune_me.rpm);
	lcd_space();
}

void lcd_ve_learn_stat_m(void)
{
	lcd_putch('m');
	printhex(engine_kpa8());
	lcd_putch('-');
	printhex(tune_me.kpa);
	lcd_space();
}

void lcd_ve_learn_stat_a(void)
{
	lcd_putch('a');
	printhex(corr.ego);
	lcd_putch('-');
	printhex(tune_me.adjust);
}

void lcd_ve_learn_conf(void)
{
	lcd_putch('c');
	printhex(tune_me.confidence);
	lcd_space();
}

void lcd_ve_learn_conf_r(void)
{
	printhex(tune_me.cr);
	lcd_5space();
}

void lcd_ve_learn_conf_m(void)
{
	printhex(tune_me.ck);
	lcd_5space();
}

void lcd_ve_learn_conf_a(void)
{
	printhex(tune_me.ce);
}

void lcd_ve_learn_weight(void)
{
	printhex(tune_me.w00);
	lcd_space();
	printhex(tune_me.w01);
	lcd_space();
	printhex(tune_me.w10);
	lcd_space();
	printhex(tune_me.w11);
}





void lcd_pw(void)
{
	uint16_t t = inj_port[0].pwcalc;
	int2BCDn( timebase_to_10usec( t ), lcd.next_pos, BCD_DIGIT4);
}

void lcd_duty_str(void)
{
        lcd_putstr_P( PSTR("dc:") );
}


/*
Prints the injector duty in %
duty = pw_in_sec / camperiod_in_sec
duty = pw/250000  /  (120 / RPM)

because
60 / RPM is the time (in sec) of a crank rotation
so 
120 / RPM is the time (in sec) of a cam rotation

that simplifies to:
100 * duty = pw * engine.rpm_vhr / 300000;
or
100 * duty = pw * engine.rpm_vhr * 56 / 2**24;

*/
void lcd_duty(void)
{
        uint32_t duty_cycle = mult16_8( inj_port[0].pwcalc, 56 );
        duty_cycle *= engine.rpm_vhr;
        duty_cycle >>= 24;

        int2BCDn(duty_cycle, lcd.next_pos, BCD_DIGIT3);
        lcd_putch('%');
}


void lcd_advstr (void)
{
    lcd_putstr_P(PSTR("adv:"));
}

void lcd_adv (void)
{	// print current ignition advance
//  if (engine.ignadv > 199)
//    lcd_putstr_P(PSTR("??"));
//  else
	uint8_t ignadv=engine.ignadv;
	if(ignadv >= IGN_BALANCE)
	{
		lcd_putch('+');
		int2BCDn((ignadv - IGN_BALANCE)>>2, lcd.next_pos, BCD_DIGIT2);
	}
	else
	{
		lcd_putch('-');
		int2BCDn((IGN_BALANCE - ignadv)>>2, lcd.next_pos, BCD_DIGIT2);
	}
}

void lcd_alien_adv (void)
{	// print alien ignition advance
  if (engine.alien_ignadv > 199)
    lcd_putstr_P(PSTR("??"));
  else
    int2BCDn(engine.alien_ignadv>>1, lcd.next_pos, BCD_DIGIT2);
}

void lcd_dwellstr(void)
{
	lcd_putstr_P(PSTR("dwell:"));
}

void lcd_dwell(void)
{
	int2BCDn( timebase_to_10usec(engine.dwell), lcd.next_pos, BCD_DIGIT3 );
}

void lcd_alien_dwell(void)
{
	int2BCDn(engine.alien_dwell, lcd.next_pos, BCD_DIGIT3);
}

void lcd_adv_diff(void)
{
	uint8_t t;
	lcd_putstr_P(PSTR("Dadv:  "));
	if (engine.alien_ignadv > (engine.ignadv>>1)) {
		t = engine.alien_ignadv - (engine.ignadv>>1);
		lcd_putch('+');
	} else {
		t = (engine.ignadv>>1) - engine.alien_ignadv;
		lcd_putch('-');
	}
	int2BCDn(t, lcd.next_pos, BCD_DIGIT3);
}

void lcd_dwell_diff(void)
{
	uint8_t t;
	lcd_putstr_P(PSTR("Ddwell:"));
	if (engine.alien_dwell > (engine.dwell>>4)) {
		t = engine.alien_dwell - (engine.dwell>>4);
		lcd_putch('+');
	} else {
		t = (engine.dwell>>4) - engine.alien_dwell;
		lcd_putch('-');
	}
	int2BCDn(t, lcd.next_pos, BCD_DIGIT3);
}

// print difference between observed and requested sparks
void lcd_spark_cnt(void)
{
#ifdef BENCHMARK
	uint16_t t;
	lcd_putch('d');
	t = benchstats.trig_sec_cnt - benchstats.trig_prim_cnt;
	printhex(t>>8);
	printhex(t);
#endif
}

// print the number of times something bad happened!
void lcd_ign_err(void)
{
#ifdef BENCHMARK
	lcd_putch('e');
	printhex(benchstats.ign_err>>8);
	printhex(benchstats.ign_err);
#endif
}


void lcd_W(void)
{
    lcd_putch('W');
#ifdef BENCHMARK
    printhex(benchstats.wheel_err);
#endif
}

void lcd_S(void)
{
    lcd_putch('S');
    printhex(etime.tooth_cnt);
}

void lcd_I(void)
{
  uint8_t t;
    lcd_putch('I');
#ifdef BENCHMARK
    t = benchstats.trig_sec_cnt - benchstats.secondary_trigger_cnt;
#else 
	t = 0;
#endif
    printhex(t);
}

void lcd_engine_status_char(void)
{ // prints one character symbolizing the current state of the engine (priority decreases downwards)
  if(engine.status & _BV(crank))
    lcd_putch('C'); // cranking
#ifdef ALS
  else if(engine.status_ext & _BV(als_active))
    lcd_putch('A'); // AntiLagSystem is active
#endif	
  else if(engine.status_ext & _BV(crank_enable))
    lcd_putch('c'); // Enable cranking code
  else if(engine.status & _BV(startw))
    lcd_putch('w'); // warmup enrichment strarted
  else if(engine.status_ext & _BV(o2_not_ready))
    lcd_putch('O'); // O2 sensor warming up
  else if(engine.status & _BV(warmup))
    lcd_putch('W'); // warming up
//  else if(engine.status & _BV(idle))
//    lcd_putch('i'); // idling at normal temp
  else if(engine.status & _BV(tpsaen))
    lcd_putch('a'); // accelerating
  else if(engine.status & _BV(tpsden))
    lcd_putch('d'); // decelerating
  else if(engine.status_ext & _BV(o2_closed_loop))
    lcd_putch('o'); // In O2 closed loop
  else if(engine.status & _BV(running))
    lcd_putch(' '); // engine on constant speed and warmed
  else
    lcd_putch('_'); // engine stopped (shouldn't happen)
}

void lcd_engine_error_char(void)
{  // prints one character about the engine errors (priority decreases downwards)
  if(engine.status_trig & _BV(wheel_error))
    lcd_putch('E'); // Trigger error
  else if(engine.status_ext & _BV(baro_problem))
    lcd_putch('B'); // MAP problem
  else if((etime.tooth_cnt < 128) || (config.primary_trigger & _BV(ptrig_coil)))
    lcd_putch(' '); // Synchronized to wheel (or coil trigger), all OK.
  else
    lcd_putch('S'); // Wheel sync error
}

#ifdef WBO2
void lcd_wbo2_ri(void)
{
	lcd_putstr_P(PSTR("Ri:"));
	printhex(wbo2.Ri>>8);
	printhex(wbo2.Ri);
}
void lcd_wbo2_nernst(void)
{
	lcd_putstr_P(PSTR("Nernst:"));
	printhex(wbo2.nernst_dc>>8);
	printhex(wbo2.nernst_dc);
}
void lcd_wbo2_heat(void)
{
	lcd_putstr_P(PSTR("Heat:"));
	printhex(wbo2.heat);
}
void lcd_wbo2_pump(void)
{
	lcd_putstr_P(PSTR("Pump:"));
	printhex(wbo2.pump_pw);
}
void lcd_wbo2_str(void)
{
	if (wbo2.status & _BV(ACTIVE))
		lcd_putstr_P(PSTR(" ON"));
	else
		lcd_putstr_P(PSTR("OFF"));
}

void lcd_wbo2(void)
{
  printhex(wbo2.Ri>>8);
  printhex(wbo2.Ri);
  lcd_space();
  printhex(wbo2.nernst_dc>>8);
  printhex(wbo2.nernst_dc);
  lcd_space();
  printhex(wbo2.heat);
}

void lcd_wbo2_a(void)
{
  printhex(wbo2.Ri>>8);
  printhex(wbo2.Ri);
  lcd_space();
  printhex(wbo2.heat);
  lcd_putch(',');
}
void lcd_wbo2_b(void)
{
  lcd_space();
  printhex(wbo2.nernst_dc>>8);
  printhex(wbo2.nernst_dc);
  lcd_space();
  printhex(wbo2.pump_pw);
}
void lcd_wbo2_flags(void)
{
  lcd_space();
  if (wbo2.status & _BV(IN_WARMUP))
	  lcd_putch('W');
  else
	  lcd_space();

  if (wbo2.status & _BV(PUMP_ENABLE))
	  lcd_putch('P');
  else
	  lcd_space();

  if (wbo2.status & _BV(FALLBACK_ACTIVE))
	  lcd_putch('F');
  else
	  lcd_space();

}
void lcd_wbo2_benchstats(void)
{
	printhex(benchstats.wbo2_overflow_Ri);
	lcd_space();
	printhex(benchstats.wbo2_overflow_nernst);
	lcd_space();

	printhex(benchstats.ocr0cnt>>8);  //heater overflow
	lcd_space();
	printhex(benchstats.ocr0cnt);  // pump overflow
	
	lcd_space();
	printhex(pump_pid.integral>>8);
	printhex(pump_pid.integral);

//	printhex(benchstats.wbo2_overflow_pumppid);
//	lcd_space();
//	printhex(benchstats.wbo2_overflow_heaterpid);
}


// print afr, make wrapper functions for 'target' and 'current'
// format: xx.yy
void lcd_print_afr(uint16_t afr) {
	uint8_t val, frac;

	val = div1024(afr);
	frac = mult16_8(afr & 0x03ff, 100) >> 10;

	int2BCDn(val, lcd.next_pos, BCD_DIGIT2);
	lcd_putch('.');
	int2BCDn(frac, lcd.next_pos, BCD_DIGIT2);
}

void lcd_print_current_afr(void) {
 	lcd_putstr_P(PSTR("AFR="));
	if (wbo2.current_afr == 0xffff) {
		lcd_putch('?');
		lcd_4space();
	} else {
		lcd_print_afr(wbo2.current_afr);
	}
}

void lcd_print_target_afr(void) {
	uint16_t target;
	lcd_putstr_P(PSTR("AFR="));
	target = calc_target_afr();

	if (target == 0xffff) {
		lcd_putch('?');
		lcd_4space();
	} else {
		lcd_print_afr(target);
	}
}


#endif // WBO2

#ifdef GENBOARDv3
void lcd_knock_window(void)
{
	lcd_putstr_P(PSTR("S:"));
	printhex(knock.window_start>>8);
	printhex(knock.window_start);

	lcd_putstr_P(PSTR(" L:"));
	printhex(knock.window_length>>8);
	printhex(knock.window_length);
}

void lcd_knock_data(void)
{
	lcd_putstr_P(PSTR("D:"));
	printhex(knock_cyl[0].knock_val>>8);
	printhex(knock_cyl[0].knock_val);

	lcd_space();
	printhex(knock.channel);

	lcd_space();
	printhex(knock.status);

#if 0
// this is a dirty hack to test the cam trigger
	lcd_space();
	printhex(engine.cam_period>>8);
	printhex(engine.cam_period);
	lcd_space();
	printhex(engine.engphase);
#endif
}
void lcd_knock_act(void) 
{
	lcd_putstr_P(PSTR("R:"));
	printhex(knock_cyl[0].noise_val>>8);
	printhex(knock_cyl[0].noise_val);

	lcd_putstr_P(PSTR(" ka:"));
	printhex(knock_cyl[0].knock_adjust);

	lcd_space();
	printhex(knock_cyl[0].state);

	lcd_space();
	printhex(knock_stats.adv_err);
}

void lcd_knock_diff(void)
{
	uint16_t t, kv, nv;
	lcd_putstr_P(PSTR(" d: "));
	kv = knock_cyl[0].knock_val;
	nv = knock_cyl[0].noise_val;
	if (kv > nv) {
		t = kv - nv;
		lcd_putch('+');
	} else {
		t = nv - kv;
		lcd_putch('-');
	}
	int2BCDn(t, lcd.next_pos, BCD_DIGIT3);
}

void lcd_knock_stat(void)
{
	lcd_putstr_P(PSTR(" C:"));
	printhex(knock_stats.neg_intensity_cnt>>8);
	printhex(knock_stats.neg_intensity_cnt);
//	printhex(knock_stats.count>>8);
//	printhex(knock_stats.count);

//	lcd_space();
//	printhex(knock.status);
#if 0
// this is a dirty hack to test the cam trigger
#ifdef BENCHMARK
	lcd_space();
	printhex(benchstats.engine_phase_err>>8);
	printhex(benchstats.engine_phase_err);
	lcd_space();
	printhex(benchstats.engphase_nosync);
#endif // BENCHMARK
#endif // 0
}

#endif // GENBOARDv3


// just print some silly test string to see if it works
void lcd_empty(void)
{
	lcd_putstr_P(PSTR("Game over, insert 1 Euro to play the \"tune-the-engine game\" again..."));
}

//#define FLASH __attribute__ ((progmem))

// using the flash (PROGRAM-MEMORY) to store these pointers is non-trivial.
// http://www.avr1.org/pipermail/avr-gcc-list/2002-August/002518.html

// typedef void(*f)(void) funcptr;

#if LCD_COLS==20

// these are just debug dump, they will go away when things are tested throughly
void mik_lineB(void) //11 chars
{
#ifdef BENCHMARK
  lcd_putch('c'); // [c]rank triggers
  printhex(benchstats.trig_sec_cnt>>8);
  printhex(benchstats.trig_sec_cnt);
  lcd_space();
  lcd_putch('i'); // co[i]l triggers
  printhex(benchstats.secondary_trigger_cnt>>8);
  printhex(benchstats.secondary_trigger_cnt);
#endif // BENCHMARK
}

void mik_lineA(void) // 7 chars
{
	uint8_t t;
    t = ego_voltage();
		
    t <<= 1;
    if (t > 99)
		t = 99;
    int2BCDn(t, lcd.next_pos, BCD_DIGIT2);
    lcd_space();

	print_ego_plus_minus(corr.ego);

    lcd_space();
}

#endif // LCD_COLS==20

void lcd_booststr(void) {
#if LCD_COLS==16
	lcd_putstr_P(PSTR("boost T"));
#else
	lcd_putstr_P(PSTR("boost Tgt "));
#endif
}

void lcd_boosttarget(void) {
    int2BCDn(boost_target() >> 4, lcd.next_pos, BCD_DIGIT3);
}

void lcd_boostmap(void) {
	lcd_putstr_P(PSTR(" MAP"));
	int2BCDn(engine.kpa250, lcd.next_pos, BCD_DIGIT3);
}

void lcd_boostpwm(void) {
#ifdef BOOSTCONTROL
	//lcd_putstr_P(PSTR("PWM XX "));
printhex(boost_tps_downscale());
  lcd_putch('i'); // co[i]l triggers
printhex(engine.boostmax_target);
// XXX needs to be implemented XXX
#endif
}

void lcd_egtstr(void) {
 	lcd_putstr_P(PSTR("egt:"));
}

void lcd_egt1(void) // 4chars
{	
	uint16_t egtreading=engine.egt;
	if(egtreading > 9999){	
	// can only happen if overflowed (most likely for config.egt1_offs == 0x80)
 		egtreading = 0;
	}	
	
	// not equal-time, and not read when LCD on another page
	int2BCDn(egtreading, lcd.next_pos, BCD_DIGIT4);	
}



// NOTES: we store function pointers by lines, not by pages.
// 4x20 and 2x20 has shared maintenance, and page up/down will be possible

// LINEEND1 is an indication that lcd_clr_rest_of_line could be needed. Could be value=0x0001. 
// LINEEND2 means lcd_clr_rest_of_line not needed, since line is filled (can result in linend garbage if misused)
// for now, lcd_clr_rest_of_line always called
#define LINEEND1 NULL
#define LINEEND2 NULL

/* ======================= short lines: used for both LCD_COLS== 16 or 20 =============== */
//PGM_VOID_P lcd_line_intro[] PROGMEM = { lcd_introstr, lcd_firmware_versionstr, lcd_firmware_version, LINEEND1 };

PGM_VOID_P lcd_line_80[] PROGMEM = { lcd_egtstr, lcd_egt1, LINEEND1 }; // EGT view
PGM_VOID_P lcd_line_ee[] PROGMEM = { lcd_space, LINEEND1 }; // empty line


// ========================================= LCD_4x20 and LCD_2x20
#if LCD_COLS==20
PGM_VOID_P lcd_line_00[] PROGMEM = { lcd_rpmstr, lcd_rpm_true_4char, lcd_space, lcd_Tstr, lcd_T, lcd_space, lcd_Pstr, lcd_P, LINEEND1 };
PGM_VOID_P lcd_line_01[] PROGMEM = { lcd_O2str, lcd_O2, lcd_space, lcd_egocorr, LINEEND1 };
PGM_VOID_P lcd_line_02[] PROGMEM = { lcd_coolantstr, lcd_coolant, lcd_2space, lcd_tpsstr, lcd_tps_percent,  LINEEND1 };

PGM_VOID_P lcd_line_03[] PROGMEM = { 
	lcd_pw, lcd_space, 
#ifdef LCD_IAC_PREFERRED
	lcd_iacstr, lcd_iacpos, 
#else	// display EGT
	lcd_egtstr, lcd_egt1,
#endif	// LCD_IAC_PREFERRED
	lcd_idlerpm, lcd_shiftcut_state, LINEEND1
};

PGM_VOID_P lcd_line_10[] PROGMEM = { lcd_rpmstr, lcd_rpm_true_4char, lcd_space, lcd_W, lcd_S, lcd_I, LINEEND1 };
PGM_VOID_P lcd_line_11[] PROGMEM = { mik_lineA, mik_lineB, LINEEND1 };
PGM_VOID_P lcd_line_12[] PROGMEM = { lcd_coolantstr, lcd_coolant, lcd_2space, lcd_advstr, lcd_adv, LINEEND1 };
PGM_VOID_P lcd_line_13[] PROGMEM = { lcd_pw, lcd_space, lcd_iacstr, lcd_iacpos, lcd_idlerpm, lcd_space, LINEEND2 };

PGM_VOID_P lcd_line_20[] PROGMEM = { lcd_rpmstr, lcd_rpm_true_4char, lcd_space, lcd_W, lcd_I, lcd_space, lcd_Pstr, lcd_P, LINEEND1 };


#ifdef WBO2
PGM_VOID_P lcd_line_21[] PROGMEM = { lcd_wbo2_a, lcd_wbo2_b, lcd_wbo2_flags, LINEEND1 };
PGM_VOID_P lcd_line_22[] PROGMEM = { lcd_wbo2_benchstats, LINEEND1 };
PGM_VOID_P lcd_line_23[] PROGMEM = { lcd_print_current_afr, lcd_space, lcd_print_o2percent_old, lcd_print_current_o2, LINEEND1 }; 
#else
PGM_VOID_P lcd_line_21[] PROGMEM = { LINEEND1 };
PGM_VOID_P lcd_line_22[] PROGMEM = { lcd_coolantstr, lcd_coolant, lcd_2space, lcd_advstr, lcd_adv,  LINEEND1 };
PGM_VOID_P lcd_line_23[] PROGMEM = { lcd_pw, lcd_space, lcd_iacstr, lcd_iacpos, lcd_idlerpm, lcd_space,  LINEEND2 };
#endif // WBO2

// IAC view
PGM_VOID_P lcd_line_30[] PROGMEM = { lcd_iacstr, lcd_iac_status, lcd_space, lcd_rpm_true_4char, lcd_idlerpm, LINEEND2 };
PGM_VOID_P lcd_line_31[] PROGMEM = { lcd_iac_pid_e, lcd_space, lcd_iac_pid_i, lcd_space, lcd_iac_pid_d, LINEEND2 };
PGM_VOID_P lcd_line_32[] PROGMEM = { lcd_iac_pid_pterm, lcd_space, lcd_iac_pid_iterm, lcd_space, lcd_iac_pid_dterm, LINEEND2 };
PGM_VOID_P lcd_line_33[] PROGMEM = { lcd_pw, lcd_space, lcd_iacpos, lcd_space, lcd_iac_advance, lcd_space, lcd_iac_pid, LINEEND2 };

// VE learn view
PGM_VOID_P lcd_line_40[] PROGMEM = { lcd_ve_learn_str, lcd_ve_learn_status, LINEEND2 };
PGM_VOID_P lcd_line_41[] PROGMEM = { lcd_ve_learn_stat_r, lcd_ve_learn_stat_m, lcd_ve_learn_stat_a, LINEEND2 };
PGM_VOID_P lcd_line_42[] PROGMEM = { lcd_ve_learn_conf, lcd_ve_learn_conf_r, lcd_ve_learn_conf_m, lcd_ve_learn_conf_a, LINEEND2 };
PGM_VOID_P lcd_line_43[] PROGMEM = { lcd_ve_learn_weight, LINEEND1 };

#ifdef GENBOARDv3
// Ignition view
PGM_VOID_P lcd_line_50[] PROGMEM = { lcd_advstr, lcd_adv, lcd_space, lcd_dwellstr, lcd_dwell, LINEEND1 };
PGM_VOID_P lcd_line_51[] PROGMEM = { lcd_knock_window, lcd_knock_stat,  LINEEND2 };
PGM_VOID_P lcd_line_52[] PROGMEM = { lcd_knock_data, lcd_knock_diff, LINEEND2 };
PGM_VOID_P lcd_line_53[] PROGMEM = { lcd_knock_act, lcd_space, LINEEND1 };

// Alien ignition advance/dwell view
PGM_VOID_P lcd_line_60[] PROGMEM = { lcd_advstr, lcd_adv, lcd_space, lcd_alien_adv, lcd_space, lcd_rpmstr, lcd_rpm_true_4char, LINEEND1 };
PGM_VOID_P lcd_line_61[] PROGMEM = { lcd_dwellstr, lcd_dwell, lcd_space, lcd_alien_dwell, lcd_space, lcd_battery_volts, LINEEND1 };
PGM_VOID_P lcd_line_62[] PROGMEM = { lcd_adv_diff, lcd_space, lcd_spark_cnt, LINEEND1 };
PGM_VOID_P lcd_line_63[] PROGMEM = { lcd_dwell_diff, lcd_space, lcd_ign_err, LINEEND1 };
#endif // GENBOARDv3

// WBO2 info
#ifdef WBO2
PGM_VOID_P lcd_line_70[] PROGMEM = { lcd_wbo2_ri, lcd_space, lcd_wbo2_nernst, LINEEND1 };
PGM_VOID_P lcd_line_71[] PROGMEM = { lcd_wbo2_heat, lcd_space, lcd_wbo2_pump, LINEEND1 };
PGM_VOID_P lcd_line_72[] PROGMEM = { lcd_print_current_afr, lcd_space, lcd_print_current_o2, LINEEND1 };
PGM_VOID_P lcd_line_73[] PROGMEM = { lcd_print_target_afr, lcd_wbo2_flags, lcd_4space, lcd_wbo2_str, LINEEND2 };
#endif // WBO2

// BOOSTCONTROL info
PGM_VOID_P lcd_line_90[] PROGMEM = { lcd_booststr, lcd_boosttarget, lcd_boostmap, LINEEND1 };
PGM_VOID_P lcd_line_91[] PROGMEM = { lcd_boostpwm, lcd_tpsstr, lcd_tps, LINEEND1 };


PGM_VOID_P lcd_line_list[] PROGMEM = {
	lcd_line_00, lcd_line_01, lcd_line_02, lcd_line_03,
	lcd_line_10, lcd_line_11, lcd_line_12, lcd_line_13,
	lcd_line_20, lcd_line_21, lcd_line_22, lcd_line_23,
	lcd_line_30, lcd_line_31, lcd_line_32, lcd_line_33,
	lcd_line_40, lcd_line_41, lcd_line_42, lcd_line_43,

#ifdef GENBOARDv3
	lcd_line_50, lcd_line_51, lcd_line_52, lcd_line_53,
	lcd_line_60, lcd_line_61, lcd_line_62, lcd_line_63,
#else
	lcd_line_ee, lcd_line_ee, lcd_line_ee, lcd_line_ee,
	lcd_line_ee, lcd_line_ee, lcd_line_ee, lcd_line_ee,
#endif // GENBOARDv3

#ifdef WBO2
	lcd_line_70, lcd_line_71, lcd_line_72, lcd_line_73,
#else
	lcd_line_ee, lcd_line_ee, lcd_line_ee, lcd_line_ee,
#endif // WBO2

	lcd_line_80, lcd_line_ee, lcd_line_ee, lcd_line_ee,
#ifdef BOOSTCONTROL
	lcd_line_90, lcd_line_91, lcd_line_ee, lcd_line_ee,
#endif // BOOSTCONTROL
	NULL
	};

#endif // LCD_COLS==20


#if LCD_COLS==16

PGM_VOID_P lcd_line_00[] PROGMEM = {lcd_rpmstr, lcd_rpm_true_4char, lcd_Tstr, lcd_T, lcd_Pstr, lcd_P, LINEEND2 };  // 4+4 + 4+4

#ifdef WBO2
PGM_VOID_P lcd_line_01[] PROGMEM = { lcd_wbo2, LINEEND2 };
#else
PGM_VOID_P lcd_line_01[] PROGMEM = { lcd_O2str, lcd_O2, lcd_egocorr, LINEEND2 }; // 5 + 3 + 6+2
#endif

PGM_VOID_P lcd_line_02[] PROGMEM = { lcd_coolantstr, lcd_coolant, lcd_tpsstr, lcd_tps_percent, LINEEND2 }; // 8+8
PGM_VOID_P lcd_line_03[] PROGMEM = { lcd_pwstr,lcd_pw, lcd_iacpos, lcd_iacdir, lcd_idlerpm, LINEEND2 };

PGM_VOID_P lcd_line_10[] PROGMEM = { lcd_empty, LINEEND1 };


#if LCD_LINES==4
PGM_VOID_P lcd_line_list[] PROGMEM = {
	lcd_line_00, lcd_line_01, lcd_line_02, lcd_line_03,
	lcd_line_10, NULL, NULL, NULL,
	NULL
};
#endif // LCD_LINES==4


// ===================================== LCD_2x16 and LCD_1x16
#if LCD_LINES<=2

PGM_VOID_P lcd_line_x1[] PROGMEM = { lcd_coolant, lcd_O2, lcd_engine_status_char, lcd_pw, lcd_engine_error_char, lcd_tps_percent, LINEEND2 }; // 4+3+1+4+1+3

PGM_VOID_P lcd_line_list[] PROGMEM = {
	lcd_line_00, lcd_line_x1,
	NULL
};
#endif // LCD_LINES<=2


#endif // LCD_COLS==16



#ifndef FUNCARR_MAXIDX
#define FUNCARR_LEN sizeof(lcd_line_list)
// word based vs. index based
#define FUNCARR_MAXIDX ((FUNCARR_LEN-4)/2)
#endif // FUNCARR_MAXIDX


// display a line "bRxx"
// b=bin number, R=rpm table, xx=weight of upper bin
// xx==0 : the upper bin (b) is used with weigth 0
// xx==ff : the upper bin (b) is used with weigth ff
static __inline__ void lcd_bin_line(void)
{
  uint8_t val;

  lcd_putch(hex_lo(rpm_idx.uindex));
  lcd_putch('R');
  if (rpm_idx.uindex == 0) // special case
    printhex(0xFF);
  else
    printhex( linear_interp_fixed_y(ve.rpm_range[rpm_idx.lindex], ve.rpm_range[rpm_idx.uindex], engine.rpm) );

  lcd_putch(' ');
  lcd_putch(hex_lo(kpa_idx.uindex));
  lcd_putch('K');
  if (kpa_idx.uindex == 0)
    printhex(0xFF);
  else
    printhex( linear_interp_fixed_y(ve.kpa_range[kpa_idx.lindex], ve.kpa_range[kpa_idx.uindex], engine_kpa8()) );

  lcd_putch(' ');
  lcd_putch(hex_lo(clt_idx.uindex));
  lcd_putch('T');
  // we don't have a rx21 table for the temperature, calculate distance by "by hand"
  
  if (clt_idx.uindex == 0)
    val = 0xff;
  else if (engine.coolant <= config.warmup_clt_range[clt_idx.lindex])
    val = 0;
  else if (engine.coolant >= config.warmup_clt_range[clt_idx.uindex])
    val = 0xff;
  else {
    uint8_t dx, diff;
    dx = engine.coolant - config.warmup_clt_range[clt_idx.lindex];
    diff = config.warmup_clt_range[clt_idx.uindex] - config.warmup_clt_range[clt_idx.lindex];
    val = ((uint16_t)dx*0xff) / diff;
  }

  printhex(val);

}
#define BINS_START (LCD_LINES-1)*LCD_COLS


/* finds the relevant string in magicstr and prints onto LCD
 * to inform the user which config setting is being updated.
 * in this case since we've already completed the update and
 * moved to the next position we have to display the "previous"
 * item
 * TODO: this is the longest executing lcd function 
 * (almost as long as table lookups?)
 */
static __inline__ void lcd_config_magicstr(struct menu_t *m)
{
	PGM_VOID_P P;
	uint8_t i, idx, *str;
	
	idx = m->posh << 4 | m->posl;	// posh:posl counts as a flat byte here
	P = (PGM_VOID_P)(magicstr + idx);
	// P = magicstr[ m->posl ];
	
	i = 0;
	while((str=(void*)pgm_read_addr(P)) == NULL){
		P -= sizeof(void*);
		i++;
	}
	
	lcd_putstr_P( PSTR("tc") );
	printhex(idx-i);	// the position of the configured byte
	lcd_putch('/');
	lcd_putch(hex_lo(i));	// this is the subposition inside (max 15)
	lcd_putch('=');
	printhex(m->tab[idx]);
	lcd_putch(' ');

	if ( debug.disable_flags & _BV(DIS_FLAG_LCDBIN)) {  // show the 'bins' on the last line
	  lcd_putstr_P_max(str, BINS_START-10);   // 10 is because of the initial "tc01/0=00 "
	  lcd_clr_to_idx(BINS_START-1);
	  lcd_bin_line();
	  lcd_clr_to_end();
	} else {  // normal behavior
	  lcd_putstr_P_max(str, LCD_CHARS-10);
	  lcd_clr_to_end();
	}
}


/* prints a few hexa numbers from the array to the lcd, bails out after maxch chars.
marker marks the 0th, 1st, etc... number with a little arrow (use 0xff for no marker)
*/
static __inline__ void lcd_hex_line(uint8_t *pt, uint8_t maxch, uint8_t marker, uint8_t space)
{
	uint8_t p;	// how many numbers have we printed ?
	p=0;
	
	do {
		if(maxch>1) {
			printhex(*pt++); 
			maxch -= 2;
		}
		if(maxch) {
			// zero is marked with a leftarrow, a little special
			lcd_putch( (p++ == 0 && marker==0 ) ? '<' : (p==marker? '>' : space));
			maxch --;
		}
	} while (maxch>1);
}

// similar, but used for displaying the difference between ve learnt and ve_ref table
static __inline__ void lcd_hex_line_diff(uint8_t *pt, uint16_t *pp, uint8_t maxch, uint8_t marker, uint8_t space)
{
	uint8_t p, diff;	// how many numbers have we printed ?
	p=0;
	
	do {
		if(maxch>1) {
			diff = ((*pp)>>8) - *pt;
			printhex(diff);
			pp++; pt++; 
			maxch -= 2;
		}
		if(maxch) {
			// zero is marked with a leftarrow, a little special
			lcd_putch( (p++ == 0 && marker==0 ) ? '<' : (p==marker? '>' : space));
			maxch --;
		}
	} while (maxch>1);
}

	// after the leading 3 chars:
	// for 20 col display we can display 6
#if LCD_COLS==20
#define DN 6
#define DN_1DIM 7   //FF_FF_FF_FF_FF_FF_FF
#endif
	// for 16 col display we can display 5, with dropping ':' at the line beginning
#if LCD_COLS==16
#define DN 5
#define DN_1DIM 5   //FF_FF_FF_FF_FF
#endif
	// numbers

	// TODO: we always print rpmrange and kparange, although this is silly...
//#define LINS VE_SIZE_KPA
#define COLS VE_SIZE_RPM

/*
LCD-linenum to display on
table-line 
table-col to start with
table pointer
*/
static __inline__ void linebeg(uint8_t lin, uint8_t pl, uint8_t pc, struct menu_t *m)
{
	uint8_t t, marker, space;
	lcd_setpos(lin * LCD_COLS); 
	printhex(ve.kpa_range[pl]);	// TODO: not always appropriate
#if LCD_COLS!=16
	lcd_putch(':');
#endif
	t = COLS*pl + pc;
	marker = (pl == m->posh ? m->posl - pc : 0xff);
	space = (pl == kpa_idx.uindex ? '-' : ' ');

	if(m->table == 'd'){ // difference of ve.ve_table (learnt) and ve_ref
		lcd_hex_line_diff(&(ve.ref_table[t]), &(ve.ve_table[t]), DN*3-1, marker, space);
	} else {
		lcd_hex_line(&(m->tab[t]), DN*3-1, marker, space);
	}
}

/* 
read its usage first
finds the window start-index for a limited subset(displayable) of the array(elements) to be displayed: 
idx is the must-be there element, at the preferred location. Thus:
idx-preferred is the preferred return value, if otherwise in range
*/
static __inline__ uint8_t find_start(uint8_t idx, uint8_t elements, uint8_t displayable, uint8_t preferred)
{
	if (idx <= preferred) // position is among the first numbers
		return 0;	
	if(idx >= (elements-displayable+preferred)){ // position is among the last numbers
		return elements-displayable;
	} else 	{
				return idx-preferred ;	// somewhere in the middle
			}
}


static __inline__ void lcd_table(void)
{
	uint8_t pl, pc;		// line and column of displayed topleft
	// uint8_t *tab;
	struct menu_t *m;
	// TODO: currently we use the last menu directly (that we got via lcd_display_pos())
	m = lcd_disp.menu;
	
	// TODO: currently we print to LCD directly (should be possible to rs232 with long lines - even VT100)

	switch(m->table){
	case 'c':	// config is special. magicstr is used
		lcd_setpos(0);
		lcd_config_magicstr(m);
		return;
	case 'd':
	case 'h':
	case 'j':
	case 'l':
	case 'n':
	/* funcarridx is a line number when we are displaying a table,
	since 2 dimensional tables are displayed in several runs (one line at a time)	
	*/
	if(lcd_disp.funcarridx >= LCD_LINES)
		lcd_disp.funcarridx = 0;

	pl = find_start(m->posh, VE_SIZE_KPA, LCD_LINES-1, LCD_LINES>2 ? 1 : 0);
	pc = find_start(m->posl, VE_SIZE_RPM, DN, 2);
	// tab = m->tab;
	switch(lcd_disp.funcarridx){
		case 0:
#if LCD_COLS!=16
#define STARTSTR "tt"
#else
#define STARTSTR "t"
#endif
			update_chars(STARTSTR, 0 * LCD_COLS);
			lcd_putch(m->table);
			// we also mark the rpm for info
			// TODO: how to mark the kpa???
			lcd_hex_line(&ve.rpm_range[pc], LCD_COLS-sizeof(STARTSTR), rpm_idx.uindex - pc, ' ');
			break;
		case 1:
			linebeg(1, pl, pc, m);
			break;
		case 2:
			linebeg(2, pl+1, pc, m);
			break;
		case 3:
			if (m->options & _BV(SHOW_BINS)) {  // show the 'bins' on the last line
				lcd_clr_to_idx(BINS_START-1);
				lcd_bin_line();
				lcd_clr_to_end();
			} else
				linebeg(3, pl+2, pc, m);
			break;
	}
	return;
	case 'k':
	case 'r':
	default:	// all one dimension tables look this way:
		update_chars(STARTSTR, 0);
		lcd_putch(m->table);
		lcd_putstr_P(PSTR(" pos="));
		printhex(m->posl);	// the position of the configured byte
		lcd_putstr_P(PSTR(" val="));
		printhex(m->tab[m->posl]);
		lcd_clr_to_end();	// it's a little waste of cycles...
		lcd_setpos(LCD_COLS); // go to 2nd line
		pc = find_start(m->posl, m->dimension+1, DN_1DIM, 2);	// m->dimension upper half is 0 anyway
		lcd_hex_line(&m->tab[pc], min(LCD_COLS, (m->dimension+1)*3-1), m->posl - pc, ' ');
		return;
	} // switch(m->table)
}

// checks lcd_line_list array of pointers to function pointer array to find what to do
static inline void lcd_nextfunc(void)
{
	voidfunc_P func;
	psize_t p;
	uint8_t lineidx, funcidx, t;

	t = lcd_disp.funcarridx;
	if( t == 0xFF ){
		// leftover behaviour: some initialization assigns FF
		t = 0;
	}

	lineidx = t >> 4;
	funcidx = t & 0x0F;

	t = lcd_disp.firstline + lineidx;
	// non-cyclic behaviour: 
	// min(t, FUNCARR_MAXIDX));

	// cyclic:
	if( t > FUNCARR_MAXIDX ){
		t = 0;
	}

	//p = pgm_read_word_near( ... ) ?
	p = pgm_read_word(lcd_line_list + (uint16_t) t);
	if ( p == 0 ){ // this is bad, recover by defaulting to the first 'menu'
		// because of above "min()" function it only happens when there are more than 1 NULL, ...
		p = pgm_read_word(lcd_line_list);
		lineidx=0;
		funcidx=0;
	}

	// p points to the array of functions for given line

	func = (PGM_VOID_P)pgm_read_word( p + ((uint16_t)funcidx << 1) );

	if (func){
		func(); // now do something
		funcidx++;
	} else {
		// func==NULL so next line
		funcidx = 0;
		if( lineidx >= LCD_LINES-1 ){
			// last line, let's go back:
			lineidx = 0;
		} else {
			lineidx++;
		}

		// don't call lcd_clr_rest_of_line() at the beginning of the line
		// because it's most likely just wrapped over. 
		// use lcd_space before this to clear line
		if( lineidx * LCD_COLS !=  lcd.next_pos ){
			lcd_clr_rest_of_line(); 
		}

		// goto beginning of the line even if we know we are there
		// somewhat suppress the effect of comm errors
		lcd_GLx( lineidx );
	}

	lcd_disp.funcarridx = (lineidx << 4) + funcidx;
}

void lcd_do_something()
{
	switch(lcd_disp.mode){
		case DISPLAY_INTRO:
                        // after appr. 2 seconds move on to DISPLAY_NORMAL
                        if (etime.now_upper > 8) {
                          lcd_disp.mode=DISPLAY_NORMAL;
                          lcd_cls();
                          lcd_dirty();
                          return;
                        }
			
			// TODO: consider calling lcd_nextfunc(); with lcd_line_list passed as argument
			lcd_setpos(0);
			lcd_introstr();
                        // line 2
			lcd_setpos(LCD_COLS);
			lcd_firmware_versionstr();
			return;
		case DISPLAY_NORMAL:
			lcd_nextfunc();
			return;
		case DISPLAY_TABLE:
			lcd_disp.funcarridx++;
			lcd_table();
			return;

		case DISPLAY_HOLD:	// nothing to do
			return;
	} // switch
}

// sets pointer to 0 and whatever mode (1 for menu, 0 for original)
// specifies which menu to use for the display
__inline__ void lcd_display_mode(uint8_t mo, struct menu_t *m)
{
	lcd_disp.funcarridx=0xFF;
	lcd_disp.mode = mo;
	lcd_disp.menu = m;
	lcd_disp.param = 0;
}

// one line down
void lcd_display_lineinc(uint8_t x)
{
	lcd_disp.firstline = ( lcd_disp.firstline > FUNCARR_MAXIDX - x) ? 0 : lcd_disp.firstline + x ;
}

// one line up
void lcd_display_linedec(uint8_t x)
{
	if(lcd_disp.firstline >= x){
		lcd_disp.firstline -= x;
	}
}

#endif // LCD

/*@}*/

