/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen mik at caffrey.dk]        *
***********************************************************
*  multitooth decoding                                    *
\*********************************************************/

#ifndef __MULTITOOTH_C__
#define __MULTITOOTH_C__

#include "global.h"
#include "eventqueue.h"
#include "fuelcontrol.h"
#include "benchmark.h"

#include "arithm.c"

#include <inttypes.h>

#define TW_TRIG    0
#define TW_MISSING 1

#define invalidate_rpm_period() engine.rpm_period = 0xFFFFffff
#define isvalid_rpm_period() ((uint16_t)engine.rpm_period) != 0xffff

#define TOOTHREL_MISSING (0)
#define TOOTHREL_SINGLE (1)
#define TOOTHREL_ERROR (0xFF)

/* ===========
 might move elsewhere, similar action for coil-type in timing.c
 only call from interrupt
*/
static inline void flag_rpm_calc( uint32_t capturedtime )
{
	engine.rpm_period = capturedtime - etime.prev_rpm_tooth;
	etime.prev_rpm_tooth = capturedtime;
	engine.status_ext |= _BV(new_rpm);  // start a rpm calculation
}

/* ===========
 *  syncronize igncount if not ignited yet
 */
static inline void syncronize_igncount( void )
{
	if( engine.ignstate==0 ){
		engine.igncount=config.ignchmax;
	}
}

// ==============

// true if b > 1.5 * ave (NOTE: avehalf must be ave/2)
#define significantly_bigger(b, ave, avehalf) ((b > ave) && ((b - ave) > avehalf))

static inline uint16_t scale16_4(uint16_t ave, uint8_t x)
{
	uint32_t t32 = mult16_8( ave, x) >> 4;
	return t32 & 0xFFFF;
}

static inline uint16_t scale16_4_clamped(uint16_t ave, uint8_t x)
{
	uint32_t t32 = mult16_8( ave, x) >> 4;
	uint8_t upper = t32 >> 16; 
	// check for overflow
	return upper == 0 ? t32 & 0xFFFF : 0xFFFF ;
}

/*
 * single tooth return TOOTHREL_SINGLE
 * missing tooth return TOOTHREL_MISSING
 * error return TOOTHREL_ERROR
 *
 * config examples:
 * 60-2: config.toothrel_normal=0x7A
 * N-1:  config.toothrel_normal=0x79
 *
 * 60-2: config.toothrel_missing=0xB5
 * N-1:  config.toothrel_missing=0x00
 */
static inline uint8_t tooth_relative_length( uint16_t b, uint16_t ave, uint16_t avehalf )
{
	uint8_t t8;
	uint16_t diff, t16;

	// Primitive multitooth test:
	if( (config.primary_trigger & _BV(trigger_multitooth_hairy)) == 0 ){
		// traditional behaviour, simple detection:
		if( significantly_bigger( b, ave, avehalf )){
			return TOOTHREL_MISSING;
		}
		return TOOTHREL_SINGLE;
	}

	// Fiat Stilo trigger
	// shortest tooth is detected as TOOTHREL_MISSING
	// see manual/images/TriggerAdvancedFilterMovingAverage.xls
	// or think about the moving average filter:
	// 60,90,30 => 56,73,51 so relative 1.17, 1.62, 0.41
	// with the screwed average 51, 71, 43 so relative 1.4, 1.74, 0.42
	//
	// so recommended accepted ranges:
	// 0.25 .. 0.75 TOOTHREL_MISSING config.toothrel_missing=0x4C
	// 1..2.25 is TOOTHREL_SINGLE	config.toothrel_normal=0x04

	if( config.primary_trigger & _BV(trigger_fiatstilo) ){

		// usually bigger, so this branch first:
		if( b >= ave ){
			t8 = config.toothrel_normal ;
			diff = b - ave;
			t16 = scale16_4_clamped( ave, 16 + (t8 & 0x0F));
			if( diff < t16 ){
				return TOOTHREL_SINGLE;
			}

			 return TOOTHREL_ERROR;
		}

		// b < ave
		t8 = config.toothrel_missing ;
		
		t16 = scale16_4( ave, (t8 & 0x0F));
		if( b > t16 ){
			return TOOTHREL_ERROR;
		}

		t16 = scale16_4( ave, (t8 >> 4));
		if( b < t16 ){
			return TOOTHREL_ERROR;
		}

		return TOOTHREL_MISSING;
	}

	// Advanced implementation of N-1:
	// Worked well from the beginning, just configuration example was bad.

	// upper and lower threshold for normal tooth length
	t8 = config.toothrel_normal ;
	if( b >= ave ){
		diff = b - ave;
		// longer than average

		// threshold for normal tooth:
		t16 = scale16_4( ave, t8>>4 );
		if( diff <= t16 ){
			// difference is lower or equal than threshold (eg. 6/16) for normal tooth
			return TOOTHREL_SINGLE;
		}

		// from now, t8 is the upper and lower threshold for missing tooth length
		t8 = config.toothrel_missing ;
		t16 = scale16_4_clamped( ave, 4 + (t8 & 0x0F));

		// take half to prevent overflow. and threshold for missing tooth specified in 1/8 (not 1/16)
		diff >>= 1;
		if( diff < t16 ){
			// diff is less than threshold (eg. 9 / 8), so not long enough for a missing tooth
			return TOOTHREL_ERROR;
		}

		t16 = scale16_4_clamped( ave, 12 + (t8 >> 4));

		if( diff <= t16 ){
			return TOOTHREL_MISSING;
		}

		// ouch, diff is longer than expected for missing tooth
		return TOOTHREL_ERROR;
	}

	diff = ave - b;
	// TODO: diff = ave - b makes no sense; change config semantix so it's eliminated!
	t16 = scale16_4( ave, (t8 & 0x0F));
	if( diff <= t16 ){
		return TOOTHREL_SINGLE;
	}

	// OOPS .. too low
	return TOOTHREL_ERROR;
}

// ========================
/*
	Tooth detection.
	Used from trigger interrupt.
  
	Bit definitions of trigger:
    bit 0: trigger
    bit 1: missing tooth
	
	tooth wheel in sync when etime.tooth_cnt < 128
	
	tooth_detection operates on 16 bit period instead of 32 bit
*/
static inline uint8_t tooth_detection(uint16_t period16, uint32_t capturedtime) {
	uint8_t trigger = 0, tooth_cnt, c_trigger_tooth, c_another_trigger_tooth;
	uint16_t etime_atp, etime_atp2;

	tooth_cnt = etime.tooth_cnt;
	
	etime_atp = etime.average_tooth_period;
	etime_atp2 = etime_atp >> 1;	// half
	if (tooth_cnt < 128) {	// wheel_in_sync
		// We could handle tooth_cnt==127 which should not normally appear
		uint8_t toothlen;
		// tooth_cnt is normally 0..57 (for 60-2)
		// 0 .. 117 for 60-2 with camsync
		// config.tooth_wheel is 58 in either case
		uint8_t ctw = config.tooth_wheel;
		toothlen = tooth_relative_length(period16, etime_atp, etime_atp2);

		switch( toothlen ){
			uint8_t tooth_check;
		case TOOTHREL_MISSING:		// http://www.vems-group.org/wiki/index.php?page=GenBoard%2FUnderDevelopment%2FFirmWare%2FTriggerRelated

			// For fiatstilo,
			// (now - etime.strig_capturedtime_prev) > 270 crankdeg condition could be used 
			// instead of _BV(strig_cam_sync) flag
			// since (config_secondary_trigger & _BV(strig_cam_sync)) not set
			if( config.secondary_trigger & _BV(strig_cam_sync) ){
				if( engine.status_trig & _BV(cam_sync3)
				// || tooth_cnt == 117 // TODO: rethink what else needed to enable limphome without camsync
				){
					engine.status_trig &= ~_BV(cam_sync3);
					tooth_check=tooth_cnt;
					if(config.primary_trigger & _BV(trigger_fiatstilo)){
						tooth_check -= 4;
					} else {
						// For 30-1, tooth_cnt==0..28 than 30..58
						// 58 is accepted as tooth_check=ctw=29
						tooth_check -= ctw;
						if( 0 == (config.primary_trigger & _BV(trigger_multitooth_Nminus1))){
						// For 60-2, tooth_cnt==0..57 than 60..117
						// 117 is accepted as tooth_check=ctw=58;
						// but 117-58 is 59 so we must decrease to become 58:
							tooth_check --;
						}
					}
					syncronize_igncount();
					tooth_cnt=0;
					trigger = _BV(TW_MISSING);
				} else {
					// without flagged camsync, let tooth_cnt count up ( over 60, to 117 )
					// (instead of +=1 or resetting).
					// Necessary for the N-1 or N-2, eg.60-2 ( 3-tooth long gap )
					tooth_check = tooth_cnt + 1;
					// syncronize_igncount();	// NOT applied here !!
					tooth_cnt+= (config.primary_trigger & _BV(trigger_multitooth_Nminus1) ? 2 : 3);
				}
			} else { // camsync not configured
				tooth_check=tooth_cnt + 1;
				syncronize_igncount();
				tooth_cnt=0;
				trigger = _BV(TW_MISSING);
			}

			if (tooth_check == ctw) {
#ifdef ALLOW_SLOPPY_TOOTHWHEEL
			// another chance:
			// if a toothwheel sometimes counts 60-3 instead of 60-2
			// this most likely means the tooth after the missing tooth was lost
			// trigger came later _LAST_TIME_:
			// therefore it _was_ timing 6 degrees later, well..
			// Makes no sense to lose the next rotation.
			} else if( tooth_check == ctw - 1 ){
#ifdef BENCHMARK
				// increase wheel_err so it gets noticed
				benchstats.wheel_err++;
#endif	// BENCHMARK
#endif // ALLOW_SLOPPY_TOOTHWHEEL
			} else {
				// something went wrong, result of the tooth wheel decoding cannot be trusted
				// leave last known good engine.rpm_period as it is
//				invalidate_rpm_period();
				engine.status_trig |= _BV(wheel_error);
#ifdef BENCHMARK
				benchstats.wheel_err++;
#endif	// BENCHMARK
				// reenter sync-mode, the only sane thing to do
				etime.tooth_cnt = 255;
				return 0;
			}

			// so averaging is not screwed that much:
			// perfect for N-1 wheel, *0.333 could be used for N-2
			// Tricky for FiatStilo: 
			// 30deg becomes 15, but it does not disturb detection at all
		        //if( 0 == (config.primary_trigger & _BV(trigger_fiatstilo)) ){
				etime_atp2 >>= 1;
			//}

			// flag_rpm_calc( capturedtime ); // happens later
			break;
		case TOOTHREL_SINGLE:
			tooth_cnt++;
			break;
		case TOOTHREL_ERROR:
		default:
#ifdef BENCHMARK
			benchstats.wheel_err++;
#endif  // BENCHMARK
			// start sync sequence all from the beginning
			etime.tooth_cnt = 255;
			return 0;
		} // switch ( toothlen )
	} else { // NOT wheel_in_sync
#define TOOTH_WHEEL_NOISE 3
#define TOOTH_WHEEL_CNT   5
#define TOOTH_WHEEL_LISTEN (TOOTH_WHEEL_NOISE + TOOTH_WHEEL_CNT)

		engine_is_running(); //turn on fuelpump etc.
// do a rather long sync sequence to avoid getting wild rpm readings during the first rotation

		if (tooth_cnt < (255 - TOOTH_WHEEL_LISTEN)) {
			// initialization done, find the missing tooth
			if (tooth_relative_length(period16, etime_atp, etime_atp2) == TOOTHREL_MISSING ) {
				// uint32_t t32;
				engine.status_trig &= ~_BV(wheel_error);

				// TODO: use timing data from more teeth
				// the >> 1 is because period16 is for 2 teeth at this point
				// would be good for 36-1 only, since 60-2 is 3 teeth !
				// t32 = mult16_8( period16, config.another_trigger_tooth ) >> 1;
				// engine.rpm_period = ((t32 >> 16) ? 65535 : (uint16_t) t32);
				// we just make sure that TW_TRIG not fired in this case:

				invalidate_rpm_period();
	
				etime.prev_rpm_tooth = capturedtime;
				trigger |= _BV(TW_MISSING);
				tooth_cnt = 0;
			} 
			// else stay unsynced tooth_cnt kept constant at 254 - TOOTH_WHEEL_LISTEN
			// note that we do update etime.average_tooth_period before return
		} else {  // assume the first pulses are noise
			--tooth_cnt;
			if (tooth_cnt > 255 - TOOTH_WHEEL_NOISE){
				invalidate_rpm_period();
				// we don't even avarage at this point.
				// (although a bogus 0xffff * 4usec  would decay fast)
				etime.average_tooth_period = period16;	// doesn't hurt to save it
				etime.tooth_cnt = tooth_cnt;
				return 0;
			} // else note the average_tooth_period averaging
		} 
	}

/* test cases naming:
Tat_t another_trigger_tooth=at trigger_tooth=t
note: RPM always applied at 0 anyway, see start of this function
T30_0 RPM: 30 	TRIG: 0,30
T30_1		TRIG: 1,31
T30_29		TRIG: 29,59
T20_0 RPM: 20,40 TRIG: 0,20,40
T20_1		TRIG: 1,21,41
T20_19		TRIG: 19,39,59
T15_0 RPM: 15,30,45 TRIG: 0,15,30,45
T15_1		TRIG: 1,16,31,46
T15_14		TRIG: 14,29,44,59
T9_0 RPM: 9,18,27 TRIG: 0,9,18,27
T9_1		TRIG: 1,10,19,28
T9_8		TRIG: 8,17,26,35
 ... is another_trigger_tooth=0 ever applied ?
T0_0 RPM: -  
T0_1
T0_11
*/
	etime.tooth_cnt = tooth_cnt;

	// the trigger-tooth can be adjusted in config
	c_trigger_tooth = config.trigger_tooth;
	c_another_trigger_tooth = config.another_trigger_tooth;
	if (c_another_trigger_tooth){
		while( tooth_cnt > c_another_trigger_tooth){
			tooth_cnt -= c_another_trigger_tooth;
		}
	// take care, tooth_cnt destroyed now, subtracted N * c_another_trigger_tooth
	// tooth_cnt range:  0 <= tooth_cnt <= c_another_trigger_tooth INCLUSIVE range !!!
	}

	if( tooth_cnt == 0 || tooth_cnt == c_another_trigger_tooth ) {
		flag_rpm_calc( capturedtime );
	}

	// see "tooth_cnt range" note above
	if( ( tooth_cnt == c_trigger_tooth || (c_trigger_tooth == 0 && tooth_cnt == c_another_trigger_tooth))
		&& isvalid_rpm_period() ) {
		trigger |= _BV(TW_TRIG);
		// updating engine.rpm_period and "start a rpm calculation" happens elsewhere
	}

	// first take the half, than add them! Overflow is more concern than 4 usec precision
	etime.average_tooth_period = etime_atp2 + (period16 >> 1);
	return trigger;
}

#endif
