/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Marcell Gal, cell at x-dsl.hu]                *
***********************************************************
*  -- History --                                          *
\*********************************************************/

   /*!
   * \addtogroup g_serial_comm Serial Communication Routines
   */
/*@{*/

/*
TODO:
The uart receive interrupt should place the byte in a circular buffer 
instead of calling the comm-function. The comm function should reside
in the main loop and poll the circular buffer.

TODO: head and tail will be fucked by interrupt unless correctly locked 
by cli()-sei() Make separate functions that has these
(or wrap the whole calling around sei-cli), since in interrupt
we don't want to waste cycles by saving SREG and reverting ...
*/

#include "global.h"
#include "comm.h"
#include "ringbuf.h"
#include "storage.h"
#include "actuators.h"
#include "ve.h"
#include "menu.h"
#include "iac.h"
#include "adc.h"
#include "lcd.h"
#include "fuelcalc.h"
#include "slcan.h"

#include <avr/io.h>
#include <inttypes.h>
#include <avr/signal.h>
#include <avr/sfr_defs.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

struct ringbuf_t comm_recbuf, comm_sndbuf;

void ringbuf_init(struct ringbuf_t *r)
{
	r->hidx = r->tidx = 0;
	r->cnt = 0;
}

/* FIFO functions named a'la perl style: push appends, shift gets first
could return true if we could insert (but what would rs232 interrupt use it??)
Must be called from interrupt
*/
__inline__ void ringbuf_push(struct ringbuf_t *r, uint8_t val)
{
	uint8_t t;
	if(r->cnt < RINGBUFLEN){ //great: we have place to push our value to
		r->buf[t = r->tidx] = val;
		
		if(++t >= RINGBUFLEN)
			t=0;
	
		r->cnt++;
		r->tidx = t;
		
		return; // success 
	}
	return;	// sorry. no more place.
}

/***********************************
 * stuff bytes if bytestuff != 0
 * bytestuff==0 allows pushing framing HDLC_END char
 * or using protocol that does not use byte stuffing
 ***********************************/
__inline__ void ringbuf_push_mightstuff(struct ringbuf_t *r, uint8_t val, uint8_t bytestuff)
{
	if( bytestuff == 0){
		ringbuf_push( r, val );
		return;
	}

	if( val == HDLC_END ){
		ringbuf_push( r, HDLC_ESC );
		ringbuf_push( r, HDLC_ESCAPE(HDLC_END) );
		return;
	}
	if( val == HDLC_ESC ){
		ringbuf_push( r, HDLC_ESC );
		ringbuf_push( r, HDLC_ESCAPE(HDLC_ESC) );
		return;
	}
}

/*
 * push max n bytes starting from buf[0] to ringbuf *r
 * returns how many bytes actually pushed
 * TODO: test ringbuf_push_n(), apparently
 * it crashes rs232 communication
 */
__inline__ uint8_t ringbuf_push_n(struct ringbuf_t *r, uint8_t *sbuf, uint8_t n)
{
	uint8_t t, *r_buf, r_cnt, n_orig;
	
	n_orig = n;
	t = r->tidx;
	r_cnt = r->cnt;	// don't worry, we are in a big cli()-sei() by caller
	r_buf = r->buf;
	while(r_cnt < RINGBUFLEN && n){ 
    //great: we have place to push our value to and bytes to push
		r_buf[t++] = *sbuf;	// yes, copy it please
		sbuf++;
		if(t >= RINGBUFLEN) t = 0;	// at the end of ringbuffer?
		r_cnt++;
		n--;
	}
	r->tidx = t;
	r->cnt = r_cnt;
	return n_orig - n;	// return how many were pushed
}


/*
Only call from userspace.
*/
void ringbuf_shift_proc(struct menu_t *m, struct ringbuf_t *r)
{
	uint8_t t;

	cli();	// if we put it below, gcc will not reload r->cnt so it could get corrupted
	if((r->cnt != 0)){ // something is in there
	r->cnt--;	// only the cnt could be tampered from irq....
	sei();
	
	menu_char(m, r->buf[t = r->hidx]);
	
	if(++t >= RINGBUFLEN)
		t=0;
			
	r->hidx = t;
	
	}
	sei();
	return ; // success
}

// return a 16 bit word. The upper byte is used for signalling buffer empty
// Called from TX interrupt
uint16_t ringbuf_shift(struct ringbuf_t *r)
{
	uint8_t t;
	uint16_t val;

	if (r->cnt != 0) { // something is in there
		r->cnt--;	// only the cnt could be tampered from irq....
	
		val = r->buf[t = r->hidx];
		
		if(++t >= RINGBUFLEN)
			t=0;
			
		r->hidx = t;

		return val;
	} else
		return 0xff00;  // msb is non-zero == buffer is empty
}


#define is_byte_stuffing (menu->state == MENU_can)

void ringbuf_fill(struct ringbuf_t *sndbuf, struct menu_t *menu)
{
/* 
process recbuf, call comm() (that puts cmds to cmdqueue) 
until recbuf empty or cmdqueue full
*/
	uint8_t cnt;
	struct cmd_queue_t *cmdqh;
	cmdqh = &cmd_queue[cmd_queue_head];
	/* process cmdqueue until empty or sndbuf full 
	* TODO: retest this with and without SLCAN
	*/

	// when byte stuffing is active, we might need to
	// push 2 characters to the buffer
	while((cmd_queue_head != cmd_queue_tail) 
		&& (sndbuf->cnt < 
		// ( is_byte_stuffing ? RINGBUFLEN-1 : RINGBUFLEN )
		// simple and safe:
		(RINGBUFLEN-1)
		)) {
		uint8_t (*f)(uint8_t);
		f = cmdqh->func;
		// cmdqh->count is NOT decreased by the TX interrupt
		// (only sndbuf->cnt is decreased - important that never increased!)
		cnt = cmdqh->count;
		cli();
		if(f != NULL){	// traditional comm function
			// at cnt==0 HDLC_END needs no escaping
			ringbuf_push_mightstuff(sndbuf, f(cnt), is_byte_stuffing ? cnt : 0);	// send only one byte
			cnt = cmdqh->count;	// the tricky TableDump ConfigDump functions screw this
			cnt++;
		} else { // NULL pointer is special: we got bytes pointed to by private to copy
			// test ringbuf_push_n() before putting to service
//			sent = ringbuf_push_n(sndbuf, cmdqh->private + cnt, cmdqh->len - cnt);
//			cnt += sent;
			ringbuf_push_mightstuff(sndbuf, srl_generic(cnt), is_byte_stuffing ? cnt : 0);
			cnt ++;
		}	
		UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
		sei();

		if ( cnt < cmdqh->len) {	// not ready yet, save the updated .count
			cmdqh->count = cnt;
		} else {	// drop the cmd_queue entry
			if(f == NULL) *cmdqh->private = 0;	// ruin the first byte (signalling back)
			cmd_queue_head = (cmd_queue_head + 1) % CMD_QUEUE_SIZE;
			cmdqh = &cmd_queue[cmd_queue_head];
		}
	} // while
}

/*@}*/

