/* Copyright 2005 (C) by Michael Richards. All rights reserved.  */

#include <avr/interrupt.h>
#include <inttypes.h>

#include "global.h"
#include "ve.h"
#include "comm.h"
#include "storage.h"
#include "ego.h"
#ifdef WBO2
#include "wbo2.h"
#endif
#include "ioinit.c"

#include "firmware_revision.h"
#include "sipr.h"

/* global for the current incoming packet */
struct sipr_t sipr_incoming;

/* SIPR 100ms timeout counter for receiving a packet 255 = none incoming */
uint8_t sipr_inc_timeout;
uint8_t sipr_state;

/* this variable is used to load word values that must be sent. This
 * will allow the other byte to be cached so we don't have sync problems
 * while we're waiting for the next byte of the packet to be sent */
uint8_t sipr_atomic;

uint8_t sipr_packet_ACK(uint8_t i);
uint8_t sipr_packet_GET_VERSION(uint8_t i);
uint8_t sipr_packet_GET_TABLESIZE_VE(uint8_t i);
uint8_t sipr_packet_GET_TABLESIZE_ZERO(uint8_t i);
uint8_t sipr_packet_GET_HWSIG(uint8_t i);
uint8_t sipr_packet_GET_TABLESIZE_RPM(uint8_t i);
uint8_t sipr_packet_GET_TABLESIZE_KPA(uint8_t i);
uint8_t sipr_packet_GET_TABLEDATA_RPM(uint8_t i);
uint8_t sipr_packet_GET_TABLEDATA_KPA(uint8_t i);
uint8_t sipr_packet_GET_TABLEDATA_IGN(uint8_t i);
uint8_t sipr_packet_GET_TABLEDATA_VE(uint8_t i);
uint8_t sipr_packet_GET_TABLEDATA_LA(uint8_t i);
uint8_t sipr_packet_GET_TABLEDATA_EGOC(uint8_t i);
uint8_t sipr_packet_GET_TABLEDATA_CONFIG(uint8_t i);
uint8_t sipr_packet_READ_SENSORS_P1(uint8_t i);

void sipr_process_packet(void);

/* init function - sets up the timeout and the incoming/outgoing packet data
 */
void sipr_init() {
  /* no packets are incoming */
  sipr_inc_timeout=255;

  /* reset the character count */
  sipr_incoming.idx=0;

  /* reset the sequence number */
  sipr_incoming.seq=0;
}


/* timeout flag. sends a NAK packet */
void sipr_timeout() {
  pushfuncsipr(&sipr_packet_NAK, 6, sipr_incoming.seq);
  sipr_inc_timeout=255;
}


/* process a completed packet from sipr_incoming */
void sipr_process_packet() {
  switch (sipr_incoming.type) {
    case SIPR_PAK_ACK: /* got an ACK return an ACK */
      pushfuncsipr(&sipr_packet_ACK, 6, sipr_incoming.seq);
      break;
    case SIPR_PAK_GET_VERSION: /* return the firmware version */
      pushfuncsipr(&sipr_packet_GET_VERSION, 9, sipr_incoming.seq);
      break;
    case SIPR_PAK_SET_BAUD: /* return the ack packet and schedule a baudrate change */
      pushfuncsipr(&sipr_packet_ACK, 6, sipr_incoming.seq);
      /* set the baudrate change flag so when the ACK is sent the baud will change */
      switch (sipr_incoming.data[0]) {
        case SIPR_BAUD_9600:
          new_baudrate=103;
          break;
        case SIPR_BAUD_19200:
          new_baudrate=51;
          break;
        case SIPR_BAUD_38400:
          new_baudrate=25;
          break;
        case SIPR_BAUD_57600:
          new_baudrate=16;
          break;
        case SIPR_BAUD_115200:
          new_baudrate=8;
          break;
        default:
          new_baudrate=103;
          break;
      }
      break;
    case SIPR_PAK_GET_TABLESIZE: /* return the size of the map tables */
      /* invalid table ID means we send a zero size */
      switch (sipr_incoming.data[0]) {
        case SIPR_TABLE_RPM:
          pushfuncsipr(&sipr_packet_GET_TABLESIZE_RPM, 8, sipr_incoming.seq);
          break;
        case SIPR_TABLE_KPA:
          pushfuncsipr(&sipr_packet_GET_TABLESIZE_KPA, 8, sipr_incoming.seq);
          break;
        case SIPR_TABLE_IGN:
        case SIPR_TABLE_VE:
        case SIPR_TABLE_LA:
          pushfuncsipr(&sipr_packet_GET_TABLESIZE_VE, 8, sipr_incoming.seq);
          break;
        default: 
          pushfuncsipr(&sipr_packet_GET_TABLESIZE_ZERO, 8, sipr_incoming.seq);
          break;
      }
      break;
    case SIPR_PAK_GET_TABLEDATA: /* return the size of the map tables */
      switch (sipr_incoming.data[0]) {
        case SIPR_TABLE_RPM:
          pushfuncsipr(&sipr_packet_GET_TABLEDATA_RPM, 6+VE_SIZE_RPM, sipr_incoming.seq);
          break;
        case SIPR_TABLE_KPA:
          pushfuncsipr(&sipr_packet_GET_TABLEDATA_KPA, 6+VE_SIZE_RPM, sipr_incoming.seq);
          break;
        case SIPR_TABLE_IGN:
          pushfuncsipr(&sipr_packet_GET_TABLEDATA_IGN, 6+VE_SIZE_RPM*VE_SIZE_KPA, sipr_incoming.seq);
          break;
        case SIPR_TABLE_VE:
          pushfuncsipr(&sipr_packet_GET_TABLEDATA_VE, 6+VE_SIZE_RPM*VE_SIZE_KPA, sipr_incoming.seq);
          break;
        case SIPR_TABLE_LA:
          pushfuncsipr(&sipr_packet_GET_TABLEDATA_LA, 6+VE_SIZE_RPM*VE_SIZE_KPA, sipr_incoming.seq);
          break;
        case SIPR_TABLE_EGOC:
          pushfuncsipr(&sipr_packet_GET_TABLEDATA_EGOC, 6+VE_SIZE_RPM*VE_SIZE_KPA, sipr_incoming.seq);
          break;
        case SIPR_TABLE_CONFIG:
          pushfuncsipr(&sipr_packet_GET_TABLEDATA_CONFIG, 22, sipr_incoming.seq);
          break;
      }
      break;
    case SIPR_PAK_SET_TABLEDATA: /* updates an element within a table */
      switch (sipr_incoming.data[0]) {
        case SIPR_TABLE_RPM:
          /* check the bounds of the offset */
          if (sipr_incoming.data[1]<VE_SIZE_RPM) {
            /* update the table in SRAM */
            ve.rpm_range[sipr_incoming.data[1]]=sipr_incoming.data[2];            
            /* update the EEPROM table entry */
          }
          break;
        case SIPR_TABLE_KPA:
          /* check the bounds of the offset */
          if (sipr_incoming.data[1]<VE_SIZE_KPA) {
            /* update the table in SRAM */
            ve.kpa_range[sipr_incoming.data[1]]=sipr_incoming.data[2];
            /* update the EEPROM table entry */
          }
          break;
        case SIPR_TABLE_IGN:
          /* check the bounds of the offset */
          if (sipr_incoming.data[1]<VE_SIZE_RPM*VE_SIZE_KPA) {
            /* update the table in SRAM */
            ign.adv_table[sipr_incoming.data[1]]=sipr_incoming.data[2];
            /* update the EEPROM table entry */
          }    

          break;
        case SIPR_TABLE_VE:
          /* check the bounds of the offset */
          if (sipr_incoming.data[1]<VE_SIZE_RPM) {
            /* update the table in SRAM */
            ve.ref_table[sipr_incoming.data[1]]=sipr_incoming.data[2];
            /* update the EEPROM table entry */
          }    

          break;
        case SIPR_TABLE_LA:
          /* check the bounds of the offset */
          if (sipr_incoming.data[1]<VE_SIZE_RPM) {
            /* update the table in SRAM */
            ve.lambda[sipr_incoming.data[1]]=sipr_incoming.data[2];
            /* update the EEPROM table entry */
          }    

          break;
        case SIPR_TABLE_EGOC:
          /* check the bounds of the offset */
          if (sipr_incoming.data[1]<VE_SIZE_RPM) {
            /* update the table in SRAM */
            /*ve.rpm_range[sipr_incoming.data[1]]=sipr_incoming.data[2]; */
            /* update the EEPROM table entry */
          }    

          break;
        case SIPR_TABLE_CONFIG:
          ((uint8_t *)&config)[sipr_incoming.data[1]]=sipr_incoming.data[2];
          break;
      }
      pushfuncsipr(&sipr_packet_ACK, 6, sipr_incoming.seq); /* acknowledge that it's been done */
      break;
    case SIPR_PAK_READ_SENSORS: /* return the sensor data for page1 */
        pushfuncsipr(&sipr_packet_READ_SENSORS_P1, 27+6, sipr_incoming.seq);
      break;
    case SIPR_PAK_GET_HWSIG:
        pushfuncsipr(&sipr_packet_GET_HWSIG, 12+6, sipr_incoming.seq);
      break;
    case SIPR_PAK_ENTER_FLASH:
      if (!(engine.status & _BV(running))) {
        cli();
        SPCR = 0x36; /* we take care of io and uart */
        ioinit();
        UCSR0A = 0;
        UCSR0B = 0x18; /*d8; */
        UCSR0C = 0x06;
        /* jump to the last address in progmem */
        asm("jmp 0x1fffc"); /* bye bye cruel world I'm leaving */
      }
      pushfuncsipr(&sipr_packet_NAK, 6, sipr_incoming.seq);
      break;
    case SIPR_PAK_SAVE_CONFIG:
      save_variables(0);
      pushfuncsipr(&sipr_packet_ACK, 6, sipr_incoming.seq);
      break;
    default: /* unrecognised packet type send a NAK */
      pushfuncsipr(&sipr_packet_NAK, 6, sipr_incoming.seq);
      break;
  }
}


/* NAK packet function */
uint8_t sipr_packet_NAK(uint8_t i) {
  /* packet type */
  return SIPR_PAK_NAK;
}


/* ACK packet function */
uint8_t sipr_packet_ACK(uint8_t i) {
  /* packet type */
  return SIPR_PAK_ACK;
}


/* GET_VERSION packet function */
uint8_t sipr_packet_GET_VERSION(uint8_t i) {
  switch (i) {
    case 0: /* packet type */
      return SIPR_PAK_GET_VERSION;
    case 1:
      return FIRMWARE_VERSION_MAJOR;
    case 2: 
      return FIRMWARE_VERSION_MINOR;
  }
  /* case 3 */
  return FIRMWARE_VERSION_BUILD;
}


/* GET_TABLESIZE packet function for all the VE tables */
uint8_t sipr_packet_GET_TABLESIZE_VE(uint8_t i) {
  switch (i) {
    case 0: /* packet type */
      return SIPR_PAK_GET_TABLESIZE;
    case 1:
      return VE_SIZE_KPA;
  }       
  /* case 2 */
  return VE_SIZE_RPM;
}

/* GET_TABLESIZE packet function for all the VE table widths */
uint8_t sipr_packet_GET_TABLESIZE_RPM(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_TABLESIZE;
  if (i==1) return 1; /* 1 row */
  return VE_SIZE_RPM;
}

/* GET_TABLESIZE packet function for all the VE table heights */
uint8_t sipr_packet_GET_TABLESIZE_KPA(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_TABLESIZE;
  if (i==2) return 1; /* 1 column */
  return VE_SIZE_KPA;
}



/* GET_TABLESIZE packet function for unknown tables - returns 0's */
uint8_t sipr_packet_GET_TABLESIZE_ZERO(uint8_t i) {
  switch (i) {
    case 0: /* packet type */
      return SIPR_PAK_GET_TABLESIZE;
  }
  /* return 0's for all the sizing data */
  return 0;
}


/* GET_HWSIG packet function to return the hardware signature if valid */
uint8_t sipr_packet_GET_HWSIG(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_HWSIG;
  /* make sure there is a valid signature byte - if not return 0's */
  if (pgm_read_byte_far(0x0001fff0)!=0xAA) 
    return 0;
  /* don't allow overflow */
  if (i>12)
    return 0;
  return pgm_read_byte_far(0x0001fff0+i-1);
}


/* GET_TABLEDATA packet function for returning the rpm table */
uint8_t sipr_packet_GET_TABLEDATA_RPM(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_TABLEDATA;
  /* return the data for the ignition table */
  return ve.rpm_range[i-1];
}


/* GET_TABLEDATA packet function for returning the map table */
uint8_t sipr_packet_GET_TABLEDATA_KPA(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_TABLEDATA;
  /* return the data for the ignition table */
  return ve.kpa_range[i-1];
}


/* GET_TABLEDATA packet function for returning the ignition table */
uint8_t sipr_packet_GET_TABLEDATA_IGN(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_TABLEDATA;
  /* return the data for the ignition table */
  return ign.adv_table[i-1];
}


/* GET_TABLEDATA packet function for returning the VE table */
uint8_t sipr_packet_GET_TABLEDATA_VE(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_TABLEDATA;
  /* return the data for the ignition table */
  return ve.ref_table[i-1];
}


/* GET_TABLEDATA packet function for returning the Lambda table */
uint8_t sipr_packet_GET_TABLEDATA_LA(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_TABLEDATA;
  /* return the data for the ignition table */
  return ve.lambda[i-1];
}


/* GET_TABLEDATA packet function for returning the ignition table */
uint8_t sipr_packet_GET_TABLEDATA_EGOC(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_TABLEDATA;
  /* return the data for the ego correction table */
/* XXX not implemented yet XXX */
  return 0;
}

/* GET_TABLEDATA packet function for returning the test table */
uint8_t sipr_packet_GET_TABLEDATA_CONFIG(uint8_t i) {
  if (i==0) return SIPR_PAK_GET_TABLEDATA;
  return ((uint8_t *)&config)[i-1];
}


/* read the sensor data */
uint8_t sipr_packet_READ_SENSORS_P1(uint8_t i) {
  uint16_t atomic;
  switch (i) {
    case 0: return SIPR_PAK_READ_SENSORS;
    case 1: return 0; // 0 packets remaining
    case 2: return 0; // 0 delay
    case 3: return 0; // page0 format
    case 4:
      atomic=engine.rpm_vhr;
      sipr_atomic=(atomic >> 8);
      return atomic & 0x00FF;
    case 5: return sipr_atomic;
    case 6:
      atomic=engine.kpa250;
      sipr_atomic=(atomic >> 8);
      return atomic & 0x00FF;
    case 7: return sipr_atomic;
    case 8:
      atomic=0; // injector pulsewidth
      sipr_atomic=(atomic >> 8);
      return atomic & 0x00FF;  
    case 9: return sipr_atomic;
    case 10:
      #ifdef WBO2
        atomic=wbo2.current_afr; // AFR1
      #else
        atomic=0; // AFR1
      #endif
      sipr_atomic=(atomic >> 8);
      return atomic & 0x00FF;
    case 11: return sipr_atomic;
    case 12:
      atomic=0; // AFR2
      sipr_atomic=(atomic >> 8);
      return atomic & 0x00FF;
    case 13: return sipr_atomic;
    case 14: 
      atomic=calc_target_afr(); // target AFR
      sipr_atomic=(atomic >> 8);
      return atomic & 0x00FF;
    case 15: return sipr_atomic;
    case 16: return engine.tps;
    case 17: return engine.coolant;
    case 18: return engine.mat;
    case 19: return engine.batt;
    case 20: return engine.ignadv;
    case 21:
      atomic=engine.egt;
      sipr_atomic=(atomic >> 8);
      return atomic & 0x00FF;
    case 22: return sipr_atomic;
    case 23:
      atomic=0;  // XXX needs EGT value for channel 2
      sipr_atomic=(atomic >> 8);
      return atomic & 0x00FF;
    case 24: return sipr_atomic;
    case 25:
      #ifdef BOOSTCONTROL
        return engine.boostmax_target;
      #else
        return 0;
      #endif
    case 26: return 0; // boost duty cycle

  }
  return 0;
}

