/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Marcell Gal cell at x-dsl.hu]                *
***********************************************************

there are PWM functions which need some care from time to time, however
it is a waste to place them into expensive interrupts like
handler2k or whatever...

LCD backlight and IDLEPWM_SOFT comes to mind, heater, etc...

This should be called from mainloop, and it will keep track of the 
time fractions
*/
   /*!
   * \addtogroup g_pwm Pulse Width Modulation Routines
   */
/*@{*/

#ifndef __SOFTPWM_C__
#define __SOFTPWM_C__

#include "eventqueue.h"
#include "softpwm.h"
#include "benchmark.h"

#include <avr/interrupt.h>
#include <inttypes.h>

#include "arithm.c"

// also include for genboard v2: (DHIGH constant is used)
#include "lbus.h"

// 85 will give large enough window for all(0/1/err) cases
// unit is [256 * 4 usec]
#ifndef SLIGHTLY_TRESHOLD
#define SLIGHTLY_TRESHOLD (85)
#endif

#define full2on(n,ratio) (mult16_8(n,ratio)>>8)

/* returns the time atomically (where should this be??) */
static inline uint16_t whats_the_time(void)
{
	uint16_t n;	// uint8_t my_sreg = SREG;
	cli();
	readTMR();
	n = etime.now;
	sei();	// SREG = my_sreg;
	return n;
}

/* receives the "percentage" 0..255 where 255 means 100 percent full-on
*/
static __inline__ void softpwm_init(struct softpwm_t *s, uint8_t onr)
{
	uint16_t n;
	s->on_ratio = onr;
	s->designated_ratio = onr;
	n = whats_the_time(); 
	s->ontime = full2on(n, onr);
	s->lasttime = n;
};

/* Call this whenever the on_ratio changes */
static __inline__ void softpwm_set(struct softpwm_t *s, uint8_t newval)
{
	s->designated_ratio = newval;
}

// changes on_time as if the 0..n happened with dr instead of onr (ratio)
// saves n to s->lasttime
// returns updated s->ontime
static inline uint16_t adjust_on_time(struct softpwm_t *s, uint8_t dr, uint8_t onr, uint16_t n, uint16_t lt)
{
	uint16_t ont;
	ont = s->ontime;
	if(n < 	lt){ // overflow of etime.now (smaller than  it was lasttime)
		// correction of s->ontime needed: -= 0x10000 * s->on_ratio/256
		ont -= onr << 8;	
	}

	if(dr != onr){	// On_ratio designated to be changed
		if(dr > onr) // unsigned
			ont += full2on(n, dr - onr);
		else
			ont -= full2on(n, onr - dr);

		s->on_ratio = dr;
	}

	s->ontime = ont;
	s->lasttime = n;
	return ont;
}

// checks if the emitted charge (or power) is higher or smaller than what would be desired
// return true (DHIGH constant) if output should be activated
static inline uint8_t compare_mean(struct softpwm_t *s, uint16_t ont, uint8_t onr, uint16_t n)
{
	uint8_t di;
	// the desired ontime (this is the key of the softPWM !!)
	ont -= full2on(n, onr);
	di = ont>>8;

	// if di is slightly positive, we need to turn the signal off.
	if(di < SLIGHTLY_TRESHOLD){
		return 0;
	}

	// if di is slightly negative, we need to turn the signal on.
    if(di > 0x100-SLIGHTLY_TRESHOLD){
		return DHIGH;
	}
#ifdef BENCHMARK
	benchstats.softpwm_err++;
#endif
	// if di is in the illegal window, some error occured, such as not been called 
	// for too much time (should not happen)
	// or softpwm_act_var(): multiplier < (on_ratio<<8) continuosly =>
	// means unreachable target => therefore we return 1;
	// And reset s->ontime anyway: lose the remainder, but what else could we do?
	// etime.now overflow is handled explicitely in adjust_on_time().
	s->ontime = full2on(n, onr);
	return DHIGH;
}

/* 
 Returns true if the device should be turned on 
 Must be called more frequent than 80 msec
 How it works:
 s->on_time is kept at appr. (etime.now * s->on_ratio/256)
 a little deviance in any direction will be corrected. Very smart, overflow
 should not cause problem, the whole game is played modulo 65536 * 4 usec
 imagine it as a "ring", an analog wall-clock! The etime.now hand of it
 goes around faster than it's on_time hand.

 Should we return (in a bit) if change happened since last time?
*/
static __inline__ uint8_t softpwm_act(struct softpwm_t *s)
{
	uint16_t n, lt, diff, ont;
	uint8_t dr, onr;

	n = whats_the_time();
	lt = s->lasttime;
	dr = s->designated_ratio;
	onr = s->on_ratio;
	ont = adjust_on_time(s, dr, onr, n, lt);
	if(s->state) {	// it was turned on
		diff = n - lt;
		ont += diff;
		s->ontime = ont;
	}

	return (s->state = compare_mean(s, ont, dr, n));
}

// #define full2on_mult(n,m,r) (uint16_t)(mult16_16(mult16_8(n,r)>>8, m)>>8)
/* 
Similar to above.
returns true if the device should be turned on.
multiplier is the instantenous measured (or assumed) voltage on the 
switched device.
It tries to keep (time_spent_on * mult /65536 == n * on_ratio)
With multiplier == 0xffff, it will be almost the same (except slower) as softpwm_act().
With smaller multiplier, it tries to spend more time in the on state:
on_ratio/mult * 65536 [/256 of all time]

If you want to keep output power (like with wbo2 heater,
unlike voltage for the wbo2 Ipump), you
specify multiplier=voltage * voltage, scaled appropriately.
If you want to keep output mean voltage, you specify first power
of voltage (scaled appropriately).
*/
static inline uint8_t softpwm_act_var(struct softpwm_t *s, uint16_t multiplier)
{	
	uint16_t n, lt, diff, ont;
	uint8_t dr, onr;

	n = whats_the_time();
	lt = s->lasttime;
	dr = s->designated_ratio;
	onr = s->on_ratio;
	ont = adjust_on_time(s, dr, onr, n, lt);
	if(s->state) {	// it was turned on
		diff = n - lt;
		ont += mult16_16(diff, multiplier)>>16; // the only difference from softpwm_act()
		s->ontime = ont;
	}

	return (s->state = compare_mean(s, ont, dr, n));
}

/* returns true if interval * 4 usec passed since last true..
*/
static __inline__ uint8_t softelapsed(struct elapsed_t *e, uint16_t interval)
{
	uint16_t d, now;
	// only used for the 100ms handler where 1msec error due to nonatomic read of now 
	// would be acceptable, but a forward-1msec error at the time of
	// returning 1 and a correct read very shortly afterwards
	// could trigger a variable reset (e->lastrunat = now;) ...
	
	now = whats_the_time();
	if((d = now - e->lastrunat) < interval) 
		return 0;	// not elapsed
		
	if(d > (interval << 1)) {	// should never happen, except at beginning
		e->lastrunat = now;
		return 1;
	} 

	e->lastrunat += interval;
	return 1;
}


static __inline__ uint16_t now16_1024usec( void )
{
	uint16_t t;
	t = etime.now_upper << 8;
	t |= etime.now >> 8;
	return t;
}

/* returns true if interval * 1024 usec passed since last true..
Does not update now (not even access it atomically), just take it as approximate value...
*/
static __inline__ uint8_t softelapsedlong(struct elapsed_t *e, uint16_t interval)
{
	uint16_t d, now;
	// only used for the 100ms handler where 1msec error due to nonatomic read of now 
	// would be acceptable, but a forward-1msec error at the time of
	// returning 1 and a correct read very shortly afterwards
	// could trigger a variable reset (e->lastrunat = now;) ...
	
	now = now16_1024usec();
	if((d = now - e->lastrunat) < interval) 
		return 0;	// not elapsed
		
	if(d > (interval << 1)) {	// should never happen, except at beginning
		e->lastrunat = now;
		return 1;
	} 

	e->lastrunat += interval;
	return 1;
}

/* similar to above, but a little rougher. Accesses now atomically, since just upper byte */
static __inline__ uint8_t softelapsed8(struct elapsed8_t *e, uint8_t interval)
{
	uint8_t d, now;
	
	now = etime.now >> 8;	// no need for atomic read, 1 byte only!
	
	d = now - e->lastrunat;
	if(d < interval) 
		return 0;	// not elapsed

	if(d > (interval << 1)) {	// should never happen, except at beginning
		e->lastrunat = now;
		return 1;
	} 

	e->lastrunat += interval;
	return 1;
}
#endif

/*@}*/

