   /*!
   * \addtogroup g_knock Knock Sensor Routines
   */
/*@{*/

// low-level spi functions

#ifdef GENBOARDv3

#include <inttypes.h>
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/sfr_defs.h>
#include <avr/interrupt.h>

#include "arithm.c"

#include "global.h"
#include "spi.h"
#include "ringbuf.h"
#include "benchmark.h"
#include "ve.h"
#include "knock.h"
#include "eventqueue.h"

/*
TODO:
the current code is almost finished!

There are no need for 32 bytes buffers (ringbuf)
 - how can they be made smaller without braking rs232 etc.?

Using a common, 8 entry, uint16_t ringbuf should do it.
One byte would be the command to send, the other select the channel (knock, ADC, ...)
This ringbuf could be filled (optionally from userspace) the same way as RS232's.
We can allow it to contain overlapping commands.
*/


/*
  knock example:
    set_knock_channel(KNOCK_CHANNEL_1);
    read_knock();
    while(knock_data[KNOCK_CHANNEL_1] == 0xffff)
      ;
    val = knock_data[KNOCK_CHANNEL_1];

  adc example (channel i):
    read_mcp3208(i);
    while(mcp3208_data[i] == 0xffff)
      ;
    val = mcp3208_data[i];
*/

// this array is not really useful this way.
// there should be a filtered set (8 elements), 
// and a communication set (1..2 elements).
// the communication set needn't have 8, rather just 1..2 channel,
// since the data is to be stored in the filtered place before the
// communication (SPI-ADC interrupt) could ruin it.
uint16_t mcp3208_data[MCP3208_CHANNELS]; // I AM OUT OF GOOD NAMES, HELP PLEASE!

uint8_t knock_channel;  // don't modify this yourself, use set_knock_channel()

struct spi_t spi;
struct ringbuf_t spi_knock, spi_adc;

void init_spi(void) {

	// SPI:
	// clock 2MHz, low clock idle, clock phase
	// TODO: if spi just doesn't work, then the clock phase/polarity is probably wrong!
	SPCR = _BV(SPE) | _BV(MSTR) | _BV(CPHA) | _BV(SPR0);
	SPSR = _BV(SPI2X);

	spi.cmd = NO_CMD;
	spi.cnt = 0;
	
	ringbuf_init(&spi_knock);
	ringbuf_init(&spi_adc);

}

void init_spi_mcp3208(void) {
	uint8_t i;

	CS_MCP_PORT |= _BV(CS_MCP);

	for (i=0; i<MCP3208_CHANNELS; i++) // just a matter of time before someone changes this to do-while :)
		mcp3208_data[i] = 0xffff;
	
}


void init_spi_knock(void) {
	uint8_t t;
	
	CS_KNOCK_PORT |= _BV(CS_KNOCK);
	CS_KNOCK_INTEGRATE_PORT &= ~_BV(KNOCK_INTEGRATE);

	knock.data = 0xffff;
	knock.channel = 0xff;
	knock.status = 0;

	t = next_knock_channel();
	set_knock_channel(t);
	// do twice to ensure the KNOCK_NOT_RESPONDING flag is initialized properly
	set_knock_channel(t);
}

void read_mcp3208(uint8_t chan) {
	uint8_t i;

	mcp3208_data[chan] = 0xffff;  // ffff == data not available

	// push data to queue
	i = SREG;
	cli();
	ringbuf_push(&spi_adc, (chan<<4) | ADC_READ);
	SREG = i;

	// start spi
	next_spi_cmd();
}

/*
  Setup knock channel before the integration is started!!!
*/
void set_knock_channel(uint8_t chan) {
	uint8_t i;

	// insert ((chan<<4) | KNOCK_SETUP) command in knock queue
	i = SREG;
	cli();
	ringbuf_push(&spi_knock, (chan<<4) | KNOCK_SETUP);
	SREG = i;
	
	// start spi
	next_spi_cmd();
}

/*
  read out the result of the integration
*/
void read_knock(void) {
	uint8_t i;

	knock.data = 0xffff;

	// insert command in knock queue
	i = SREG;
	cli();
	ringbuf_push(&spi_knock, KNOCK_READ);
	SREG = i;

	// start spi
	next_spi_cmd();
}

uint8_t spi_knock_busy(void) {
	// either there are knock-commands in the queue
	if (spi_knock.cnt) 
		return 1;
	
	// or a single knock-command is currently being processed
	if ((spi.cmd == KNOCK_SETUP) || (spi.cmd == KNOCK_READ)) {
		return spi.cnt; //if non-zero, then the command has not completed
	}

	return 0;
}

/*	
send three bytes to mcp3208.
we want MSB first, so the first byte is padded with 0's

  to be send:
    byte1: 0000011 chan[2]
    byte2: chan[1] chan[0] xxxxxx
    byte3: xxxxxxxx
*/
uint8_t adc_next_byte(uint8_t d) {
	uint8_t chan, spi_cnt;

	spi_cnt = spi.cnt;
	// gcc uses 16 bit comparisons for switch-case :-(
	if(spi_cnt <= 2){
		if (spi_cnt <= 1) {	
			// case 0: shouldn't happen
			// case 1:
			spi.tmp_data = d & 0x0f;	// pickup first result byte
			return 0; // adc interprets a leading 1 as start bit, sending 0's should be safe
		}
		// case 2: // continue to sets up channel number
		// spi.cmd bits 6:4 has the channel. bits 5:4 go to 7:6
		chan = spi.cmd; // & 0xF0; // we don't zero don't care bits
		return chan<<2;
	}  

	if(spi_cnt == 3){
		// case 3: // first command to send
		// this command sets up single end mode and channel number
		// spi.cmd bits 6:4 has the channel. bit6 (channel number msb) goes to bit0
		return MCP3208_CMD_0 | ( spi.cmd & 0x40 ? 1 : 0);
	} else return 0;
}


/*
  already in advanced mode
  transmit cmd1, cmd2 and an extra cmd2

  REMEMBER: When changing knock channel, it takes effect next time int/hold is activated!
  NOTE: knock chip wants CS to be deasserted after each byte...zzzzz
*/
uint8_t knock_next_byte(uint8_t d) {
	uint8_t t;
	CS_KNOCK_PORT |= _BV(CS_KNOCK);

	switch (spi.cnt) {
	case 3:  // first command to send
		// setup prescaler
		t = KNOCK_SET_PRESCALER | (KNOCK_OSC_FREQ << 1);
		break;
	case 2:
		// setup channel
		t = KNOCK_SET_CHANNEL | knock_channel;
		break;
	case 1:
		// *dummy command*
		// setup channel again - this is not a copy/paste error!
		// data are delayed by one shift cycle
		spi.tmp_data = d;  // grab first response byte
		t = KNOCK_SET_CHANNEL | knock_channel;
		break;
	default:
		// can't happen
		t = 0xff; // is this a safe value?
		break;
	}

	CS_KNOCK_PORT &= ~_BV(CS_KNOCK);
	return t;
}

/*
  TODO:
  each knock channel should have a seperate set of parameters
*/
uint8_t knock_setup_next_byte(uint8_t d) {
	uint8_t t, chan;
	CS_KNOCK_PORT |= _BV(CS_KNOCK);

	switch (spi.cnt) {
	case 6:  // first command to send
		// enter advanced mode
		t = KNOCK_SET_ADV_MODE;
		break;
	case 5:
		// check if knock IC responded to the previous command
#if 0
// if the avr resets, this may set the knock_not_responding flag even though
// the knock ic responds (it is already in advanced mode)

		if (!(knock.status & _BV(KNOCK_IN_ADV_MODE))) {
			if (d != KNOCK_SET_ADV_MODE) {
				// no response from knock :-(
				// if no response here, we should abort the entire initialization,
				// and never use knock 
				knock.status |= _BV(KNOCK_NOT_RESPONDING);
			}
		}
#endif
		// setup prescaler
		t = KNOCK_SET_PRESCALER | (KNOCK_OSC_FREQ << 1);
		break;
	case 4:
		// data delayed in advanced mode
		if (knock.status & _BV(KNOCK_IN_ADV_MODE)) {
			if (d != (uint8_t)~KNOCK_SET_ADV_MODE) {
				// ooops, no response from knock, this is not good
				// increment error counter
				knock.status |= _BV(KNOCK_NOT_RESPONDING);
			}
		} else {
			// in order not to unintentionally set the flag, we must wait until here:
			knock.status |= _BV(KNOCK_IN_ADV_MODE);
		}

		// setup channel (TODO: Can you explain this spi.cmd ?) 
		// if otherwise correct, use spi.cmd & 0xF0
		chan = (spi.cmd >> 4) ? 1 : 0;

		t = KNOCK_SET_CHANNEL | chan;
		knock_channel = chan;  // globally store which channel the knock is configured for
		break;
	case 3:
		// setup bandpass
		t = (knock_channel) ? config.knock2_frequency : config.knock1_frequency;
		t = KNOCK_SET_BANDPASS | (t & 0x3F);
		break;
	case 2:
		// setup gain
		t = (knock_channel) ? config.knock2_gain : config.knock1_gain;
		t = KNOCK_SET_GAIN | (t & 0x3F);
		break;
	case 1:
		// setup integrator
		t = (knock_channel) ? config.knock2_integrator : config.knock1_integrator;
		t = KNOCK_SET_INTEGRATOR | (t & 0x1F);
		break;
	default:
		// can't happen
		t = 0xff; // is this a safe value?
		break;
	}

	CS_KNOCK_PORT &= ~_BV(CS_KNOCK);
	return t;
}

/*
  This function decides what to do next.
  Are there any userspace-interrupt conflicts ? 
  (cli() needed ? PORTC, PORTG, spi. structure ? ) 
*/
void next_spi_cmd(void) {
	if (spi.cnt == 0) {  // check if spi is busy

		if (spi_knock.cnt) {  // anything in the queue?
			uint8_t c, cmd;
			
			c = ringbuf_shift(&spi_knock);
			cmd = c & 0x0f;

			if (cmd == KNOCK_SETUP) {
				spi.cnt = 6; // send 6 bytes
				spi.cmd = c; //KNOCK_SETUP;
				spi.next_byte = &knock_setup_next_byte;
				
				SPDR = knock_setup_next_byte(spi.cnt);
				SPCR |= _BV(SPIE);
				return;
			}

			if (cmd == KNOCK_READ) {
				spi.cnt = 3; // send 3 bytes
				spi.cmd = c; //KNOCK_READ;
				spi.next_byte = &knock_next_byte;
				
				SPDR = knock_next_byte(spi.cnt);
				SPCR |= _BV(SPIE);
				return;
			}

			// ooops, an unknown command in the knock queue... bad
			// kill the command... BOOOOM
			// command killed, go on and check the adc queue
		}

		if (spi_adc.cnt) { // anything in the queue?
			uint8_t c;
			c = ringbuf_shift(&spi_adc);

			spi.cnt = 3; // send 3 bytes
			spi.cmd = c; //ADC_READ;
			spi.next_byte = &adc_next_byte;

			CS_MCP_PORT &= ~_BV(CS_MCP);
			SPDR = adc_next_byte(spi.cnt);
			SPCR |= _BV(SPIE);
			return;
		}

		// nothing to do, disable spi interrupt
		spi.cmd = NO_CMD;
		SPCR &= ~_BV(SPIE);
	}
}


/***************************************************************************/
/*** INTERRUPT: SPI                                                      ***/
/***************************************************************************/

SIGNAL(SIG_SPI) {
	uint8_t t;

	t = SPDR;

	spi.cnt--;
	if (spi.cnt == 0) {  // means a command has finished, start next command (if any)
		// first pickup the received byte
		switch (spi.cmd & 0x0F) {
		case ADC_READ:
			CS_MCP_PORT |= _BV(CS_MCP);
			mcp3208_data[spi.cmd>>4] = ((uint16_t)spi.tmp_data<<8) | t;
			break;
		case KNOCK_READ:
			CS_KNOCK_PORT |= _BV(CS_KNOCK);
			knock.data = ((uint16_t)t<<2) | spi.tmp_data;
			knock.channel = knock_channel;
			break;
		case KNOCK_SETUP:
			CS_KNOCK_PORT |= _BV(CS_KNOCK);
			break;
		default:
			// how did we end up here...?
			break;
		}

        // start next command (or disable spi if no commands are pending)
		next_spi_cmd();

	} else {
		uint8_t (*next_byte)(uint8_t);
		next_byte = spi.next_byte;

		SPDR = next_byte(t);
	}

// TODO: what if spi has completed transmission before the interrupt 
//       is ended. Will the interrupt flag then be cleared (FATAL)?
}



#endif
/*@}*/

