/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Marcell Gal, cell at x-dsl.hu]               *
***********************************************************
*  -- History --                                          *
*  					    		                          *
\*********************************************************/


#include "fuelcontrol.h"
#include "fuelcalc.h"
#include "global.h"
#include "adc.h"
#include "comm.h"
#include "iac.h"
#include "eventqueue.h"
#include "dispatcher.h"
#include "injconf.h"
#include "ignconf.h"
#include "edis.h"
#include "benchmark.h"
#include "timing.h"

#include <avr/io.h>
#include <inttypes.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/sfr_defs.h>

extern volatile struct engine_t engine;

#include "injector.c"
#include "reenable.c"

#include "edis.c"

#include "multitooth.c"

#ifdef AUDITRIGGER
/**
 * Auditrigger related thigs are all behind AUDITRIGGER flag
 * use Manmla to enable trigger event log;
 * ..PppPppCPppSPpppppcPpp...
 * P : rpm trigger
 * p : skipped rpm trigger
 * C : falling CAM edge (ie. cam sync)
 * c : rising CAM edge
 * S : secodary trigger (62BTDC) with cam sync
 * s : skipped secondary trigger
 *
 * CAM signal is NOT physically connected (eg. to PIND0 INT0),
 * just used to mask out every 2nd crank-home pulse.
 * So we have 135 evenly spaced crankteeth and one precisely positioned (no cambelt sloppyness)
 * pulse every second crankrot, at 62deg BTDC cyl1.
 * Config:
 * WARNING: http://www.vems.hu/wiki/index.php?page=InputTrigger%2FAudiTrigger is more uptodate!!
 * primary_trigger=FB	# coil-type, rising edge, but filtering disabled (bit2=0),
 * secondary_trigger=19	# enabled; rising edge; camsync mode
 * tooth_wheel=87  		# dec 135
 * the above triggers the AUDITRIGGER 135 + 1 tooth code
 * trigger_tooth=1C
 * another_trigger_tooth=36 # (dec 54). 54*4 < 216 , so tooth_cnt fits well-> we just stop counting at 0xFF;
 * ignchmax=04		# 5 cyl
 * alternate=14		# or 04; since 5 injector banks used.
 * 
 * Multitooth code is disabled, so:
 ** We use the struct multitooth_t to store any internal state / counter.
 ** There is MUCH space there, like tooth_cnt, uint16_t average_tooth_period
 ** (actually, currently in etime in stable, only split to struct multitooth in trunk)
 **  We can even use up engine.status_trig & _BV(cam_sync3) if we like.
 * 
 * Return early (not even time captured!) from PRIMARY_TRIGGER irq,
 * except for mt->tooth_cnt == 54,108,162,216
 * Since we don't actually care about actual tooth_cnt (like in multitooth), we might just count in 2 bytes,
 * lower byte from 53..0, higher byte counts cylinders. But igncount does this for us anyway
 * For these "trigger-teeth" we do everything exactly like coil-type traditional trigger
 * We make sure igncount is updated properly (in dispatcher.c igndeact).
 * We do it the same way (actually same code) as the working coil-type trig + camsync code.
 * 
 * TODO: check config.tooth_wheel_twidth1 and config.tooth_wheel_twidth2 and find out the proper values;
 */
#endif

/* let's outsorce (static inline) the 
if ((config.trigger & _BV(coil_trig)) || got_missing_tooth)
part, at least the beginning of it.
We should support at least 2 ign output combinations:
The active period can be AFTERTRIGGER (like EDIS) or 
BEFORESPARK, like dummyignition. 
We want separate config bits for
AFTER / BEFORE and TRIGGER / SPARK, 
but it's OK to only allow 11 and 00 combinations for now.
The length of the active period should be in dwell, filled by userspace.
That calculation is different for EDIS, no problem:
I suggest an independant config bit for that.
No problem if there are useless combinations
of config, we'll write checklists for the popular ones.

This should be read together with dispatcher.c actions and edis.c.
*/
// pass lower 16 bits of etime.capturedtime_prev
// returns new engine.ignstate

/*!
 * \ingroup g_ign_timing
 * \brief Determine when to do ignition spark.
 *
 * \param capturedt  lower 16 bits of etime.capturedtime_prev
 *
 * \return engine.ignstate  new state of ignition
 */
static inline uint8_t ign_trigger_here(uint16_t capturedt){
	uint32_t t32;
	uint16_t per, delay, dwell, sparkat;
	uint8_t igntype, new_early, en_status_ext;
	uint8_t ignstate_copy = engine.ignstate;  // make a copy of ignstate
	
	// dwell needed in EDIS branch too (holds SAWlen in that case)
	dwell = engine.dwell;
	igntype = config.ign_out & 0xF0;
switch(igntype){
  case IGNTYPE_DUMMY:	// the most laborsome case first
	// only BEFORESPARK is allowed for now, 
	// although this could be split according to config.ign_out bits 
	// (if other combinations are useful)

	// the idea is that IGNACT can be scheduled from either previous spark,
	// or from ign_trigger_here(), but
	// IGNDEACT should not be missed either: 
	// IGNDEACT will be scheduled usually from IGNACT
	// but from trigger if IGNACT came before trigger.
	// In any case: using engine.sparkat timing value (which is calculated here) for IGNDEACT .

	// draw a diagram that shows outputsignal vs time, 
	// make some vertical lines (milestones) visible:
	// previous spark, trigger, spark
	// 0, 1, 2 state transition is trivial
	// 3, 1, 2 signal looks similar (dwell intended to be started before trig, but trig won)
	// 3, 4, 2 dwell is really started before trig

	// Interesting situ is when dwell becomes > (tdcdelay-ignadv)
	// individual flags in a bitmask BAD (confusing and complex, believe me, mik)
	// state-machine good (write these states into the diagram)
	// engine.ignstate:
	// 0: nothing done (the trigger will come and schedule IGNACT)
	// 1: IGNACT scheduled
	// 2: IGNACT done, IGNDEACT scheduled
	// 3: prev. spark scheduled next IGNACT (because it's preferred before next trigger is expected)
	// 4: spark-scheduled IGNACT done before trigger (so trigger will need to schedule IGNDEACT)
	// the race between trigger and IGNACT-event means 3=>4 or 3=>1.
	// at 3=>1: (the rare case) when trigger comes before IGNACT (scheduling from prevspark was not necessary) 
	// Trig does not schedule IGNACT, since IGNACT already in the heap. Nor IGNDEACT, since
	// IGNACT will schedule IGNDEACT
	// at 3=>4: IGNACT executed before trigger. Note: in this case
	// IGNACT did NOT schedule IGNDEACT, so trigger will

	// The whole thing is simpler than it sounds if your diagram is correct and shows
	// the possible states and what schedules what.
/* This ASCII art shows the trigger (t) between 2 sparks (s).
s     t       s
-_0____1__--2--_	lowRPM, simple case: ignact scheduled from trigger

-_3____1__--2--_	ignact was scheduled from prevspark, but trig came earlier

-_3_-4-2-------_	ignact was scheduled from prevspark, and ignact before trigger
*/
	// ------------ now let's do the thing: --------------
	// Check when spark should be done (spark always timed relative to trigger):
	// We could do it in userspace, but 
	// multiplying with the fresh period value should not be too prohibitive

	// we take care about delay fitting into uint16_t
#define DSCL (2)
#define PDSCL (1<<DSCL)
	per = engine.rpm_period >> 16;
	if((per>>8) == 0){ // rpm_period fits into 24 bits
		if((uint8_t)per == 0){	// we like the period that fits into 16 bits
			per = (uint16_t)engine.rpm_period;
			t32 = mult16_16(per, engine.trig2spark);
			if(lsr24(t32) < 16){	// the NICE branch, no return
				delay = div4096(t32);
			} else {
#ifdef BENCHMARK
				benchstats.ocr0cnt++;
#endif
				ign_deactivate();	// it should not normally have been started though
				return 0;
			}

		}else if((uint8_t)per < PDSCL){	//  another chance via DownSCaLing
			per = (uint16_t)( engine.rpm_period >> DSCL);
			t32 = mult16_16(per, engine.trig2spark);
			if(lsr24(t32) >= (16/PDSCL)){	// too much delay, sorry
				ign_deactivate();	// it should not normally have been started though
				return 0;
			}
			// at this low RPM value doing <<DSCL after div (on 16bits) would be OK
			delay = div4096(t32 << DSCL);	// still acceptable
		} else { // sorry, with this low RPM try no further
			ign_deactivate();	// it should not normally have been started though
			return 0;
		}
	} else {
		ign_deactivate();
		return 0;
	}
	
	// actually, appr. (delay - dwell) will be scheduled, that could be checked
	// would gain appr. 1..4 more msec on top of 131 msec
	// this way 100 degree will work down to 127 RPM
	if(delay > EVENT_MOST_FUTURE){
		return 0;
	}
	// we must not assume that the flag previously seen by
	// dispatcher was the same, as we just calculated

	// was_early = engine.status_ext & ~_BV(early_dwellstart);

	// now delay tells (4 usec) when spark is needed
	// make a precalculated condition:
	// condition referred to as early_dwellstart is just a hint (not from userspace)
	// even a few hundred microsec error in the threshold which determines 
	// would be a noissue, since
	// dwell is not critical usually (EDIS is another branch altogether)

	// reusing variable per: dwell enlarged with measured interrupt latency:
	per = dwell + etime.now - capturedt;
	
	// do we want some histeresys in the early_dwellstart flag? NO.
	en_status_ext = engine.status_ext;
	
	if(per >= delay){
		new_early = _BV(early_dwellstart);
		en_status_ext |= new_early;
	} else {
		new_early = 0;
		en_status_ext &= ~_BV(early_dwellstart);
	}
	// tell IGNDEACT (dispatcher.c) how to act:
	engine.status_ext = en_status_ext;

	// last capturedtime_prev is also suitable, if we are called
	// from the Nth tooth after the missing tooth (in a multitooth setup)
	sparkat = capturedt + delay;
	engine.sparkat = sparkat;	// stored globally, for dispatcher

	// the coices in short:
	// engine.ignstate == 0: schedule IGNACT; engine.ignstate = 1;
	// engine.ignstate == 3: IGNACT already scheduled; engine.ignstate = 1;
	// engine.ignstate == 4: schedule IGNDEACT; engine.ignstate = 2;

#ifdef BENCHMARK
	// TODO: make a log or histogram for 
	// (ignstate_copy + new_early>>N)
	// since new_early is 128 or 0
#endif
	// TODO: when to cut spark? (eg. crazy period value?)
	if(ignstate_copy | new_early){	// bitwise | faster than logical ||
		ign_activate();  // why is this here?
#ifdef V3_IGN_DRIVE_A
		test_ign_turnon();
		test_ign_deactivate_at(sparkat);
#endif
		event_insert(sparkat, IGNDEACT);
		return 2;
	} else {
		// (ignstate_copy == 0) && (new_early == 0)
		// ign_deactivate();	// not needed
		event_insert(sparkat - dwell, IGNACT);
		return 1;
	} 

			// What should we do, if trigger comes while state==2 (insane, but...)
			// What if it comes when state==1? (even more insane)
			// Note that in state 3 it's guaranteed that IGNACT was 
			// scheduled from prev spark (and not executed yet)
			// since IGNACT would set engine.ignstate to 4 
			// the scheduled IGNACT event will schedule IGNDEACT according to globally stored sparkat
			// so there's no real need to do anything here.
	
	// this replaces the edis stuff, no need for #ifdef EDIS:
  case IGNTYPE_EDIS:
		// hopefully we are at the falling edge of PIP,
		// and configured in coil (not multitooth) mode.
		// dwell holds SAW
		ign_activate();
		// maybe event_insert_relative instead?
		event_insert(etime.now + dwell, EDIS_STOP_SAW);
		return 6;
  default:	// nothing to do
		return 5;
} // switch

} // ign_trigger_here()


/*
  Keep track of engine phase
  the argument is 0 when not using a tooth wheel trigger
*/
static inline void update_engphase(uint8_t wheel_stat) {
	uint8_t phase, rea;
	phase = engine.engphase;

	if(wheel_stat & _BV(TW_MISSING)) {
		// compensate for missing tooth (teeth)
		phase += config.tooth_wheel_twidth2;
	} else {
		phase += config.tooth_wheel_twidth1;
	}

	rea = config.reset_engphase_after;
	if (phase >= rea) {
		phase = phase - rea;
	}

	engine.engphase = phase;
}

/*
  called for every TW_TRIG (or coil-type event)
  Synhronize engine phase to cam trigger
  updates engine.status_trig (est)
*/
static inline void cam_trigger_sync(void) {
	uint8_t est = engine.status_trig;
		
	if (est & _BV(cam_sync)) {
		uint8_t engine_engphase_sync;
		est &= ~_BV(cam_sync);
		// only called for TW_TRIG, so 2 camrot is exaclty (config.ignchmax << 1) + 2, we might want more:
		etime.cam_sync_cnt = (config.ignchmax << 2) + 7;
		// (config.tooth_wheel << 1) + 1;

		engine_engphase_sync = engine.engphase_sync;
		if (est & _BV(engine_in_sync)) {
			// engine already in sync, verify that nothing bad happened
			if (engine.engphase != engine_engphase_sync) {
				// lost sync, how did this happen?
				// what should be trusted, the crank trigger or the cam trigger?
				engine.engphase = engine_engphase_sync;  // get back on track

				// if the crank trigger is trusted, then fallback to wasted spark?
				//engine.status_trig &= ~_BV(engine_in_sync);
#ifdef BENCHMARK
				benchstats.engine_phase_err++;
#endif
			}
		} else { // sync with secondary trigger to calculate engine phase
			engine.engphase = engine_engphase_sync;
			est |= _BV(engine_in_sync);
		}
	} else {
#ifdef BENCHMARK
		if (!(est & _BV(engine_in_sync))) {
			// measure the time to get cam sync
			benchstats.engphase_nosync++;
		}
#endif

		// if no trigger from the cam for N primary triggers,
		// then we can't guarantee that the engine still is in sync
		if (etime.cam_sync_cnt) {
			etime.cam_sync_cnt--;
		} else {
			// lost cam sync
			est &= ~_BV(engine_in_sync);
		}
	}
	engine.status_trig = est;
}

/** 
 * Debug functions to dump trigger information to serial comm
 */
#ifdef AUDITRIGGER
// simple 'p' is not enough for serious debugging.
// Binary output makes more verbose debugging possible, use gtkterm or brayterm to view:
// 128..255: toothcount + 128 (view in hex, like 0x80..0xFF)
// 'x' : toothcount > 128 (not yet in sync)
// 'P' : primary trigger tooth
// "\r\nS" : secondary trigger
uint8_t audi_pri_trig_toothcnt(uint8_t i)
{
	uint8_t toothcount;
	// fetch toothcnt in privch.a;
	toothcount = cmd_queue[cmd_queue_head].privch.a;
	if( toothcount < 128 ){
		return (0x80 | toothcount);
	}

	return 'x';
}
uint8_t audi_pri_trig_P(uint8_t i)	{ return('P'); }
uint8_t audi_sec_trig(uint8_t i)	{ return('\n'); }

#endif

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
// PRIMARY TRIGGER
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////


/* cranksensor
* coiltype (one pulse for one ignition-event - every 180 crankdegree on a 4cyl)
* multitooth eg. 60-2
*/
/*!
 * \ingroup g_engine_timing
 * \brief Interrupt Service for PRIMARY_TRIGGER interrupt TCNT3 drives eventqueue.
 *
 * \return none
 */
SIGNAL (SIG_PRIMARY_TRIGGER) {

	uint16_t latency;
	uint32_t capturedtime, period32; // 16 bits could be OK but we must take care of overflow to distinguish very low rpm
	uint8_t sched_squirt;
	uint16_t ch1_time, ch2_time;
	uint8_t tooth_wheel_trigger;
	uint8_t new_event;
	uint8_t config_primary_trigger=config.primary_trigger;

#ifdef AUDITRIGGER
	if( auditrigger_configdisabled() == 0 ){ // auditrigger enabled
		uint8_t tooth_cnt = etime.tooth_cnt;	// write back just before return

		if( tooth_cnt ){
			// TODO: this if should be dropped after DEBUGGING
			if (audi_trigger_debug_verbose()){
				// we store toothcnt in privch.a;
				pushfuncpar2( audi_pri_trig_toothcnt, 1, NULL, tooth_cnt );
			}

			if( (engine.status_trig & _BV(cam_sync3)) == 0 ){
				// if no camsync yet (tooth_cnt in [another_trigger_tooth..255] range), we decrease
				// tooth_cnt every N'th time (using true random generator ICR16)
				// to avoid unintentional sync if no cam shows up within 201 teeth.
				// We might still go "random"-sync-mode after appr. 804 teeth to allow limphome
				// (maybe not, if camsync code prevents operation without camsync)
				if( (tooth_cnt < config.another_trigger_tooth )
				// should be fine without the ICR16 part, toothcnt stays at 0xFF before camsync
				// || ((ICR16 & 3)==0)
				){
					// FASTPATH:
					tooth_cnt --;
				}
			} else { // YEAH, CAMSYNC flagged !!!
				// http://www.sjmautotechnik.com/trouble_shooting/wave/RPMTDCId.jpg
				// shows that cramhome comes very short after tooth_cnt==1 (remember downcounting)
				// If "cramhome" came after tooth_cnt==2 (somewhat early)
				// we don't want to reset to 0 one-tooth early,
				// so we postpone the operation for 1 tooth.
				// note that this is NOT -1,0,+1 tooth allowance,
				// just 1 (cramhome coming early) or 0 (normal)
				
				// tooth_cnt is normally == 1
				// the >= config.another_trigger_tooth allows initial sync
				if( tooth_cnt == 1 || (tooth_cnt >= config.another_trigger_tooth ) ){
					tooth_cnt = 0;	// sync here, fine
					engine.status_trig &= ~_BV(cam_sync3);
				} else if( tooth_cnt == 2 ){
					// we also accept 2, don't even fire wheel_err
					// maybe count this branch elsewhere ?
					tooth_cnt --;	// == 1, actually
					// note that we do NOT clear the flag, leave it for next primtrig pulse:
					// engine.status_trig &= ~_BV(cam_sync3);
				} else {
					// start a new sequence; dropped events are better than blown engines
					tooth_cnt = 0xFF;
#ifdef BENCHMARK
					benchstats.wheel_err ++;
#endif// BENCHMARK
					engine.status_trig &= ~_BV(cam_sync3);
				}
			}

			// because we already decreased tooth_cnt,
			// trigger_tooth == 0 actually means the 2nd pulse after cramhome is the trigger tooth
			if( tooth_cnt != config.trigger_tooth ){
      			// FASTPATH: note that the interrupt overhead was minimal!
      			etime.tooth_cnt = tooth_cnt;
				return;
			}
		} else { // tooth_cnt counted down to 0
			if (audi_trigger_debug()){
				 pushfunc(&audi_pri_trig_P, 1);
			}
			if( (engine.status_trig & _BV(cam_sync3)) == 0){
      			// for every 54th tooth, act like it's a normal coil type trigger pulse
				tooth_cnt = config.another_trigger_tooth - 1;
			} else {
				// this is a hard decision, probably best this way
				tooth_cnt = 0;
				engine.status_trig &= ~_BV(cam_sync3);
			}
      		etime.tooth_cnt = tooth_cnt;
      		return;
      		// trigger_tooth= decimal 53 not allowed, nor does it make sense.
		}
    etime.tooth_cnt = tooth_cnt;
    }

#endif

	new_event = 0;

	latency = ICR16; // just reusing variable names, do not get confused!
	readTMR();

	ch1_time = etime.now_upper;
	ch2_time = etime.now;
	if(ch2_time < latency){ // now_upper has incremented
		ch1_time --;
	}	
		
	capturedtime = ((uint32_t)ch1_time<<16) | latency;

	// log every trigger event:
	if(debug.enable_flags & _BV(6)){
		pushfuncpar( &srl_direct, 2, (uint8_t *)latency );
	}

/* Engine can be in one of several modes:
   - running, rpm above a threshold => normal_edge + igndelay for ign
   - rpm below a threshold => other_edge + igndelay for ign (note that rpm threshold is settable, so this can be disabled!)
   - no rpm calculation yet, because only 1 edge in the last reasonable window (ign right at other_edge if enabled,)

   Injection can be started right from here for non-timed(simultaneous or alternating)
   or scheduled with some delay (like ignition) if sequential is needed.
   Some immediate injection will be helpful in the last case as well (very early in cranking).
*/


	/* event_insert(capturedtime + (engine.rpm_period*150) >>10);
	   will give you a 150/1024 trigger-period delay.
	*/

  	period32 = capturedtime - etime.ptrig_capturedtime_prev; // 4 usec resolution 
	etime.ptrig_capturedtime_prev = capturedtime;

	if ( 0 == (config_primary_trigger & _BV(ptrig_coil))) { // multitooth
		/* finally we want to keep track of engine (cam) phase like:
		0..720 degrees is mapped to an 8 bit value called engphase.
		phase modulo +toothwidth toothcount reset_igncount_after (dist8cyl/dist4cyl/wasted/fullyseq)
		0..215 216 +3	(36 crankteeth)	27/54/108/216
		0..239 240 +2	(60 crankteeth)	30/60/120/240
		0..239 240 +30	(4	crankteeth)	30/60/120/240
		0..134 135 +3   (45 crankteeth) (5cyl with AUDITRIGGER #defined)
		This will make it possible to time ALAP (As Late As Possible) 
		from the last input pulse before the event - since we 
		always know when the next pulse is expected and when output is due.
		Note that separate injphase and ignphase must be maintained,
		since reset_injselector_after can be different from reset_ignselector_after:
		basically any combinations are allowed, 
		(TBI / seqinj) injection and 
		(distributor / wasted / fullyseq) ignition are independent.
		
		Odd operation: (like oddfire V6)
		an array shows when the TDC-s happen in terms of above engphase variable
		
		*/

	// it is very unlikely that period32 > 0xffff except for the first pulse
	// currently we examine just bits 23:16 with lsr16()
		tooth_wheel_trigger = tooth_detection(
			((lsr16(period32) == 0) ? (uint16_t) period32 : 0xffff), capturedtime);

#ifdef GENBOARDv3
/*
  The following condition is to avoid a bad outcome of a potential
  race at first engine revolution.
  If we get camsync very quickly, but the primary tooth wheel has not
  sync'ed yet, then the trigger-tooth which engine.engphase_sync
  refers to, might be skipped. This results in the next trigger tooth
  will be synced with a wrong engine phase (==bad).
  By expiring the cam sync after N teeth on the primary trigger wheel,
  this problem cannot happen.
*/
		if (etime.cam_trigger_expire) {
			etime.cam_trigger_expire--;
		} else { // this should be rare
			engine.status_trig &= ~cam_sync_mask;
		}
#endif		
	} else {	// normal coil trigger
		engine.rpm_period = period32;
		engine.status_ext |= _BV(new_rpm);  // start a rpm calculation

		// act like it was a TW_TRIG with multitooth, to simplify related conditional expressions
		tooth_wheel_trigger = _BV(TW_TRIG);
	}	// normal coil trigger

#ifdef GENBOARDv3
	// if (config_secondary_trigger & _BV(strig_cam_sync))
	update_engphase(tooth_wheel_trigger);
#endif

	if ( tooth_wheel_trigger & _BV(TW_TRIG)) {
		etime.primary_trigger = capturedtime;  // store when the trigger occured

#ifdef BENCHMARK
		benchstats.trig_prim_cnt++;
#endif

#ifdef GENBOARDv3
		// if (config_secondary_trigger & _BV(strig_cam_sync))
		cam_trigger_sync();
#endif

		// Allow firing if engine_in_sync or not using camsync:
		// We could also allow firing if no camsync for several primary triggers, and user allowed limphome
		// that way chance to start the engine is 50% for 60-2 and 25% for coil-type 4 cylinder
		if( (engine.status_trig & _BV( engine_in_sync ))
			|| (0==(config.secondary_trigger & _BV(strig_cam_sync))) ){
		// when camsync is configured, don't fire anything until camsync
		
#ifndef NOIGN
			// ign_trigger_here() must be called
			// even if(engine.status_ext & _BV(ign_cut_req)) 
			// otherwise we fry coils !!! (think: early_dwellstart)		
			engine.ignstate = ign_trigger_here((uint16_t)capturedtime);

			// we need no alarm2core() if ignoldstate + ignnewstate ==4
			// 3=>1 particularly
			// but it's a rare case, doesn't hurt to call always
			new_event=1;
#endif
		engine_is_running(); //turn on fuelpump etc.

		if (engine.status & _BV(crank)) { // TODO: drop this ??
			engine.status &= ~(_BV(tpsaen) | _BV(tpsden));
			sched_squirt = true;

		} else {

			if (--etime.trigcount == 0) {
				etime.trigcount = config.divider;
				sched_squirt = true;

				if (etime.egocount) // TODO: drop this
					etime.egocount--;

			} else {
				sched_squirt = false;
			}

			if (etime.asecount) {
				// decrease asecount every engine cycle

				// this will eventually leave; currently we don't know the start
				// of an engine cycle and are therefore doing this instead
				if (--etime.ase_cyl_count == 0) {
					etime.ase_cyl_count = (config.config11>>4) + 1;
					etime.asecount--; 
				}
			}
		}

		if (sched_squirt) {
#ifdef GENBOARDv3
			if ((engine.status & _BV(crank)) && (config.alternate & _BV(BATCH_FIRE_CRANKING))) {
				injector_start_all();
			} else {
				uint8_t injch;
				injch = etime.altcount;
				if(injch){	// step down if possible
					injch --;					
				} else {	// or reset to max channel
					injch = config.alternate & 0x0f;
				}
				etime.altcount = injch;
				injector_start(injch); // Starting primary injectors
				
				#ifdef INJECTOR_STAGING
				uint8_t i;
				injch=INJECTORS-SECONDARY_INJ_CHANNELS_NUM;
				i=SECONDARY_INJ_CHANNELS_NUM;
				while(i)
				{
					i--;
					injector_start(injch); // Starting second-stage injectors
					injch++;
				}
				#endif // INJECTOR_STAGING
			}
#else
			if ( (engine.status & _BV(crank)) || (config.alternate == 0) ) {

				// schedule both injectors SIMULTANEOUSLY
				// This will probably be dropped and a more generic port_act
				// scheme will be used (so anyone can turn on any port she likes)

				// no counting and peeping at the pwcalc, pwrun value from time to time. 
				// The only thing to take care is to write the 16 bit (not 8!) value
				// from userspace with interrupts disabled

				injector_start(&inj_port[1], startinj2,  INJBGOPWM, INJBSTOP);
				injector_start(&inj_port[0], startinj1,  INJAGOPWM, INJASTOP);
			} else { // ALTERNATING
				etime.altcount++;
				if (etime.altcount & 0x01) {	// B channel (inj_port2)
					injector_start(&inj_port[1], startinj2,  INJBGOPWM, INJBSTOP);
				} else {	// A channel (inj_port1)
					injector_start(&inj_port[0], startinj1,  INJAGOPWM, INJASTOP);
				}
			} // SIMULTANEOUS or ALTERNATING
#endif
			new_event=1;
		}	// sched_squirt 
	} // allow firing

	}

#ifdef BENCHMARK
	benchstats.primary_trigger_cnt++; // redundant since we have a histogram:
//benchmark_histogram(ic3_histogram, latency, IC3HISTOGRAM_BUCKETS);
#endif

	// Disabling triggering and reenabling somewhat later provides some anti-glitch. 
	if (config_primary_trigger & _BV(trigger_filter)) {
		uint16_t t;
		t = reenable_time(period32);

		if (tooth_wheel_trigger & _BV(TW_MISSING)){
			/* This can only happen with multitooth => crazy user ?
			 * At the missing tooth, the period32 is large at low RPM
			 * => high risk to miss the following tooth. This can work with
			 * 2 missing teeth (reenable after 3/2/2 = 0.75 tooth-time)
			 * but anyone crazy enough to use multitooth with reenable?
			 */
			t >>= 1;
		}

		event_insert(etime.now + t, PRIMARY_TRIGGER_REENABLE);

		if (!(t>>8)) // handler2k will fire in max 256*4us and run alarm2core()
			new_event=1;

		// disable input-capture interrupt (reenabled by the event)
#ifdef GENBOARDv3	
		TIMSK &= ~_BV(TICIE1);
#else
		ETIMSK &= ~_BV(TICIE3);
#endif
		// No need to clear ICF flag: done after reenabling!
		// 1..2 msec reenable delay is ok up to
		// about 7000 RPM 8 cylinder engine or 14000 RPM 4 cyl,

	} else {	// other values: FF=don't disable, 
		// clear interrupt flag
#ifdef GENBOARDv3
		TIFR = _BV(ICF1);
#else
		ETIFR = _BV(ICF3);
#endif
	}


	if (new_event)
		alarm2core();
		

} // ICR handler




///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
// SECONDARY TRIGGER
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

// the interrupt handler for int6 and ic3 will probably be very different.
// they are split from the beginning...


/***************************************************************************/
/*** INTERRUPT: engine-position signal (triggers on falling edge)        ***/
/* TODO:
 check bit TACH_ON_INT6_OR_7 in config13: 
 - If 1, use ICR3 (timing.c) for more precise (eliminated interrupt latency) position signal.
 - If 0, we have no ignition, and injection can be initiated from here, same rpm_period must 
 be provided (4usec) as from timing.c (almost the same handler, except ICR3 is not available,
 must use readTMR2() and disable the actual interrupt.
 - theoretically we could accept INT6 or IC3 interrupt regardless of configuration.
 The user NEEDS to make sure that only one gets triggered, the other NOT 
 see ICR3 in timing.c 
*/
/***************************************************************************/

uint8_t trig_on_dwellstart;
#define ALIEN_DWELL_REENABLE_DELAY 60   // 60 * 4 us = 240 us

#ifdef GENBOARDv3 

// setup cam sync trigger edge
static inline void cam_sync_set_trigger_edge(void) {
	if (config.secondary_trigger & _BV(strig_both_edges)) {
#warning "TODO: check if the cranking flag is set correctly"
		if (engine.status & _BV(crank)) {
			TCCR3B ^= _BV(ICES3);  // the other edge
			return;
		}
	}

	// be sure the edge is set correctly
	if (config.secondary_trigger & _BV(trigger_edge))
		TCCR3B |= _BV(ICES3);  // rising edge
	else
		TCCR3B &= ~_BV(ICES3);  // falling edge
}

/*!
 * \ingroup g_engine_timing
 * \brief Interrupt Service for SECONDARY_TRIGGER interrupt engine-position signal.
 *
 * \return none
 */
SIGNAL (SIG_SECONDARY_TRIGGER) {
	uint32_t capturedtime, period32; // 16 bits could be OK but we must take care of overflow to distinguish very low rpm
	uint16_t ch1_time, ch2_time;
	uint16_t t, reenable;
	uint8_t trig_edge;
	uint8_t config_secondary_trigger, filter_active, conf_sectrig_edge;
#ifndef WBO2
	uint16_t latency;
#endif

#ifdef AUDITRIGGER

       if (audi_trigger_debug()) pushfunc(&audi_sec_trig, 1);

#endif

	readTMR();
	ch1_time = etime.now_upper;
	ch2_time = etime.now;

#ifndef WBO2
	// when not using tmr3 as 8-bit pwm, the capture unit can be
	// used to get better precision

	latency = ICR3 >> 6; // convert to 4 usec time unit
	t = ch2_time & 0x03FF;
	if (t < latency) { // now has incremented, take care of both "now" and "now_upper"
		uint16_t k;
		k = ch2_time - 0x0400;
		k = (k & 0xFC00) | latency;

		if (k > ch2_time) // now_upper has incremented
			ch1_time--;

		ch2_time = k;
	} else {
		ch2_time = (ch2_time & 0xFC00) | latency;
	}

#endif

	capturedtime = ((uint32_t)ch1_time<<16) | ch2_time;

#ifdef BENCHMARK
	benchstats.secondary_trigger_cnt++;
#endif

	period32 = capturedtime - etime.strig_capturedtime_prev;
	etime.strig_capturedtime_prev = capturedtime;

	// cache variable and its bits:
	config_secondary_trigger = config.secondary_trigger;
	
	conf_sectrig_edge = config_secondary_trigger & _BV(trigger_edge);
	
	filter_active = config_secondary_trigger & _BV(trigger_filter);
	if (filter_active) {
		reenable = reenable_time(period32);
	}

	// determine which edge trigged this interrupt handler
	trig_edge = TCCR3B & _BV(ICES3);

	if (config_secondary_trigger & _BV(strig_cam_sync)) {
//********************************************************************************//
// cam sync
//********************************************************************************//
	// for fiatstilo, only consider this pulse, if( period32 > 1.5 * engine.rpm_period )
	  if( 0 == (config.secondary_trigger & _BV(strig_fiatstilo))
	   	|| (period32 > (engine.rpm_period + (engine.rpm_period >> 1)))
	  ){
		if (trig_edge) { // rising edge
			engine.engphase_sync = config.cam_sync_r_edge_phase;
			// doing the XOR with trig_edge here (with the branch help):
			conf_sectrig_edge ^= _BV(trigger_edge);
		} else { // falling edge
			engine.engphase_sync = config.cam_sync_f_edge_phase;
		}

		cam_sync_set_trigger_edge();
		engine.status_trig |= cam_sync_mask;
		etime.cam_trigger_expire = config.tooth_wheel | 0x0F;

		// only update cam_period on the primary trigger edge
#if(ICES3 == trigger_edge)
#warning "We could use bitwise XOR, no need for trig_edge_shifted"
#endif
		// if one configures falling edge, he'll never notice
		// that ^ becomes | (same as ||) in earlier revision condition
		// TODO: rethink this
		if (conf_sectrig_edge == 0) {
			// never used together with etime.strig_capturedtime_dwell (reuse?) 
			engine.cam_period = period32;
		}
	   }
	} else {
//********************************************************************************//
// alien ignition advance
//********************************************************************************//
		// no logical XOR operator in C, right?
		if ((trig_edge | conf_sectrig_edge)	
			&& ((trig_edge==0) || (conf_sectrig_edge ==0))){
			// start of dwell

			etime.strig_capturedtime_dwell = ch2_time;

			// disable interrupt for a short time
			reenable = ALIEN_DWELL_REENABLE_DELAY;

		} else {
			// coil on fire
#ifdef BENCHMARK
			benchstats.trig_sec_cnt++;
#endif

			spark_time = capturedtime - etime.primary_trigger;
			spark_rpm = engine.rpm_vhr;
			new_spark = 1; // signal function in userspace to calculate advance

			t = ch2_time - etime.strig_capturedtime_dwell;
			t >>= 4;  // convert to base 64 us
	
			if (t>>8)
				engine.alien_dwell = 0xff;
			else
				engine.alien_dwell = t;
		}

		// setup trigger edge
		TCCR3B ^= _BV(ICES3);  // the other edge
	}

//********************************************************************************//
// interrupt filtering
//********************************************************************************//

	if (filter_active) {
		ETIMSK &= ~_BV(TICIE3);  // disable interrupt

		event_insert(ch2_time + reenable, SECONDARY_TRIGGER_REENABLE);
		if (!(reenable>>8)) // handler2k will fire in max 256*4us and run alarm2core()
			alarm2core();

	} else {
		ETIFR = _BV(ICF3);
	}

}


#else

// GENBOARDv2 interrupt 6
/*!
 * \ingroup g_engine_timing
 * \brief Interrupt Service for interrupt 6 engine-position signal.
 *
 * \return none
 */
SIGNAL(SIG_ALIEN_MEASURE) {
	uint32_t period32, capturedtime;
	uint16_t t;

	readTMR();
	capturedtime = now32(); 

#ifdef BENCHMARK
	benchstats.secondary_trigger_cnt++;
#endif

	if (trig_on_dwellstart) {  // start of dwell

		// set interrupt to trigger on the normal edge
		if (config.secondary_trigger & _BV(trigger_edge))
			EICRB |= _BV(ISC60);  // rising edge
		else
			EICRB &= ~_BV(ISC60);  // falling edge
		
		etime.strig_capturedtime_dwell = etime.now;
		trig_on_dwellstart = 0;

		if (config.secondary_trigger & _BV(trigger_filter)) {
			EIMSK &= ~_BV(INT6);  // disable interrupt
			event_insert(etime.now + ALIEN_DWELL_REENABLE_DELAY, SECONDARY_TRIGGER_REENABLE);
                        // ALIEN_DWELL_REENABLE_DELAY < 256*4us, therefore it is
                        // necessary to run alarm2core()
			alarm2core();
		} else {
			EIFR = _BV(INTF6);
		}
		return;

	} else {  // end of dwell, coil was fired

		// set interrupt to trigger on the inverted edge
		if (config.secondary_trigger & _BV(trigger_edge))
			EICRB &= ~_BV(ISC60);  // falling edge
		else
			EICRB |= _BV(ISC60);  // rising edge

		trig_on_dwellstart = 1;
	}


	spark_time = capturedtime - etime.primary_trigger;
	spark_rpm = engine.rpm_vhr;
	new_spark=1; // signal function in userspace to calculate advance

	t = etime.now - etime.strig_capturedtime_dwell;
	t >>= 4;  // convert to base 64 us
	
	if (t>>8)
		engine.alien_dwell = 0xff;
	else
		engine.alien_dwell = t;
	

	period32 = capturedtime - etime.strig_capturedtime_prev;
	etime.strig_capturedtime_prev = capturedtime;

	if (config.secondary_trigger & _BV(trigger_filter)) {
		uint16_t t;
		EIMSK &= ~_BV(INT6);  // disable interrupt

		t = reenable_time(period32);
		event_insert(etime.now + t, SECONDARY_TRIGGER_REENABLE);
		if (!(t>>8)) // handler2k will fire in max 256*4us and run alarm2core()
			alarm2core();

	} else {
		EIFR = _BV(INTF6);
	}
}

#endif



/*!
 * \ingroup g_engine_timing
 * \brief Initialize primary timing input capture.
 *
 * \return none
 */
static inline void primary_trigger_init(void) {

#ifdef GENBOARDv3
// configure timer 1
	uint8_t t;

	TCCR1A = 0;  // this should be zero by default

	// ICNC1: Input Capture Noise Canceller ON (page 133.)
	// CS1 2:0=001 sets prescaler to 1 (page 134.)
	t = _BV(ICNC1) | _BV(CS11) | _BV(CS10);
	
	if (config.primary_trigger & _BV(trigger_edge)) 
		t |= _BV(ICES1);

	TCCR1B |= t;

	TIMSK |= _BV(TICIE1);	// ICR1 enable
	TIFR = _BV(ICF1); 	// clear interrupt flag

#else
// configure timer 3
	uint8_t t;

	TCCR3A = 0;  // this should be zero by default

	// ICNC3: Input Capture Noise Canceller ON (page 133.)
	// CS3 2:0=001 sets prescaler to 1 (page 134.)
	t = _BV(ICNC3) | _BV(CS31) | _BV(CS30);
	
	if (config.primary_trigger & _BV(trigger_edge)) 
		t |= _BV(ICES3);

	TCCR3B |= t;

	ETIMSK |= _BV(TICIE3);	// ICR3 enable
	ETIFR = _BV(ICF3); 	// clear interrupt flag

#endif

}





/*!
 * \ingroup g_engine_timing
 * \brief Initialize secondary timing input capture.
 *
 * \return none
 */
static inline void secondary_trigger_init(void) {

#ifdef GENBOARDv3
// configure timer 3
// BEWARE: wbo2 uses same timer
	uint8_t t;

/*
Found the problem with the bad behaving alien ignition measurement:
When WBO2 is enabled (defined), tmr3 is configured as 8-bit pwm with
0xff as top and div1 prescaler - 16 usec period:
The secondary trigger interrupt must be
serviced within 16 usec zzzzzzzzzzz

Solutions:
- disable wbo2 when doing alien ignition logging
- can the timer be configured differently while still allowing wbo2 to
  be functional? Our LM324 and 510 Ohm * 2.2uF filter allows 256 usec if needed
- use only readTMR, looses some precision but that ain't critical
  for ex. camsync. (or ignadv where we use statistical averaging)
- use readTMR and only use TCNT3 for correcting the lower bits

*/

#ifndef WBO2
	TCCR3A = 0;  // this should be zero by default
#endif
	// ICNC3: Input Capture Noise Cancaller ON (page 133.)
	// CS3 2:0=001 sets prescaler to 1 (page 134.)

#ifdef WBO2  // take care not to destroy anything
	t = _BV(ICNC3);
#else
	t = _BV(CS30) | _BV(ICNC3); // no wbo2, setup prescaler too
#endif

	if (config.secondary_trigger & _BV(trigger_edge)) {
		// if cam sync, trig on both edges and IC3-pin is low,
		// then setup interrupt to trig on rising edge
		if ((config.secondary_trigger & _BV(strig_cam_sync)) &&
		    (config.secondary_trigger & _BV(strig_both_edges))) {
			if (!(PINE & _BV(PINE7))) // IC3 pin
				t |= _BV(ICES3);
		} else {
			t |= _BV(ICES3);
		}
	}

	TCCR3B |= t;

	if (!(config.secondary_trigger & _BV(strig_disable))) {
		ETIMSK |= _BV(TICIE3);	// ICR3 enable
	}

	ETIFR = _BV(ICF3); 	// clear interrupt flag
#else
// configure interrupt 6
	if (!(config.secondary_trigger & _BV(strig_disable))) {
		if (config.secondary_trigger & _BV(trigger_edge))
			EICRB |= _BV(ISC61) | _BV(ISC60);  // rising edge
		else
			EICRB |= _BV(ISC61);  // falling edge

		EIMSK = _BV(INT6);  // enable interrupt
	} 
	EIFR = _BV(INTF6);  // clear interrupt flag

#endif

}









/* ---------------------------- */
/*!
 * \ingroup g_engine_timing
 * \brief Initialize timing variables.
 *
 *
 * \return none
 */
void timing_init()
{
	// the toothwheel is initially not in sync.
	// works for AUDITRIGGER as well (0 == auditrigger_configdisabled())
	etime.tooth_cnt = 255;

	etime.trigcount = config.divider;
	etime.ase_cyl_count = (config.config11>>4) + 1;
	etime.cam_sync_cnt = 0;
	etime.cam_trigger_expire = 0x0F;
	
	engine.igncount = (config.secondary_trigger & _BV(strig_cam_sync)) ? 0 : config.ignchmax;

	/* 
	if(config.primary_trigger & _BV(ptrig_coil){
		engine.igncount = 0;
	} else { 		
		engine.igncount = config.ignchmax;
		
		// TODO: check that we don't use this before sane vale
		// it seems that TW_MISSING will set it fine before first usage
		// engine.igncount = 0xFF;
		// the FF value was ment as a flag for multitooth out-of-sync 
		// but it wasn't ever used that way, it just bit coil-type guys		
	}
	*/

	trig_on_dwellstart = 0;

	primary_trigger_init();
	secondary_trigger_init();

}






